/*
 * Copyright (C) 2002  Robert Morris (rtm@lcs.mit.edu)
 *                     Thomer Gil (thomer@lcs.mit.edu)
 * 		       Benjie Chen (benjie@lcs.mit.edu)
 *                     Athicha Muthitacharoen (athicha@lcs.mit.edu)
 *                     Massachusetts Institute of Technology
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.
 */

#include "nfstrans.h"
#include "ranges.h"
#include "log.h"
#include "logiter.h"
#include "debug.h"
#include "cfsrw.h"
#include "sfscrypt.h"

unsigned ss_thresh = ENV("SNAPSHOT_THRESH",20);
unsigned ss_interval = ENV("SNAPSHOT_INTERVAL",60);

unsigned cfsrw::blocksize = 4096;

static inline ptr<fattr3>
new_attr()
{
  ptr<fattr3> fa = New refcounted<fattr3>;
  fa->mode = fa->nlink = fa->uid = fa->gid 
           = fa->size = fa->fsid = fa->fileid
           = fa->rdev.major = fa->rdev.minor = 0;
  fa->atime.seconds = fa->atime.nseconds = 0;
  fa->mtime.seconds = fa->mtime.nseconds = 0;
  fa->ctime.seconds = fa->ctime.nseconds = 0;
  return fa;
}

static inline ptr<fattr3>
new_attr(sattr3 *sa)
{
  ptr<fattr3> fa = New refcounted<fattr3>;
  fa->mode = fa->nlink = fa->uid = fa->gid =
    fa->size = fa->fsid = fa->fileid = 
    fa->rdev.major = fa->rdev.minor = 0;
  fa->atime.seconds = fa->atime.nseconds = 0;
  fa->mtime.seconds = fa->mtime.nseconds = 0;
  fa->ctime.seconds = fa->ctime.nseconds = 0;

  if(sa->mode.set)
    fa->mode = *sa->mode.val;
  if(sa->uid.set)
    fa->uid = *sa->uid.val;
  if(sa->gid.set)
    fa->gid = *sa->gid.val;
  if(sa->size.set) {
    fa->size = *sa->size.val;
  }
  return fa;
}

static inline void
fix_fattr (recordp r, fattr3& f)
{
  inode_log_record *e;
  if (!(e = r->template record<inode_log_record>()))
    return;
  if (e->itype() == inode_log_record::T_FILE)
    f.type = NF3REG;
  else if (e->itype() == inode_log_record::T_DIR) {
    f.type = NF3DIR;
    f.size = 512;
  }
  else if(e->itype() == inode_log_record::T_SYMLINK)
    f.type = NF3LNK;
  else
    f.type = NF3REG;
  // this is a hack; if type is DIR and link count is 0, don't
  // increment it for "." since it has been removed already.
  // otherwise, we increment nlink by 1.
  if(f.type == NF3DIR && f.nlink > 0)
    f.nlink += 1; // for . (.. doesn't refer to this i-number)
  f.fsid = 0;
  f.used = f.size;
  f.fileid = make_fileid(e->inum());
}

static inline ptr<fattr3>
get_fattr3 (recordp e, ptr<cattr> ca)
{
  ptr<fattr3> fa = New refcounted<fattr3>;
  ca->ca2fa (*fa);
  fix_fattr (e, *fa);
  return fa;
}


cfsrw::cfsrw (ref<view> v, chordID vid, bigint& verf,
              callback<void, nfscall *, void *, nfs_fh3>::ref r)
  : _view(v), _vid(vid), _verf(verf), getreply(r)
{
}

bool
cfsrw::uid (nfscall *sbp, chordID &u)
{
  ptr<log> l = mylog (sbp);
  if (l) {
    u = l->head ();
    return true;
  }
  return false;
}

bool
cfsrw::readonly (nfscall *sbp)
{
  ptr<log> l = mylog (sbp);
  int proc = sbp->proc();
  if (!l && 
      (proc == NFSPROC3_CREATE || proc == NFSPROC3_WRITE ||
       proc == NFSPROC3_REMOVE || proc == NFSPROC3_SETATTR ||
       proc == NFSPROC3_SYMLINK || proc == NFSPROC3_MKDIR ||
       proc == NFSPROC3_RMDIR || proc == NFSPROC3_RENAME ||
       proc == NFSPROC3_LINK || proc == NFSPROC3_MKNOD ||
       proc == NFSPROC3_COMMIT))
    return true;
  return false;
}

void
cfsrw::dispatch (nfscall *sbp, ptr<materialized_view> mv, bool need_snapshot)
{
  sessionp s = get_session (sbp, mv, need_snapshot);
  if (!s)
    return;

  sfs_aid aid = sbp->getaid ();
  ptr<log> l = mylog (sbp);
  int proc = sbp->proc();

  if (readonly (sbp)) {
    sbp->error (NFS3ERR_ROFS);
    return;
  }
 
  switch(proc) {
  case NFSPROC3_GETATTR:
    {
      WARN(1) << "NFSPROC3_GETATTR\n";
      nfs_fh3 *nfh = sbp->template getarg<nfs_fh3> ();
      chordID cid = nfs2cid (nfh);
      find_inode (s, cid, wrap(this, &cfsrw::getattr_found_inode, s, cid));
    }
    break;
  case NFSPROC3_ACCESS:
    {
      WARN(1) << "NFSPROC3_ACCESS\n";
      access3args *a = sbp->template getarg<access3args> ();
      chordID cid = nfs2cid (&(a->object));
      find_inode (s, cid, wrap(this, &cfsrw::access_found_inode, s, cid));
    }
    break;
  case NFSPROC3_LOOKUP:
    {
      diropargs3 *dirop = sbp->template getarg<diropargs3> ();
      chordID cid = nfs2cid (&(dirop->dir));
      if (dirop->name == ".") {
        lookup3res nfsres(NFS3_OK);
        nfsres.resok->object = dirop->dir;
	reply (sbp, &nfsres);
      }
      else if (dirop->name == "..") {
        WARN(1) << "NFSPROC3_LOOKUP .. in " << cid << "\n";
	find_dir_parent(s, cid, wrap(this, &cfsrw::find_dir_parent_done, s));
      }
      else {
        WARN(1) << "NFSPROC3_LOOKUP\n";
        lookup(s, cid, dirop->name,
	       wrap(this, &cfsrw::lookup_found_link, s));
      }
    }
    break;
  case NFSPROC3_CREATE:
    {
      create3args *ca = sbp->template getarg<create3args> ();
      WARN(1) << "NFSPROC3_CREATE, mode " << ca->how.mode << "\n";

      chordID dir = nfs2cid (&(ca->where.dir));
      if (ca->how.mode == GUARDED || ca->how.mode == UNCHECKED) {
	if (check_exclusive_link(dir, ca->where.name))
          sbp->error (NFS3ERR_EXIST);
	else
          lookup(s, dir, ca->where.name,
	         wrap(this, &cfsrw::lookup_found_link_dup, s, 1), true);
      }
      else if (ca->how.mode == EXCLUSIVE)
        sbp->error (NFS3ERR_NOTSUPP);
    }
    break;
  case NFSPROC3_LINK:
    {
      WARN(1) << "NFSPROC3_LINK\n";
      link3args *a = sbp->template getarg<link3args> ();
      chordID dir = nfs2cid (&(a->link.dir));
      if (check_exclusive_link(dir, a->link.name))
        sbp->error (NFS3ERR_EXIST);
      else
	lookup(s, dir, a->link.name,
	       wrap(this, &cfsrw::lookup_found_link_dup, s, 1), true);
    }
    break;
  case NFSPROC3_RMDIR:
    {
      WARN(1) << "NFSPROC3_RMDIR\n";
      diropargs3 *dirop = sbp->template getarg<diropargs3> ();
      chordID dir = nfs2cid (&(dirop->dir));
      lookup(s, dir, dirop->name, 
	     wrap(this, &cfsrw::rmdir_found_link, s, dir));
    }
    break;
  case NFSPROC3_REMOVE:
    {
      WARN(1) << "NFSPROC3_REMOVE\n";
      diropargs3 *dirop = sbp->template getarg<diropargs3> ();
      chordID dir = nfs2cid (&(dirop->dir));
      lookup
	(s, dir, dirop->name,
	 wrap(this, &cfsrw::remove_found_link, s, inode_log_record::T_FILE));
    }
    break;
  case NFSPROC3_WRITE:
    {
      WARN(1) << "NFSPROC3_WRITE\n";
      write3args *a = sbp->template getarg<write3args> ();
      chordID cid = nfs2cid (&(a->file));
      write (s, cid, a->offset, a->count, a->data.base(),
	     wrap(this, &cfsrw::write_done, s, a->count));
    }
    break;
  case NFSPROC3_MKDIR:
    {
      WARN(1) << "NFSPROC3_MKDIR\n";
      mkdir3args *mk = sbp->template getarg<mkdir3args> ();
      chordID dir = nfs2cid (&(mk->where.dir));
      if (check_exclusive_link(dir, mk->where.name))
        sbp->error (NFS3ERR_EXIST);
      else
	lookup (s, dir, mk->where.name,
	        wrap(this, &cfsrw::lookup_found_link_dup, s, 1), true);
    }
    break;
  case NFSPROC3_READDIR:
    {
      readdir3args *readdirop = sbp->template getarg<readdir3args> ();
      chordID dir = nfs2cid (&(readdirop->dir));
      WARN(1) << "NFSPROC3_READDIR " << dir << "\n";
      find_inode (s, dir, wrap(this, &cfsrw::readdir_found_inode, s, dir));
    }
    break;
  case NFSPROC3_SETATTR:
    {
      WARN(1) << "NFSPROC3_SETATTR\n";
      setattr3args *sa = s->sbp->template getarg<setattr3args> ();
      chordID cid = nfs2cid (&(sa->object));
      getattr (s, cid, wrap(this, &cfsrw::setattr_got_attr, s, cid));
    }
    break;
  case NFSPROC3_SYMLINK:
    {
      symlink3args *sa = sbp->template getarg<symlink3args> ();
      chordID dir = nfs2cid (&(sa->where.dir));
      if (check_exclusive_link(dir, sa->where.name))
        sbp->error (NFS3ERR_EXIST);
      else
	lookup(s, dir, sa->where.name,
	       wrap(this, &cfsrw::lookup_found_link_dup, s, 1), true);
    }
    break;
  case NFSPROC3_RENAME:
    {
      rename3args *rn = sbp->template getarg<rename3args> ();
      chordID dir = nfs2cid (&(rn->from.dir));
      lookup(s, dir, rn->from.name,
	     wrap(this, &cfsrw::rename_found_link_old, s, dir));
    }
    break;
  case NFSPROC3_READ:
    {
      WARN(1) << "NFSPROC3_READ\n";
      read3args *ra = sbp->template getarg<read3args> ();
      chordID cid = nfs2cid (&(ra->file));
      uint32 count = ra->count;
      if (count > 8192)
	count = 8192;
      ptr<read_state> rs = New refcounted<read_state>(cid, ra->offset, count);
      read(s, rs, wrap(this, &cfsrw::read_done, s));
    }
    break;
  case NFSPROC3_READLINK:
    {
      nfs_fh3 *nfh = sbp->template getarg<nfs_fh3> ();
      chordID cid = nfs2cid (nfh);
      ptr<read_state> rs = New refcounted<read_state>(cid, 0, 1024);
      read(s, rs, wrap(this, &cfsrw::readlink_got_data, s));
    }
    break;
  default:
    {
      warn << "Procedure " << sbp->proc() << ": not implemented!\n";
      sbp->error (NFS3ERR_NOTSUPP);
      return;
    }
    break;
  }

  build_snapshot (aid, mv->version());
}

void
cfsrw::find_inode (sessionp s, chordID inum, cb_inode cb)
{
  ptr<cattr> ca = New refcounted<cattr>;
  logscan::find_inode
    (s->mv, s->stop, inum, ca,
     wrap(this, &cfsrw::find_inode_check_snapshot, s, inum, cb));
}

void
cfsrw::find_inode_check_snapshot
(sessionp s, chordID i, cb_inode cb, recordp e, ptr<cattr> ca, nfsstat3 status)
{
  if (status == NFS3ERR_ABORTED && s->image)
    snapshot_operation::find_inode (s->image, i, ca, cb);
  else {
    if (status == NFS3ERR_ABORTED)
      status = NFS3ERR_STALE;
    cb(e, ca, status);
  }
}

void
cfsrw::getattr (sessionp s, chordID inum, cb_fa cb)
{
  ptr<cattr> ca = New refcounted<cattr>;
  logscan::find_inode
    (s->mv, s->stop, inum, ca,
     wrap(this, &cfsrw::getattr_check_snapshot, s, inum, cb));
}

void
cfsrw::getattr_check_snapshot (sessionp s, chordID i, cb_fa cb,
                               recordp e, ptr<cattr> ca, nfsstat3 status)
{
  if (status == NFS3ERR_ABORTED && s->image)
    snapshot_operation::find_inode (s->image, i, ca,
	                            wrap (this, &cfsrw::getattr_inode_cb,
				          s, i, cb));
  else {
    if (status == NFS3ERR_ABORTED)
      status = NFS3ERR_STALE;
    getattr_inode_cb (s, i, cb, e, ca, status);
  }
}

void
cfsrw::getattr_inode_cb (sessionp s, chordID i, cb_fa cb,
                         recordp e, ptr<cattr> ca, nfsstat3 status)
{
  if (e && e->template record<inode_log_record>()) {
    ptr<fattr3> fa = get_fattr3(e, ca); 
    cb (fa, status);
  }
  else
    cb ((ptr<fattr3>)0, status);
}

void
cfsrw::find_dir_parent (sessionp s, chordID dir, cb_chordid cb)
{
  logscan::find_dir_parent(s->mv, s->stop, dir,
                           wrap(this, &cfsrw::find_dir_parent_check_snapshot,
			        s, dir, cb));
}

void
cfsrw::find_dir_parent_check_snapshot
(sessionp s, chordID dir, cb_chordid cb, chordID p, nfsstat3 status)
{
  if (status == NFS3ERR_ABORTED && s->image)
    snapshot_operation::find_dir_parent(s->image, dir, cb);
  else {
    if (status == NFS3ERR_ABORTED)
      status = NFS3ERR_NOENT;
    cb(p, status);
  }
}

void
cfsrw::find_dir_parent_done (sessionp s, chordID p, nfsstat3 status)
{
  if (status == NFS3_OK) {
    lookup3res nfsres(NFS3_OK);
    nfs_fh3 nfh;
    cid2nfs (p, &nfh);
    nfsres.resok->object = nfh;
    reply (s->sbp, &nfsres);
  }
  else
    s->sbp->error (NFS3ERR_NOENT);
}

void
cfsrw::lookup
(sessionp s, chordID dir, str name, cb_lookup cb, bool check_intend)
{
  logscan::lookup(s->mv, s->stop, dir, name,
                  wrap(this, &cfsrw::lookup_check_snapshot, s, dir, name, cb),
		  check_intend);
}

void
cfsrw::lookup_check_snapshot
(sessionp s, chordID dir, str name, cb_lookup cb, recordp e, nfsstat3 status)
{
  if (status == NFS3ERR_ABORTED && s->image)
    snapshot_operation::lookup(s->image, dir, name, cb);
  else {
    if (status == NFS3ERR_ABORTED)
      status = NFS3ERR_NOENT;
    cb(e, status);
  }
}

void
cfsrw::read(sessionp s, ptr<read_state> rs, cb_read cb)
{
  logscan::read(s->mv, s->stop, rs,
                wrap(this, &cfsrw::read_check_snapshot, s, cb));
}

void
cfsrw::read_check_snapshot
(sessionp s, cb_read cb, ptr<read_state> rs, nfsstat3 status)
{
  if (status == NFS3ERR_ABORTED && s->image)
    snapshot_operation::read(s->image, rs, cb);
  else {
    if (status == NFS3ERR_ABORTED)
      status = NFS3_OK;
    cb(rs, status);
  }
}

void
cfsrw::readdir(sessionp s, chordID dir, ptr<readdir_state> rs, cb_readdir cb)
{
  logscan::readdir(s->mv, s->stop, dir, rs,
                   wrap(this, &cfsrw::readdir_check_snapshot, s, dir, cb));
}

bool
cfsrw::readdir_check_snapshot
(sessionp s, chordID d, cb_readdir cb,
 ptr<readdir_state> rs, str n, nfsstat3 status)
{
  if (status == NFS3ERR_ABORTED && s->image) {
    snapshot_operation::readdir(s->image, d, rs, cb);
    return true;
  }
  else {
    if (status == NFS3ERR_ABORTED)
      status = NFS3_OK;
    return cb(rs, n, status);
  }
}

// {{{ EXCLUSIVE_CREATE

bool
cfsrw::check_exclusive_link(chordID dir, str name)
{
  strbuf s;
  sid(s, dir);
  s << name;
  str ss = s;
  bigint id = compute_hash(ss.cstr(), ss.len());
  bool *hasit = _locks[id];
  if (hasit && *hasit)
    return true;
  _locks.insert(id, true);
  return false;
}

void
cfsrw::clear_exclusive_link(sessionp s)
{
  nfscall *sbp = s->sbp;
  chordID dir;
  str name;
  if (sbp->proc() == NFSPROC3_MKDIR) {
    mkdir3args *mk = sbp->template getarg<mkdir3args> ();
    dir = nfs2cid (&(mk->where.dir));
    name = mk->where.name;
  }
  else if (sbp->proc() == NFSPROC3_CREATE) {
    create3args *ca = sbp->template getarg<create3args> ();
    dir = nfs2cid (&(ca->where.dir));
    name = ca->where.name;
  }
  else if (sbp->proc() == NFSPROC3_SYMLINK) {
    symlink3args *sa = sbp->template getarg<symlink3args> ();
    dir = nfs2cid (&(sa->where.dir));
    name = sa->where.name;
  }
  else if (sbp->proc() == NFSPROC3_LINK) {
    link3args *la = sbp->template getarg<link3args> ();
    dir = nfs2cid (&(la->link.dir));
    name = la->link.name;
  }
  else
    assert(0);

  strbuf x;
  sid(x, dir);
  x << name;
  str xx = x;
  bigint id = compute_hash(xx.cstr(), xx.len());
  _locks.remove(id);
}

void
cfsrw::cancel_zombie_intend
(sessionp s, int phase, chordID dir, str name, bool ok)
{
  if (!ok) {
    warn << "cancel zombie intend record failed\n";
    s->sbp->error (NFS3ERR_SERVERFAULT);
    return;
  }
  lookup(s, dir, name,
         wrap(this, &cfsrw::lookup_found_link_dup, s, phase), true);
}

void
cfsrw::lookup_found_link_dup(sessionp s, int phase, recordp e, nfsstat3)
{
  chordID dir;
  str name;
  bool unchecked = false;
  if (s->sbp->proc() == NFSPROC3_MKDIR) {
    mkdir3args *mk = s->sbp->template getarg<mkdir3args> ();
    dir = nfs2cid (&(mk->where.dir));
    name = mk->where.name;
  }
  else if (s->sbp->proc() == NFSPROC3_CREATE) {
    create3args *ca = s->sbp->template getarg<create3args> ();
    dir = nfs2cid (&(ca->where.dir));
    name = ca->where.name;
    if (ca->how.mode == UNCHECKED)
      unchecked = true;
  }
  else if (s->sbp->proc() == NFSPROC3_SYMLINK) {
    symlink3args *sa = s->sbp->template getarg<symlink3args> ();
    dir = nfs2cid (&(sa->where.dir));
    name = sa->where.name;
  }
  else if (s->sbp->proc() == NFSPROC3_LINK) {
    link3args *la = s->sbp->template getarg<link3args> ();
    dir = nfs2cid (&(la->link.dir));
    name = la->link.name;
  }
  else
    assert(0);

  ptr<log> l = s->l;
  if (e) {
    intend_log_record *ie = e->template record<intend_log_record>();
    if (ie) {
      if (ie->verf() == _verf) {
	e = 0;
      }
      else {
        // found another intend record, by us, with a different
	// verifier. this can mean two things: someone is running
	// using the same key on another host, which Ivy DOES NOT
	// support, or we crashed from a previous server. in either
	// case, CANCEL this record, and try again. we need to TRY
	// again because someone else may still be trying to create
	// the file, and her INTEND record, hidden behind this bogus
	// INTEND record, has not been canceled or fulfilled yet.
        ptr<intend_log_record> r =
          New refcounted<intend_log_record>(ie->verf(),dir,name,true);
        warn << "cancel after dup intend\n";
        l->append
	  (_version, r,
	   wrap(this, &cfsrw::cancel_zombie_intend, s, phase, dir, name));
	return;
      }
    }
  }

  if (e) {
    if (phase == 2) {
      ptr<intend_log_record> r =
	New refcounted<intend_log_record>(_verf,dir,name,true);
      warn << "cancel after exists\n";
      l->append(_version, r, wrap(this, &cfsrw::noop_bool));
    }
    clear_exclusive_link(s);
    if (unchecked) {
      if (e->template record<intend_log_record>()) {
	// XXX right now we do the wrong thing, which is return file
	// exists. need to tie this with the exponential backoff
	// scheme later
	s->sbp->error (NFS3ERR_EXIST);
	return;
      }
      else {
	// an actual link record exists, truncate it
        assert(e->template record<link_log_record>() ||
	       e->template record<rename_log_record>());
	chordID inum = e->dirop_record()->inum();
        create_found_inode(s, inum);
        return;
      }
    }
    warn << "existing link record found\n";
    s->sbp->error (NFS3ERR_EXIST);
    return;
  }

  if (phase == 2) {
    if (s->sbp->proc() == NFSPROC3_MKDIR)
      do_mkdir(s);
    else if (s->sbp->proc() == NFSPROC3_CREATE)
      check_dir_exists(s);
    else if (s->sbp->proc() == NFSPROC3_SYMLINK)
      do_symlink(s);
    else if (s->sbp->proc() == NFSPROC3_LINK)
      check_dir_exists(s);
    else
      assert(0);
  }
  else {
    ptr<intend_log_record> e =
      New refcounted<intend_log_record>(_verf,dir,name,false);
    l->append
      (_version, e, wrap(this, &cfsrw::exclusive_phase_two, s, dir, name));
  }
}

void
cfsrw::exclusive_phase_two(sessionp s, chordID dir, str name, bool ok)
{
  if (!ok) {
    clear_exclusive_link(s);
    s->sbp->error (NFS3ERR_IO);
    return;
  }
  lookup
    (s, dir, name, wrap(this, &cfsrw::lookup_found_link_dup, s, 2), true);
}

void
cfsrw::setattr_got_attr (sessionp s, chordID obj,
                         ptr<fattr3> fa, nfsstat3 status)
{
  if (fa && status == NFS3_OK) {
    setattr3args *sa = s->sbp->template getarg<setattr3args> ();
    recordp ale = 
      New refcounted<attr_log_record>(obj, &(sa->new_attributes));
    s->l->append(_version, ale, wrap(this, &cfsrw::setattr_reply, s, fa));
  }
  else
    s->sbp->error (NFS3ERR_STALE);
}

void
cfsrw::setattr_reply (sessionp s, ptr<fattr3> fa, bool ok)
{
  if (ok) {
    assert (fa);
    fattr3 newfa = *fa;
    setattr3args *sa = s->sbp->template getarg<setattr3args> ();
    apply_sattr (newfa, sa->new_attributes);
    
    wccstat3 res (NFS3_OK);
    res.wcc->before.set_present(true);
    res.wcc->before.attributes->size = fa->size;
    res.wcc->before.attributes->mtime = fa->mtime;
    res.wcc->before.attributes->ctime = fa->ctime;
    res.wcc->after.set_present(true);
    *(res.wcc->after.attributes) = newfa;
    reply (s->sbp, &res);
  }
  else
    s->sbp->error (NFS3ERR_IO);
}

void
cfsrw::getattr_found_inode
(sessionp s, chordID inum, recordp e, ptr<cattr> ca, nfsstat3 status)
{
  if (e && e->template record<inode_log_record>()) {
    ptr<fattr3> fa = get_fattr3(e, ca);
    getattr3res nfsres (NFS3_OK);
    (*nfsres.attributes) = *fa;
    reply (s->sbp, &nfsres);
  } 
  else
    s->sbp->error (status);
}

void
cfsrw::access_found_inode
(sessionp s, chordID inum, recordp e, ptr<cattr> ca, nfsstat3 status)
{
  inode_log_record *r;
  if(e && (r = e->template record<inode_log_record>())) {
    ptr<fattr3> fa = get_fattr3(e, ca);
    access3args *aa = s->sbp->template getarg<access3args> ();
    uint32 access_req = aa->access;
    access3res nfsres (NFS3_OK);
    nfsres.resok->obj_attributes.set_present(true);
    *(nfsres.resok->obj_attributes.attributes) = *fa;
    nfsres.resok->access = access_check(r, access_req);
    reply (s->sbp, &nfsres);
  } 
  else
    s->sbp->error (status);
}

uint32
cfsrw::access_check(inode_log_record *e, uint32 access_req)
{
  uint32 r = 0;

  switch (e->itype()) { 
  case inode_log_record::T_DIR:
    r = ACCESS3_READ | ACCESS3_LOOKUP | ACCESS3_MODIFY | 
        ACCESS3_EXTEND | ACCESS3_DELETE;
    break;
  case inode_log_record::T_SYMLINK:
    r = ACCESS3_READ;
    break;
  case inode_log_record::T_FILE:
    r = ACCESS3_READ | ACCESS3_MODIFY | ACCESS3_EXTEND | ACCESS3_EXECUTE;
    break;
  }
  return (access_req & r);
}

void
cfsrw::create_found_inode(sessionp s, chordID inum)
{
  ptr<log> l = s->l;
  cattr a;
  a.valid_size = 1;
  a.size = 0;
  recordp ae = 
    New refcounted<attr_log_record>(inum, a);
  l->append
    (_version,
     ae, wrap(this, &cfsrw::link_inode_to_file_done, s, (recordp)0, inum));
}

void
cfsrw::check_dir_exists(sessionp s)
{
  chordID dir;
  str name;
  if (s->sbp->proc() == NFSPROC3_CREATE) {
    create3args *ca = s->sbp->template getarg<create3args> ();
    dir = nfs2cid (&(ca->where.dir));
    name = ca->where.name;
  } else {
    assert(s->sbp->proc() == NFSPROC3_LINK);
    link3args *la = s->sbp->template getarg<link3args> ();
    dir = nfs2cid (&(la->link.dir));
    name = la->link.name;
  }
  find_inode
    (s, dir, wrap(this, &cfsrw::link_inode_found_dir, s, dir, name));
}

void
cfsrw::link_inode_found_dir
(sessionp s, chordID dir, str name, recordp dirnode,
 ptr<cattr> ca, nfsstat3 status)
{
  if (!dirnode || ca->nlink == 0) {
    if (!dirnode)
      warn << "no dirnode\n";
    // clear lock, cancel intend record
    clear_exclusive_link(s);
    ptr<log> l = s->l;
    ptr<intend_log_record> e =
      New refcounted<intend_log_record>(_verf,dir,name,true);
    l->append(_version, e, wrap(this, &cfsrw::noop_bool));
    s->sbp->error (NFS3ERR_NOENT);
    return;
  }
  if (s->sbp->proc() == NFSPROC3_CREATE) {
    ptr<inode_log_record> e = make_inode
      (s, inode_log_record::T_FILE, wrap(this, &cfsrw::noop_recordp));
    link_inode_to_file(s, e, dir, name);
  }
  else if (s->sbp->proc() == NFSPROC3_LINK) {
    link3args *a = s->sbp->template getarg<link3args> ();
    chordID cid = nfs2cid (&(a->file));
    find_inode (s, cid, wrap(this, &cfsrw::link_found_inode, s, dir, name));
  }
}

void
cfsrw::link_found_inode
(sessionp s, chordID dir, str name, recordp e, ptr<cattr> ca, nfsstat3 status)
{
  link_inode_to_file(s, e, dir, name);
}

void
cfsrw::link_inode_to_file(sessionp s, recordp node, chordID dir, str name)
{
  assert(s->sbp->proc() != NFSPROC3_MKDIR);
  ptr<log> l = s->l;
  if (!node || !(node->template record<inode_log_record>())) {
    // clear lock, cancel intend record
    clear_exclusive_link(s);
    ptr<intend_log_record> e =
      New refcounted<intend_log_record>(_verf,dir,name,true);
    warn << "cancel in link_inode_to_file\n";
    l->append(_version, e, wrap(this, &cfsrw::noop_bool));
    s->sbp->error (NFS3ERR_IO);
    return;
  }
  inode_log_record *e = node->template record<inode_log_record>();
  if (e->itype() == inode_log_record::T_DIR) {
    // clear lock, cancel intend record
    clear_exclusive_link(s);
    ptr<intend_log_record> e =
      New refcounted<intend_log_record>(_verf,dir,name,true);
    l->append(_version, e, wrap(this, &cfsrw::noop_bool));
    warn << "cancel in link_inode_to_file 2\n";
    s->sbp->error (NFS3ERR_ISDIR);
    return;
  }
  ptr<link_log_record> le =
    New refcounted<link_log_record> (e->inum(), inode_log_record::T_FILE,
	                             dir, name);
  l->append(_version, le,
            wrap(this, &cfsrw::link_inode_to_file_done, s, node, e->inum()));
}

void
cfsrw::link_inode_to_file_done(sessionp s, recordp e, chordID id, bool ok)
{
  clear_exclusive_link(s);
  if (!ok) {
    // cancel intend record
    chordID dir;
    str name;
    if (s->sbp->proc() == NFSPROC3_CREATE) {
      create3args *ca = s->sbp->template getarg<create3args> ();
      dir = nfs2cid (&(ca->where.dir));
      name = ca->where.name;
    }
    else if (s->sbp->proc() == NFSPROC3_LINK) {
      link3args *la = s->sbp->template getarg<link3args> ();
      dir = nfs2cid (&(la->link.dir));
      name = la->link.name;
    }
    else
      assert(0);
    ptr<log> l = s->l;
    ptr<intend_log_record> e =
      New refcounted<intend_log_record>(_verf,dir,name,true);
    warn << "cancel in link_inode_to_file_done\n";
    l->append(_version, e, wrap(this, &cfsrw::noop_bool));
    s->sbp->error (NFS3ERR_IO);
    return;
  }
  if (s->sbp->proc() == NFSPROC3_CREATE) {
    diropres3 nfsres(NFS3_OK);
    nfs_fh3 nfh;
    cid2nfs (id, &nfh);
    nfsres.resok->obj.set_present(true);
    *nfsres.resok->obj.handle = nfh;
    (nfsres.resok->obj_attributes).set_present(false);
    if (e) {
      create3args *ca = s->sbp->template getarg<create3args> ();
      if (ca->how.mode != EXCLUSIVE) {
        ptr<fattr3> fa = new_attr
	  ((s->sbp->template getarg<create3args>())->how.obj_attributes);
	fix_fattr (e, *fa);
        fa->mtime = e->ts(); // mtime is now.
        (nfsres.resok->obj_attributes).set_present(true);
        *(nfsres.resok->obj_attributes.attributes) = *fa;
      }
    }
    reply (s->sbp, &nfsres);
  }
  else if (s->sbp->proc() == NFSPROC3_LINK) {
    link3res nfsres(NFS3_OK);
    reply (s->sbp, &nfsres);
  }
  else
    assert(0);
}

void
cfsrw::lookup_found_link(sessionp s, recordp r, nfsstat3 status)
{
  if (!r) {
    s->sbp->error (status);
    return;
  }
  lookup3res nfsres(NFS3_OK);
  nfs_fh3 nfh;
  
  chordID fileID = r->dirop_record()->inum();
  cid2nfs (fileID, &nfh);
  nfsres.resok->object = nfh;
  nfsres.resok->obj_attributes.set_present(false);
  reply (s->sbp, &nfsres);
}

ptr<inode_log_record>
cfsrw::make_inode(sessionp s, int type, cb_recordp cb)
{
  sattr3 *sa = 0;
  if(s->sbp->proc() == NFSPROC3_CREATE)
    sa = (s->sbp->template getarg<create3args>())->how.obj_attributes;
  else if(s->sbp->proc() == NFSPROC3_SYMLINK)
    sa =
      &((s->sbp->template getarg<symlink3args>())->symlink.symlink_attributes);
  else if(s->sbp->proc() == NFSPROC3_MKDIR)
    sa = &((s->sbp->template getarg<mkdir3args>())->attributes);
  else
    assert(false);

  ptr<inode_log_record> e = New refcounted<inode_log_record>(type, sa);
  s->l->append(_version, e, wrap(this, &cfsrw::echo, e, cb));
  return e;
}

void
cfsrw::do_mkdir(sessionp s)
{
  mkdir3args *mk = s->sbp->template getarg<mkdir3args> ();
  chordID pid = nfs2cid (&(mk->where.dir));
  
  ptr<inode_log_record> e = make_inode
    (s, inode_log_record::T_DIR, wrap(this, &cfsrw::noop_recordp));
  
  ptr<link_log_record> le = New refcounted<link_log_record>
    (e->inum(), inode_log_record::T_DIR, pid, mk->where.name);
  s->l->append(_version, le,
               wrap(this, &cfsrw::link_inode_to_dir_done, s, e, e->inum()));
}

void
cfsrw::link_inode_to_dir_done(sessionp s, recordp e, chordID id, bool ok)
{
  clear_exclusive_link(s);
  if(!ok) {
    s->sbp->error (NFS3ERR_IO);
    return;
  }
  diropres3 nfsres(NFS3_OK);
  nfs_fh3 nfh;
  cid2nfs (id, &nfh);
  nfsres.resok->obj.set_present(true);
  *nfsres.resok->obj.handle = nfh;

  ptr<fattr3> fa = new_attr();
  fix_fattr (e, *fa);
  fa->nlink = 2;   // XXX
  fa->used = 1024; // XXX
  fa->size = 1024; // XXX
  fa->mtime = e->ts();
  fa->fileid = make_fileid(id);

  nfsres.resok->obj_attributes.set_present(true);
  *(nfsres.resok->obj_attributes.attributes) = *fa;
  nfsres.resok->dir_wcc.before.set_present(false);
  nfsres.resok->dir_wcc.after.set_present(false);
  
  reply (s->sbp, &nfsres);
}

void
cfsrw::rename_found_link_old
(sessionp s, chordID dir, recordp e, nfsstat3 status)
{
  if (!e) {
    s->sbp->error (status);
    return;
  }
  rename3args *rn = s->sbp->template getarg<rename3args> ();
  chordID todir = nfs2cid (&(rn->to.dir));
  lookup(s, todir, rn->to.name,
         wrap(this, &cfsrw::rename_found_link_new, s, e));
}

void
cfsrw::rename_found_link_new
(sessionp s, recordp from_e, recordp to_e, nfsstat3 status)
{
  // XXX what should we do with an intend record?
 
  assert (from_e->dirop_record());
  chordID inum = from_e->dirop_record ()->inum ();
  chordID dir = from_e->dirop_record ()->dir ();
  int itype = from_e->dirop_record ()->itype ();
  str name = from_e->dirop_record ()->name ();
 
  rename3args *rn = s->sbp->template getarg<rename3args> ();
  chordID todir = nfs2cid (&(rn->to.dir));

  if (to_e) {
    // XXX need to check if src and target are compatible. i.e., both
    // are files or both are dirs and target is empty.
    WARN(2) << "rename_found_link_new: found an existing link\n";
    ptr<rename_log_record> re =
      New refcounted<rename_log_record> (inum, itype,
	                                 dir, name, todir, rn->to.name,
					 to_e->dirop_record ()->inum ());
    s->l->append (_version, re,
	          wrap (this, &cfsrw::rename_done, s, from_e, to_e));
  }

  else {
    // this is a simple rename: target does not exist
    WARN(2) << "simple rename, fromdir = " << dir << " to = " << todir << "\n";
    ptr<rename_log_record> re =
      New refcounted<rename_log_record> (inum, itype,
	                                 dir, name, todir, rn->to.name);
    s->l->append (_version, re,
	          wrap (this, &cfsrw::rename_done, s, from_e, to_e));
  }
}

void
cfsrw::rename_done (sessionp s, recordp from, recordp to, bool ok)
{
  if(!ok) {
    s->sbp->error (NFS3ERR_IO);
    return;
  }
  rename3res nfsres (NFS3_OK);

  if (to)
    reply (s->sbp, &nfsres, to->dirop_record ()->inum ());
  else
    reply (s->sbp, &nfsres);
}

void
cfsrw::read_done (sessionp s, ptr<read_state> rs, nfsstat3 status)
{
  if (rs->buf && status == NFS3_OK) {
    read3res res (NFS3_OK);
    res.resok->count = rs->rcount;
    res.resok->data.setsize (rs->rcount);
    memcpy(res.resok->data.base(), rs->buf, rs->rcount);
    res.resok->eof = rs->eof;
    reply (s->sbp, &res);
  }
  else
    s->sbp->error (status);
}

bool
cfsrw::readdir_reply
(sessionp s, chordID d, ptr<readdir_state> rs, str name, nfsstat3 status)
{
  if (status != NFS3_OK) {
    s->sbp->error (status);
    return true;
  }
  if (name == "") {
    rs->nfsres.resok->reply.eof = true;
    reply (s->sbp, &rs->nfsres);
    return true; // stop dir scan
  }
  if (rs->count < (name.len()+30)) {
    reply (s->sbp, &rs->nfsres);
    return true;
  }
  else
    rs->count -= (name.len() + 30);
  return false; // don't stop
}

void
cfsrw::readdir_found_inode
(sessionp s, chordID cid, recordp e, ptr<cattr> ca, nfsstat3 status)
{
  if (e) {
    readdir3args *nfsargs = s->sbp->template getarg<readdir3args> ();
    ptr<readdir_state> rs = New refcounted<readdir_state>(nfsargs);
    if (ca) {
      ptr<fattr3> fa = get_fattr3(e, ca);
      rs->nfsres.resok->dir_attributes.set_present(true);
      *(rs->nfsres.resok->dir_attributes.attributes) = *fa;
    }
    else
      rs->nfsres.resok->dir_attributes.set_present(false);
    readdir(s, cid, rs, wrap(this, &cfsrw::readdir_reply, s, cid));
  }
  else
    s->sbp->error (status);
}

void
cfsrw::rmdir_found_link(sessionp s, chordID dir, recordp e, nfsstat3 status)
{
  if (!e) {
    s->sbp->error (status);
    return;
  }

  assert(e->dirop_record());
  ptr<readdir_state> rs = New refcounted<readdir_state>();
  readdir(s, e->dirop_record()->inum(), rs,
          wrap(this, &cfsrw::do_rmdir, s, e));
}

bool
cfsrw::do_rmdir
(sessionp s, recordp e, ptr<readdir_state> rs, str name, nfsstat3)
{
  if (name != "")
    s->sbp->error (NFS3ERR_NOTEMPTY);
  else
    remove_found_link(s, inode_log_record::T_DIR, e, NFS3_OK);
  return true;
}

void
cfsrw::do_symlink(sessionp s)
{
  ptr<inode_log_record> e = make_inode
    (s, inode_log_record::T_SYMLINK, wrap(this, &cfsrw::noop_recordp));
  
  symlink3args *sa = s->sbp->template getarg<symlink3args> ();
  chordID dir = nfs2cid (&(sa->where.dir));
  str name = sa->where.name;

  ptr<link_log_record> le = New refcounted<link_log_record>
    (e->inum(), inode_log_record::T_SYMLINK, dir, name);
  s->l->append(_version, le, wrap(this, &cfsrw::noop_bool));

  nfspath3 sdata = sa->symlink.symlink_data;
  recordp we = New refcounted<write_log_record>
    (e->inum(), 0, str(sdata.cstr(), sdata.len()));
  s->l->append
    (_version, we, wrap(this, &cfsrw::symlink_done, s, e->inum(), e));
}

void 
cfsrw::symlink_done(sessionp s, chordID id, recordp e, bool ok)
{
  clear_exclusive_link(s);
  if (!ok) {
    s->sbp->error (NFS3ERR_IO);
    return;
  }
  diropres3 nfsres(NFS3_OK);
  nfs_fh3 nfh;
  cid2nfs (id, &nfh);
  nfsres.resok->obj.set_present(true);
  *nfsres.resok->obj.handle = nfh;

  sattr3 *sa = 
    &((s->sbp->template getarg<symlink3args>())->symlink.symlink_attributes);
  ptr<fattr3> fa = new_attr(sa);

  fix_fattr (e, *fa);
  fa->fileid = make_fileid(id);
  nfsres.resok->obj_attributes.set_present(true);
  *(nfsres.resok->obj_attributes.attributes) = *fa;
  nfsres.resok->dir_wcc.before.set_present(false);
  nfsres.resok->dir_wcc.after.set_present(false);
  
  reply (s->sbp, &nfsres);
}

void
cfsrw::readlink_got_data (sessionp s, ptr<read_state> rs, nfsstat3 status)
{
  if (rs->buf && status == NFS3_OK) {
    str xx(rs->buf, rs->rcount);
    WARN(2) << "readlink read " << xx << "\n";
    readlink3res res(NFS3_OK);
    res.resok->data = xx;
    reply (s->sbp, &res);
  }
  else
    s->sbp->error (status);
}

void
cfsrw::remove_reply (sessionp s, recordp e, bool ok)
{
  if (!e || !ok) {
    WARN(1) << "reply oops\n"; 
    s->sbp->error (NFS3ERR_IO);
    return;
  }
  WARN(2) << "reply\n"; 
  wccstat3 res (NFS3_OK);
  reply (s->sbp, &res, e->dirop_record ()->inum ());
}

void
cfsrw::remove_found_link(sessionp s, int itype, recordp e, nfsstat3 status)
{
  if (!e) {
    s->sbp->error (status);
    return;
  }

  assert (e->dirop_record ());
  ptr<unlink_log_record> ue =
    New refcounted<unlink_log_record> (e->dirop_record ()->inum (), itype,
	                               e->dirop_record ()->dir (),
				       e->dirop_record ()->name ());
  s->l->append (_version, ue, wrap (this, &cfsrw::remove_reply, s, e));
}

void
cfsrw::write (sessionp s, chordID inum, uint64 offset, uint32 count,
              const char *data, cb_bool cb)
{
  recordp e = New refcounted<write_log_record>
    (inum, offset, str(data, count));
  s->l->append(_version, e, cb);
}

void
cfsrw::write_done(sessionp s, uint32 count, bool ok)
{
  if (!ok) {
    s->sbp->error (NFS3ERR_IO);
    return;
  }
  write3args *a = s->sbp->template getarg<write3args> ();
  write3res res(NFS3_OK);
  res.resok->count = count;
  res.resok->committed = a->stable;
  writeverf3 verf3;
  mpz_get_rawmag_be (verf3.base(), NFS3_WRITEVERFSIZE, &_verf);
  res.resok->verf = verf3;
  // we cannot return wcc, because a correct wcc would require
  // re-scanning the log
  reply (s->sbp, &res);
}

void
cfsrw::noop_bool(bool ok)
{
}

void
cfsrw::noop_recordp(recordp e)
{
}

void
cfsrw::echo(recordp e, cb_recordp cb, bool ok)
{
  WARN(2) << "echo\n"; 
  if(ok) {
    WARN(2) << "echo calling ok\n"; 
    cb(e);
  } else
    cb((recordp) 0);
}

// try to create a session pointer, which contains a snapshot pointer.
// if a snapshot can't be found, try to load it. if there isn't one to
// load neither, return a session pointer w/o a snapshot.
sessionp
cfsrw::get_session (nfscall *sbp, ptr<materialized_view> mv,
                    bool need_snapshot)
{
  if (mv->version().gt(_version)) _version = mv->version();

  ptr<log> ol = mylog (sbp);
  ptr<log> l = ol;
 
  if (need_snapshot) {
    if (!l || !l->has_snapshot()) {
      WARN(2) << "readonly user, looking for any log with snapshot\n";
      for (unsigned i=0; i<mv->nlogs(); i++) {
        if (mv->head_record(i)->has_snapshot()) {
          l = New refcounted<log> (mv->head_recordp(i));
	  break;
        }
      }
    }

    if (l) {
      ptr<snaprec>* ss = 0;
      if ((ss = _snapshots[l->head()]) && (*ss)->set) {
        sessionp s = New refcounted<session>;
        s->sbp = sbp;
        s->l = ol;
        s->mv = mv;
        s->image = (*ss)->image;
        s->stop = (*ss)->version;
        return s;
      }
      else if (!ss && l->has_snapshot()) {
        // need to load snapshot
        ptr<snaprec> s = New refcounted<snaprec>;
        s->image = 0;
        s->set = false;
        s->building = true;
        _snapshots.insert(l->head(), s);
        warn << "loading snapshot " << l->snapshot() << "\n";
        vNew refcounted<snapshot>
	  (l->snapshot(), wrap(this, &cfsrw::load_snapshot, sbp, mv,
			       l->head()));
        return 0;
      }
    }
  }

  // no snapshot
  sessionp s = New refcounted<session>;
  s->sbp = sbp;
  s->l = ol;
  s->mv = mv;
  s->image = 0;
  return s;
}

void
cfsrw::load_snapshot (nfscall *sbp, ptr<materialized_view> mv,
                      chordID head, chordID loc, ptr<snapshot> s)
{
  if (!s) {
    warn << "cannot load snapshot, running w/o snapshot\n";
    dispatch(sbp, mv, false);
    return;
  }

  ptr<snaprec>* ss = _snapshots[head];
  assert(ss);

  if (!((*ss)->set)) {
    warn << "snapshot loaded " << loc << "\n";
    (*ss)->building = false;
    (*ss)->location = loc;
    (*ss)->version = s->version();
    (*ss)->image = s;
    (*ss)->ts = timenow;
    (*ss)->set = true;
  }

  dispatch(sbp, mv);
}

// build new snapshot and load them into _snapshots hashtable

static unsigned
nrecords (const vervec &from, const vervec &to)
{
  unsigned n = 0;
  for (unsigned i=0; i < from.size () && i < to.size (); i++) {
    if (to.seqnum (i) > from.seqnum (i))
      n += (to.seqnum (i) - from.seqnum (i));
  }
  return n;
}

void
cfsrw::build_snapshot (sfs_aid aid, const vervec &v)
{
  ptr<log> l = mylog (aid);
 
  bool doit = false;
  if (l) {
    ptr<snaprec> *ss = _snapshots[l->head()];
    if (!ss ||
	((*ss)->set && 
	 (nrecords((*ss)->version, v) > ss_thresh) ||
	 (!(*ss)->version.equals(v) && ((timenow - (*ss)->ts) > ss_interval))))
      doit = true;
  }
  if (!doit)
    return;

  ptr<snaprec> *ss = _snapshots[l->head()];
  ptr<snaprec> s = 0;
  if (ss)
    s = *ss;
  if (s && s->set && !s->building && v.gt(s->version)) {
    assert(s->image);
    s->building = true;
    WARN(1) << "build snapshot from previous.\n";
    vNew refcounted<snapshot>
      (_vid, s->location, s->image, wrap (this, &cfsrw::snapshot_created, l));
  }
  else if (!s) {
    s = New refcounted<snaprec>;
    s->image = 0;
    s->set = false;
    s->building = true;
    WARN(1) << "build snapshot from scratch.\n";
    vNew refcounted<snapshot>
      (_vid, (chordID)0, (ptr<snapshot>)0,
       wrap (this, &cfsrw::snapshot_created, l));
    _snapshots.insert(l->head(),s);
  }
}

void
cfsrw::build_snapshot (sfs_aid aid)
{
  _view->materialize(wrap (this, &cfsrw::build_snapshot_with_mv, aid));
}

void
cfsrw::build_snapshot_with_mv (sfs_aid aid, ptr<materialized_view> mv)
{
  build_snapshot (aid, mv->version ());
}

void 
cfsrw::snapshot_created (ptr<log> l, chordID sID, ptr<snapshot> image)
{
  ptr<snaprec> *ss = _snapshots[l->head()];
  assert(ss);
  assert(!l->readonly());
  ptr<snaprec> s = *ss;
  if (image != 0) {
    warn << "snapshot created " << sID << "\n";
    s->location = sID;
    l->add_snapshot(sID);
    s->version = image->version();
    s->image = image;
    s->ts = timenow;
    s->building = false;
    s->set = true;
  }
  else {
    warn << "snapshot creation failed, use old one\n";
    s->building = false;
    s->set = true;
  }

  // in case we fell behind
  build_snapshot (l->writer());
}

ptr<log>
cfsrw::mylog (sfs_aid aid)
{
  ptr<log> *p = _logs[aid];
  if (p && *p != 0) {
    ptr<log> l = *p;
    if (!l->readonly())
      return l;
  }
  return 0;
}

ptr<log>
cfsrw::mylog (nfscall *sbp)
  // contacts cfsagent and returns a log object that cfsagent can sign
{
  sfs_aid aid = sbp->getaid();
  ptr<log> *p = _logs[aid];
  if (p && *p != 0) {
    ptr<log> l = *p;
    if (!l->readonly())
      return l;
    else
      return 0;
  }

  if (!p || (sbp->proc() == NFSPROC3_READDIR)) {
    sfs_aid a = sbp->getaid();
    ptr<aclnt> agentclnt = log::ca(a);
    if (agentclnt) {
      sfs_keylist kl;
      if (clnt_stat err = agentclnt->scall (AGENTCTL_DUMPKEYS, NULL, &kl)) {
        warn << "connecting to agent: " << err << "\n";
        _logs.insert(sbp->getaid(), 0);
        return 0;
      }
      for(sfs_keylistelm *e = kl; e; e = e->next) {
	ptr<sfspub> pk = sfscrypt.alloc (e->desc);
	if (!pk)
	  continue;
        strbuf b;
	pk->export_pubkey (b, false);
	str pk_raw = b;
        chordID v = compute_hash (pk_raw.cstr (), pk_raw.len());
        for (unsigned i=0; i< _view->nlogs(); i++) {
          if (v == _view->head(i) && _view->latest(i) == _view->head(i)) {
            ptr<log> l = New refcounted<log> (_view->head(i));
            l->writer(a);
            _logs.insert(a,l);
            return 0;
          }
        }
      }
    }
    _logs.insert(sbp->getaid(), 0);
  }
  return 0;
}

