
/*
 * Copyright (C) 2002  Benjie Chen (benjie@lcs.mit.edu),
 *                     Massachusetts Institute of Technology
 *
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.
 */

#include "sfscd_prot.h"
#include "cfsa_prot.h"
#include "nfs3close_prot.h"
#include "nfs3_nonnul.h"
#include "rxx.h"

#include "nfstrans.h"

#include "dhash.h"
#include "ranges.h"
#include "log.h"
#include "logiter.h"

#include "nfsxfattr.h"
#include "fcache.h"
#include "cfsrwcd.h"

int cfsrw_verbose = ENV("CFSRW_VERBOSE",0);
int print_stat    = ENV("CFSRW_STAT", 0);
int use_fcache    = ENV("CFSRW_FCACHE", 1);
int use_acache    = ENV("CFSRW_ACACHE", 1);

int ncacheflush = 0;
int nsyncwrite = 0;

void
cfsrwcd::dispatch (ref<nfsserv> ns, nfscall *sbp)
{
  dispatch_nfs (sbp);
}

void
cfsrwcd::dispatch_nfs (nfscall *sbp)
{
  int proc = sbp->proc();

  if (_cfsrw.readonly (sbp)) {
    sbp->error (NFS3ERR_ROFS);
    return;
  }

  if (proc == NFSPROC3_NULL) {
    sbp->reply (NULL);
    return;
  }

  else if (proc == NFSPROC3_FSSTAT) {
    fsstat3res res (NFS3_OK);
    res.resok->obj_attributes.set_present(false);
    res.resok->tbytes = INT64 (0x7fffffffffffffff);
    res.resok->fbytes = INT64 (0x7fffffffffffffff);
    res.resok->abytes = INT64 (0x7fffffffffffffff);
    res.resok->tfiles = INT64 (0x7fffffffffffffff);
    res.resok->ffiles = INT64 (0x7fffffffffffffff);
    res.resok->afiles = INT64 (0x7fffffffffffffff);
    res.resok->invarsec = 0;
    sbp->reply (&res);
    return;
  }

  else if (proc == NFSPROC3_FSINFO) {
    int blocksize = cfsrw::blocksize;
    fsinfo3res res (NFS3_OK);
    res.resok->rtmax = blocksize;
    res.resok->rtpref = blocksize;
    res.resok->rtmult = 512;
    res.resok->wtmax = blocksize;
    res.resok->wtpref = blocksize;
    res.resok->wtmult = 4096;
    res.resok->dtpref = blocksize;
    res.resok->maxfilesize = INT64 (0x7fffffffffffffff);
    res.resok->time_delta.seconds = 0;
    res.resok->time_delta.nseconds = 1;
    res.resok->properties = (FSF3_LINK | FSF3_SYMLINK | FSF3_HOMOGENEOUS);
    sbp->reply (&res);
    return;
  }

  if (proc == NFSPROC3_ACCESS && print_stat) {
    access3args *a = sbp->template getarg<access3args> ();
    chordID dir = nfs2cid (&a->object);
    if (dir == _view->root()) {
      cfsrw_dhashclt::stat();
      warn << nsyncwrite << " synchronous writes\n";
      warn << ncacheflush << " cache flushes\n";
    }
  }

  if (use_fcache && use_acache) {
    if (fcache_dispatch (sbp))
      return;
  }
  else if (sbp->proc () == cl_NFSPROC3_CLOSE) {
    sbp->error (NFS3_OK);
    return;
  }
  else if (sbp->proc () == NFSPROC3_COMMIT) {
    commit3res res (NFS3_OK);
    res.resok->verf = _verf3;
    sbp->reply (&res);
    return;
  }

  _view->materialize(wrap(this, &cfsrwcd::dispatch_with_view, sbp));
}

static inline bool
ctime_recent (timespec ctime)
{
  time_t t = tsnow.tv_sec - ctime.tv_sec;
  if (t > CACHE_RECENT)
    return false;
  return true;
}

void
cfsrwcd::dispatch_with_view(nfscall *sbp, ptr<materialized_view> mv)
  // check if RPC can be answered using the attribute cache
{
  if (!mv || mv->nlogs() == 0) {
    if (!mv)
      warn << "cannot materialize view\n";
    else
      warn << "view has zero logs\n";
    sbp->error(NFS3ERR_SERVERFAULT);
    return;
  }

  if (use_acache) {
    // we update the attribute cache after every modification to the
    // file system. if we are the only writer to the file system, then
    // the attribute cache can be used.
    //
    // flush attr cache if
    //  - current fs version does not equal to attr cache version
    //    AND
    //    - current user cannot write to file system OR
    //    - two versions have different sequence numbers other than the
    //      current user's sequence number OR
    //    - attr cache is more than CACHE_RECENT seconds old
    chordID id;
    bool rw = _cfsrw.uid (sbp, id);
    if (!_ac.current_version ().equals (mv->version ()) &&
        (!rw ||
         !_ac.current_version ().equals_except (mv->version (), id) ||
         !ctime_recent (_ac.cache_time ())))
      _ac.flush_attr ();
  
    // update the attr cache version to the current version
    _ac.current_version (mv->version ());
  
    switch (sbp->proc()) {
    case NFSPROC3_GETATTR:
      {
        nfs_fh3 fh = *sbp->template getarg<nfs_fh3> ();
        const fattr3 *f = _ac.attr_lookup (mv->version (), fh);
        if (f) {
	  WARN(1) << "cached getattr: " << fh << "\n";
	  getattr3res res (NFS3_OK);
	  *res.attributes = *reinterpret_cast<const fattr3 *> (f);
	  sbp->reply (&res);
	  return;
        }
        WARN(1) << "sending getattr: " << fh << "\n";
      }
      break;
    case NFSPROC3_ACCESS:
      {
        access3args *a = sbp->template getarg<access3args> ();
        int32_t perm = _ac.access_lookup
	  (mv->version (), a->object, sbp->getaid (), a->access);
        if (perm > 0) {
          fattr3 fa = *reinterpret_cast<const fattr3 *>
	    (_ac.attr_lookup (mv->version (), a->object));
	  
	  WARN(1) << "cached access: " << a->object << "\n";
          if (fa.type == NF3REG) {
            file_cache *e = file_cache_lookup(a->object);
            if (!e) {
              file_cache_insert (a->object);
              e = file_cache_lookup(a->object);
              assert(e);
              e->open();
            }
            else if (e->is_idle())
              e->open();
	  }
  
          access3res res(NFS3_OK);
          res.resok->obj_attributes.set_present (true);
          *res.resok->obj_attributes.attributes = fa;
          res.resok->access = perm;
          sbp->reply (&res);
          return;
        }
        WARN(1) << "sending access: " << a->object << "\n";
      }
      break;
    default:
      break;
    }
  }
    
  _cfsrw.dispatch (sbp, mv);
}

// extract attributes from replies from cfsrw and put them into the
// attribute cache

void
cfsrwcd::reply (nfscall *sbp, void *res, nfs_fh3 a)
{
  if (use_acache) {
    if (use_fcache)
      fcache_check_reply (sbp, res);

    // flush objects this RPC modified
    if (!nfs_constop (sbp->proc ()))
      flushfattr (sbp->proc (), sbp->getvoidarg (), res);

    // these two RPCs may change the attribute of objects whose file
    // handles are not included in either the arg or the res. the file
    // handle is passed back to us from cfsrw.
    if (sbp->proc () == NFSPROC3_REMOVE ||
	sbp->proc () == NFSPROC3_RENAME) {
      WARN(1) << "removing attr for " << a << "\n";
      _ac.attr_enter(a, 0L, 0L);
    }
    
    // enter new attributes into attribute cache; if this rpc modified
    // the file system, flush objects it changed
    getfattr (sbp->proc (), sbp->getaid (), sbp->getvoidarg (), res);
  }

  sbp->reply (res);
}

void
cfsrwcd::flushfattr (unsigned int proc, void *arg, void *res)
{
  xfhvec xv;
  nfs3_getxfh (&xv, proc, arg, res);
  for (xfh *x = xv.base (); x < xv.lim (); x++) {
    WARN(1) << "removing attr for " << *x->fh << "\n";
    _ac.attr_enter(*x->fh, 0L, 0L);
  }
}

void
cfsrwcd::getfattr (unsigned int proc, sfs_aid aid, void *arg, void *res)
{
  xfattrvec xv;
  nfs3_getxfattr (&xv, proc, arg, res);
  for (xfattr *x = xv.base (); x < xv.lim (); x++) {
    const fattr3 *fa = reinterpret_cast<const fattr3 *> (x->fattr);
    if (fa) {
      WARN(1) << "adding attr for " << *x->fh << "\n";
      _ac.attr_enter (*x->fh, fa, x->wattr);
    }
    if (proc == NFSPROC3_ACCESS) {
      access3res *ares = static_cast<access3res *> (res);
      access3args *a = static_cast<access3args *>(arg);
      if (ares->status)
        _ac.flush_access (a->object, aid);
      else
        _ac.access_enter (a->object, aid, a->access, ares->resok->access);
    }
  }
}

cfsrwcd::cfsrwcd
(sfsprog &p, chordID v, callback<void, cfsrwcd*, nfs_fh3 *>::ref cb)
  : _prog(p), _vid(v),
    _fc(128, wrap (this, &cfsrwcd::file_cache_gc_remove)),
    _view(New refcounted<view>(wrap(this, &cfsrwcd::got_view, cb), v)),
    _cfsrw(_view, v, _verf, wrap(this, &cfsrwcd::reply))
{
  char xxb[20];
  rnd.getbytes (xxb, 20);
  mpz_set_rawmag_be (&_verf, xxb, 20);
  mpz_get_rawmag_be (_verf3.base(), NFS3_WRITEVERFSIZE, &_verf);

  strbuf s;
  sid(s, _vid);
  str ss = armor64A(s);
  _cdir = "/var/tmp/ivy/fcache/" << ss;
  if (mkdir(_cdir.cstr(), 0755) < 0 && errno != EEXIST)
    fatal ("cannot create cache directory %s\n", _cdir.cstr());
  for (int i=0; i<254; i++) {
    str f = _cdir << "/" << i;
    if (mkdir(f.cstr(), 0755) < 0 && errno != EEXIST)
      fatal ("cannot create cache directory %s\n", f.cstr());
  }
}

void
cfsrwcd::got_view
(callback<void, cfsrwcd*, nfs_fh3 *>::ref cb, ptr<view> ok)
{
  if(!ok) {
    warn << "could not fetch view block\n";
    (*cb)(this, NULL);
    return;
  }
  WARN(1) << "got view block. id = " << _vid << "\n";
  
  nfs_fh3 *fh = New nfs_fh3;
  chordID n = _view->root();
  warn << "root inum " << n << "\n";
  strbuf s;
  sid(s, n);
  str ss = armor64A(s);
  warn << "root inum " << ss << "\n";
  cid2nfs (n, fh);
  cb (this, fh);
}
  

void
gotrootfh (ref<nfsserv> ns, sfsserver::fhcb cb, cfsrwcd *server, nfs_fh3 *fh)
{
  if (fh) {
    if (!ns->encodefh (*fh))
      (*cb) (NULL);
    else {
      (*cb) (fh);
      ns->setcb (wrap (server, &cfsrwcd::dispatch, ns));
    }
  }
  else
    (*cb) (NULL);
}

void
cfsrwcd_alloc(sfsprog *prog, ref<nfsserv> ns, int tcpfd,
              sfscd_mountarg *ma, sfsserver::fhcb cb)
{
  if (!ma->cres ||
      (ma->carg.civers == 5 && !sfs_parsepath (ma->carg.ci5->sname))) {
    warn << "request to mount " << ma->carg.ci5->sname << "\n";
  
    rxx path_regexp ("chord:([0-9a-zA-Z+-]*)$");
    if (path_regexp.search (ma->carg.ci5->sname)) {
      chordID view;
      str rootfhstr = path_regexp[1];
      str dearm = dearmor64A(rootfhstr);
      mpz_set_rawmag_be (&view, dearm.cstr(), dearm.len());
      vNew cfsrwcd (*prog, view, wrap (&gotrootfh, ns, cb));
      return;
    }
    
    warn << "malformated filesystem name: "
         << ma->carg.ci5->sname << "\n";
    (*cb)(NULL);
    return;
  }
  fatal << "Ivy must run in a connectinfo version 5 SFS environment\n";
}

int
main (int args, char **argv) 
{
  setprogname (argv[0]);
  
  cfsrw_dhashclt::init();
  sfsconst_init ();
  random_init_file (sfsdir << "/random_seed");

  if (mkdir("/var/tmp/ivy", 0755) < 0 && errno != EEXIST) {
    warn << "cannot create directory /var/tmp/ivy\n";
    return -1;
  }
  if (mkdir("/var/tmp/ivy/scache", 0755) < 0 && errno != EEXIST) {
    warn << "cannot create directory /var/tmp/ivy/scache\n";
    return -1;
  }
  if (mkdir("/var/tmp/ivy/fcache", 0755) < 0 && errno != EEXIST) {
    warn << "cannot create directory /var/tmp/ivy/fcache\n";
    return -1;
  }
  for (unsigned i=0; i<240; i++) {
    strbuf s;
    s << "/var/tmp/ivy/scache/";
    s << i;
    if (mkdir(str(s).cstr(), 0755) < 0 && errno != EEXIST) {
      warn << "cannot create directory " << str(s) << "\n";
      return -1;
    }
  }

  warn << "cfsrwcd version 0.1 running under PID " << getpid () << "\n";
  sfsconst_init ();
  if (ptr<axprt_unix> x = axprt_unix_stdin ())
    vNew sfsprog (x, &cfsrwcd_alloc, false, true);
  else
    fatal ("Couldn't allocate unix transport.\n");
  sigcb(SIGTERM, wrap(exit, 0)); // for profiling.
  amain();
}

