/*
 *
 * Copyright (C) 2002 Benjie Chen (benjie@lcs.mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "record.h"
#include "scan.h"
#include "ranges.h"
#include "fcache.h"
#include "cfsrwcd.h"

struct write_obj {
  static const unsigned int PARALLEL_WRITES = 32;
  typedef callback<void,bool>::ref cb_t;

  cb_t cb;
  cfsrw& c;
  sessionp sn;
  file_cache *fe;
  nfs_fh3 fh;
  chordID cid;
  uint64 start;
  uint64 size;
  uint64 written;
  unsigned int outstanding_writes;
  bool callback;
  sfs_aid aid;
  unsigned int blocks;

  void
  write_reply (bool success) {
    blocks++;
    // only make sense to build once per PARALLEL_WRITES blocks
    // anyways, because that's when our log head is likely to be
    // flushed to DHash. we try to build at a slightly faster pace
    // since building snapshot is slower than speed at which we can
    // write to DHash.
    if (blocks == PARALLEL_WRITES/2) {
      c.build_snapshot (aid);
      blocks = 0;
    }
    outstanding_writes--;
    if (!callback && success) {
      do_write();
      ok ();
    }
    else
      fail();
  }

  void cfsrw_write (uint64 off, uint32 cnt) {
    ptr<aiobuf> buf = cfsrwcd::a->bufalloc (cnt);
    if (!buf) {
      cfsrwcd::a->bufwait
	(wrap (this, &write_obj::cfsrw_write, off, cnt));
      return;
    }
    outstanding_writes++;
    fe->afh->read (off, buf,
	           wrap (this, &write_obj::cfsrw_write_read, off, cnt));
  }

  void cfsrw_write_read (uint64 off, uint32 cnt, ptr<aiobuf> buf,
                         ssize_t sz, int err)
  {
    if (err || (unsigned)sz != cnt) {
      outstanding_writes--;
      if (err)
        warn << "flush_cache: read failed: " << err << "\n";
      else 
	warn << "flush_cache: short read: got "
	     << sz << " wanted " << cnt << "\n";
      fail ();
      return;
    }
    c.write (sn, cid, off, cnt, buf->base(),
	     wrap (this, &write_obj::write_reply));
  }

  bool do_write () {
    if (written < size) {
      unsigned s = size-written;
      s = s > cfsrw::blocksize ? cfsrw::blocksize : s;
      cfsrw_write (start+written, s);
      written += s;
    }
    if (written < size)
      return false;
    return true;
  }

  static void file_closed (int) { }

  void fail () {
    if (!callback) {
      fe->afh->close (wrap (&write_obj::file_closed));
      fe->afh = 0;
      callback = true;
      cb (false);
    }
    if (outstanding_writes == 0)
      delete this;
  }

  void ok () {
    if (outstanding_writes == 0) {
      if (!callback) {
	c.build_snapshot (aid);
        fe->afh->close (wrap (&write_obj::file_closed));
        fe->afh = 0;
        callback = true;
	cb (true);
      }
      delete this;
    }
  }

  write_obj (str fn, file_cache *fe, uint64 start, uint64 size, cfsrw& c,
             ptr<materialized_view> mv, write_obj::cb_t cb, nfscall *sbp)
    : cb(cb), c(c), fe(fe), fh(fe->fh), start(start), size(size), written(0),
      outstanding_writes(0), callback(false)
  {
    if (!mv || mv->nlogs () == 0) {
      fail ();
      return;
    }
    assert (fe->afh);

    cid = nfs2cid (&fh);
    sn = c.get_session (sbp, mv, false);
    if (!sn) {
      fail ();
      return;
    }

    blocks = 0;
    aid = sbp->getaid ();

    for (unsigned i=0; i<PARALLEL_WRITES; i++)
      do_write ();
    ok ();
  }

  ~write_obj() {}
};

void
cfsrwcd::flush_cache (file_cache *e, nfscall *sbp)
{
  assert (e->is_dirty());
  e->flush ();
  str fn = fh2fn (e->fh);
  _view->materialize(wrap (this, &cfsrwcd::flush_cache_with_mv, e, sbp));
}

void
cfsrwcd::flush_cache_with_mv (file_cache *e, nfscall *sbp,
                              ptr<materialized_view> mv)
{
  uint64 start = e->mstart;
  uint64 size = e->mend - e->mstart;

  if (sbp->proc () == NFSPROC3_COMMIT) {
    commit3args *a = sbp->template getarg<commit3args> ();
    start = a->offset;
    size = a->count;
  }

  if (start <= e->mstart)
    e->mstart = start + size;
  if (start + size >= e->mend)
    e->mend = start;
  if (e->mstart >= e->mend) {
    e->mstart = 0;
    e->mend = 0;
  }

  vNew write_obj (fh2fn (e->fh), e, start, size, _cfsrw, mv,
                  wrap(this, &cfsrwcd::flush_done, e, sbp), sbp);
}

