
#include "async.h"  // timenow and tsnow
#include "cache.h"

//
// caching policy
//
// three separate LRU caches. one for log records. one for sfsro
// blocks. and one for all the other type of dhash blocks.
//
// we cache both immutable log records and public key log head
// records. we only cache public key log head records that we insert.
// this means we will NOT cache any log head records that we did not
// insert. furthermore, cached log head records expire after 60
// seconds. cached immutable log records are returned to the caller
// when retrieved. copies of cached log head records are returned to
// the caller when retrieved.
//
// we provide two methods for retrieving sfsro blocks, one that makes
// a copy of the block before returning (retrieve_sfsro), and one that
// does not (retrieve_sfsro_readonly).
//
// the has_sfsro(chordID) method returns true if the sfsro block with
// the given chordID exists in the cache (i.e.  it has been recently
// inserted or retrieved). calling this method also moves the sfsro
// block to the tail of the LRU list (i.e. it is touched). the
// has_block(chordID) method has similar semantics.
//

#include "debug.h"

#define PK_CACHETIME 60  // cache pk blocks for 60 seconds

static int precsz = ENV("CFSRW_PRINT_RECORD_SIZE", 0);
static int rcsize = ENV("CFSRW_RCACHE", 512);
static int scsize = ENV("CFSRW_SCACHE", 256);
static int bcsize = ENV("CFSRW_BCACHE", 16);

static int pk_skip_lookup = ENV("CFSRW_PK_SKIPLOOKUP", 1);
static int recinsert_skip_lookup = ENV("CFSRW_RECINSERT_SKIPLOOKUP", 0);

lrucache<chordID, cached_block, hashID> cfsrw_dhashclt::record_cache(rcsize);
lrucache<chordID, cached_block, hashID> cfsrw_dhashclt::sfsro_cache(scsize);
lrucache<chordID, cached_block, hashID> cfsrw_dhashclt::block_cache(bcsize);

dhashclient *cfsrw_dhashclt::dhc = 0;

int cfsrw_dhashclt::record_fetch = 0;
int cfsrw_dhashclt::record_insert = 0;
int cfsrw_dhashclt::cached_record_fetch = 0;
int cfsrw_dhashclt::sfsro_fetch = 0;
int cfsrw_dhashclt::sfsro_insert = 0;
int cfsrw_dhashclt::cached_sfsro_fetch = 0;
int cfsrw_dhashclt::ch_insert = 0;
int cfsrw_dhashclt::pk_insert = 0;
int cfsrw_dhashclt::pk_fetch = 0;

ptr<sfsro_data>
sfsro_decode (const char *d, size_t len)
{
  ptr<sfsro_data> data = New refcounted<sfsro_data>;
  xdrmem x (d, len, XDR_DECODE);
  if (xdr_sfsro_data (x.xdrp (), data))
    return data;
  else
    return NULL;
}

ptr<sfsro_data>
sfsro_decode (ptr<dhash_block> blk)
{
  return sfsro_decode(blk->data, blk->len);
}

void
cfsrw_dhashclt::init()
{
  char *sn = getenv("LSD_SOCKET");
  if (sn == 0) {
    warnx << "LSD_SOCKET is not set\n";
    exit(1);
  }
  dhc = New dhashclient(getenv("LSD_SOCKET"));
}

void
cfsrw_dhashclt::retrieve_cb
(cbretrieve_t cb, bool insert, chordID id, ptr<dhash_block> b)
{
  if (b && insert && bcsize) {
    cached_block blk;
    blk.type = cached_block::DHASH;
    blk.block = b;
    blk.record = 0;
    blk.sfsro = 0;
    block_cache.insert (id, blk);
    ptr<dhash_block> n = New refcounted<dhash_block>(b->data, b->len);
    cb (n);
  }
  else
    cb (b);
}

void
cfsrw_dhashclt::retrieve_ch(chordID id, cbretrieve_t cb)
{
  bool insert = true;
  cached_block *b = block_cache[id];
  if (b && bcsize) {
    if (b->type == cached_block::DHASH) {
      timecb (tsnow,
	      wrap (&cfsrw_dhashclt::retrieve_cb, cb, false, id, b->block));
      return;
    }
    else
      insert = false;
  }
  dhc->retrieve
    (id, wrap (&cfsrw_dhashclt::retrieve_cb, cb, insert, id));
}

void
cfsrw_dhashclt::retrieve_record_cb
(int tries, cb_recordp cb, bool insert, chordID id, ptr<dhash_block> b)
{
  if (b) {
    ptr<log_record> e = log_record::parse (str(b->data, b->len));
    if (insert && rcsize) {
      cached_block blk;
      blk.type = cached_block::LOG;
      blk.block = 0;
      blk.record = e;
      blk.sfsro = 0;
      record_cache.insert (id, blk);
      cb (e);
    }
    else
      cb (e);
  }
  else
    cb ((ptr<log_record>)0);
}

void
cfsrw_dhashclt::return_record (cb_recordp cb, ptr<log_record> e)
{
  cb (e);
}

void
cfsrw_dhashclt::retrieve_record (chordID id, cb_recordp cb)
{
  record_fetch++;
  bool insert = true;
  cached_block *b = record_cache[id];
  if (b && rcsize) {
    if (b->type == cached_block::LOG) {
      cached_record_fetch++;
      timecb (tsnow, wrap (&cfsrw_dhashclt::return_record, cb, b->record));
      return;
    }
    else
      insert = false;
  }
  dhc->retrieve
    (id, wrap (&cfsrw_dhashclt::retrieve_record_cb, 0, cb, insert, id));
}

void
cfsrw_dhashclt::retrieve_sfsro_cb
(cbsfsro cb, bool insert, chordID id, ptr<dhash_block> b)
{
  if (b) {
    ptr<sfsro_data> s = sfsro_decode (b);
    if (insert && scsize) {
      cached_block blk;
      blk.type = cached_block::SFSRO;
      blk.block = 0;
      blk.record = 0;
      blk.sfsro = s;
      sfsro_cache.insert (id, blk);
      copy_sfsro(cb, s);
    }
    else
      cb (s);
  }
  else
    cb ((ptr<sfsro_data>)0);
}

bool
cfsrw_dhashclt::has_sfsro (chordID n)
{
  cached_block *b = sfsro_cache[n];
  if (b && b->type == cached_block::SFSRO) {
    b->touches++;
    if (b->touches < 1024)
      return true;
    else
      sfsro_cache.remove (n);
  }
  return false;
}

void
cfsrw_dhashclt::copy_sfsro (cbsfsro cb, ptr<sfsro_data> s)
{
  ptr<sfsro_data> n = New refcounted<sfsro_data> (*s);
  cb (n);
}

void
cfsrw_dhashclt::retrieve_sfsro_readonly_cb
(cbsfsro cb, bool insert, chordID id, ptr<dhash_block> b)
{
  if (b) {
    ptr<sfsro_data> s = sfsro_decode (b);
    if (insert && scsize) {
      cached_block blk;
      blk.type = cached_block::SFSRO;
      blk.block = 0;
      blk.record = 0;
      blk.sfsro = s;
      sfsro_cache.insert (id, blk);
      cb (s);
    }
    else
      cb (s);
  }
  else
    cb ((ptr<sfsro_data>)0);
}

void
cfsrw_dhashclt::return_sfsro (cbsfsro cb, ptr<sfsro_data> s)
{
  cb (s);
}

void
cfsrw_dhashclt::retrieve_sfsro(chordID id, cbsfsro cb)
{
  sfsro_fetch++;
  bool insert = true;
  cached_block *b = sfsro_cache[id];
  if (b && scsize) {
    if (b->type == cached_block::SFSRO) {
      cached_sfsro_fetch++;
      timecb (tsnow, wrap (&cfsrw_dhashclt::copy_sfsro, cb, b->sfsro));
      return;
    }
    else
      insert = false;
  }
  dhc->retrieve
    (id, wrap (&cfsrw_dhashclt::retrieve_sfsro_cb, cb, insert, id));
}

void
cfsrw_dhashclt::retrieve_sfsro_readonly (chordID id, cbsfsro cb)
{
  sfsro_fetch++;
  bool insert = true;
  cached_block *b = sfsro_cache[id];
  if (b && scsize) {
    if (b->type == cached_block::SFSRO) {
      cached_sfsro_fetch++;
      timecb (tsnow, wrap (&cfsrw_dhashclt::return_sfsro, cb, b->sfsro));
      return;
    }
    else
      insert = false;
  }
  dhc->retrieve
    (id, wrap (&cfsrw_dhashclt::retrieve_sfsro_readonly_cb, cb, insert, id));
}

void
cfsrw_dhashclt::insert_ok_cb (cbinsert_t cb, chordID n)
{
  cb (false, n);
}

bool
cfsrw_dhashclt::has_block (chordID n)
{
  cached_block *b = block_cache[n];
  if (b && b->type == cached_block::SFSRO) {
    b->touches++;
    if (b->touches < 1024)
      return true;
    else
      block_cache.remove (n);
  }
  return false;
}

void
cfsrw_dhashclt::insert_ch_cb
(ptr<dhash_block> b, cbinsert_t cb, bool failed, ptr<insert_info> i)
{
  assert (i);
  block_cache.remove (i->key);
  if (!failed && bcsize) {
    cached_block blk;
    blk.type = cached_block::DHASH;
    blk.block = b;
    blk.record = 0;
    blk.sfsro = 0;
    block_cache.insert (i->key, blk);
  }
  cb (failed, i->key);
}

void
cfsrw_dhashclt::insert_ch
(chordID hash, const char *data, size_t len, cbinsert_t cb)
{
  if (!has_block (hash)) {
    ch_insert++;
    ptr<dhash_block> b = New refcounted<dhash_block> (data, len);
    dhc->insert (hash, data, len, wrap (&cfsrw_dhashclt::insert_ch_cb, b, cb));
  }
  else
    timecb (tsnow, wrap (&cfsrw_dhashclt::insert_ok_cb, cb, hash));
}

void
cfsrw_dhashclt::insert_sfsro
(const char *data, size_t len, cbinsert_t cb, ptr<sfsro_data> b)
{
  if (b == 0)
    b = sfsro_decode (data, len);
  ch_insert++;
  dhc->insert (data, len, wrap (&cfsrw_dhashclt::insert_sfsro_cb, b, cb));
}

void
cfsrw_dhashclt::insert_sfsro
(chordID hash, const char *data, size_t len, cbinsert_t cb, ptr<sfsro_data> b)
{
  if (!has_sfsro (hash)) {
    if (b == 0)
      b = sfsro_decode (data, len);
    sfsro_insert++;
    dhc->insert (hash, data, len,
	         wrap (&cfsrw_dhashclt::insert_sfsro_cb, b, cb));
  }
  else
    timecb (tsnow, wrap (&cfsrw_dhashclt::insert_ok_cb, cb, hash));
}

void
cfsrw_dhashclt::insert_sfsro_cb
(ptr<sfsro_data> b, cbinsert_t cb, bool failed, ptr<insert_info> i)
{
  assert (i);
  sfsro_cache.remove (i->key);
  if (!failed && scsize) {
    cached_block blk;
    blk.type = cached_block::SFSRO;
    blk.block = 0;
    blk.record = 0;
    blk.sfsro = b;
    sfsro_cache.insert (i->key, blk);
  }
  cb (failed, i->key);
}

void
cfsrw_dhashclt::insert_record
(const char *data, size_t len, cbinsert_t cb, recordp e)
{
  if (e == 0) 
    e = log_record::parse (str(data, len));
  assert(!e->template record<head_log_record>());
  record_insert++;
 
  if (precsz)
    warn << "insert record " << len << " bytes\n";
  chordID id = compute_hash (data, len);
  dhc->insert (id, data, len,
               wrap (&cfsrw_dhashclt::insert_record_cb, 0, e, cb),
               recinsert_skip_lookup > 0);
}

void
cfsrw_dhashclt::insert_record
(chordID id, const char *data, size_t len, cbinsert_t cb, recordp e)
{
  if (e == 0)
    e = log_record::parse (str(data, len));
  assert(!e->template record<head_log_record>());
  record_insert++;

  if (precsz)
    warn << "insert record " << len << " bytes\n";
  dhc->insert (id, data, len,
               wrap (&cfsrw_dhashclt::insert_record_cb, 0, e, cb),
	       recinsert_skip_lookup > 0);
}

void
cfsrw_dhashclt::insert_record_cb
(int tries, recordp e, cbinsert_t cb, bool failed, ptr<insert_info> i)
{
  assert (i);
  record_cache.remove (i->key);
  if (!failed && rcsize) {
    cached_block blk;
    blk.type = cached_block::LOG;
    blk.block = 0;
    blk.record = e;
    blk.sfsro = 0;
    record_cache.insert (i->key, blk);
  }
  cb (failed, i->key);
}

void
cfsrw_dhashclt::insert_ch(const char *data, size_t len, cbinsert_t cb)
{
  ch_insert++;
  ptr<dhash_block> b = New refcounted<dhash_block> (data, len);
  dhc->insert (data, len, wrap (&cfsrw_dhashclt::insert_ch_cb, b, cb));
}

void
cfsrw_dhashclt::insert_pk
(ptr<sfspriv> k, const char *data, size_t len, cbinsert_t cb, recordp e)
{
  if (e == 0)
    e = log_record::parse (str(data, len));
  assert(e->template record<head_log_record>());
  pk_insert++;
  dhc->insert (k, data, len, wrap (&cfsrw_dhashclt::insert_pk_cb, e, cb));
}

void
cfsrw_dhashclt::insert_pk (chordID hash, sfs_pubkey2 k, sfs_sig2 s,
                           const char *data, size_t len,
			   cbinsert_t cb, recordp e)
{
  if (e == 0)
    e = log_record::parse (str(data, len));
  assert(e->template record<head_log_record>());
  pk_insert++;
  dhc->insert (hash, k, s, data, len,
               wrap (&cfsrw_dhashclt::insert_pk_fast_cb, hash, k, s, e, cb),
	       pk_skip_lookup > 0);
}

void
cfsrw_dhashclt::insert_pk_fast_cb (chordID hash, sfs_pubkey2 k, sfs_sig2 s,
				   recordp e, cbinsert_t cb,
				   bool failed, ptr<insert_info> i)
{
  assert (i);
  if (!failed)
    insert_pk_cb (e, cb, failed, i);
  else {
    warn << "fast insert of public key failed, try regular insert\n";
    str data = e->serialize ();
    dhc->insert (hash, k, s, data.cstr (), data.len (),
	         wrap (&cfsrw_dhashclt::insert_pk_cb, e, cb));
  }
}


void
cfsrw_dhashclt::insert_pk (sfs_pubkey2 k, sfs_sig2 s,
                           const char *data, size_t len,
			   cbinsert_t cb, recordp e)
{
  if (e == 0)
    e = log_record::parse (str(data, len));
  assert(e->template record<head_log_record>());
  pk_insert++;
  dhc->insert (k, s, data, len, wrap (&cfsrw_dhashclt::insert_pk_cb, e, cb));
}

void
cfsrw_dhashclt::insert_pk_cb (recordp e, cbinsert_t cb,
                              bool failed, ptr<insert_info> i)
{
  assert (i);
  record_cache.remove (i->key);
  if (!failed && rcsize) {
    cached_block blk;
    blk.type = cached_block::PKEY;
    blk.block = 0;
    blk.record = e;
    blk.sfsro = 0;
    record_cache.insert (i->key, blk);
  }
  cb (failed, i->key);
}

void
cfsrw_dhashclt::copy_record (cb_recordp cb, ptr<log_record> e)
{
  ptr<log_record> n = e->dup ();
  if (!n)
    warn << "dup failed in cache\n";
  cb (n);
}

void
cfsrw_dhashclt::retrieve_pk(chordID id, cb_recordp cb)
{
  cached_block *b = record_cache[id];
  if (rcsize && b && b->type == cached_block::PKEY) {
    // cache only contains public key records we inserted. i.e. we
    // never cache other people's public key records.
    if (timenow - b->record->ts ().seconds < PK_CACHETIME) {
      timecb (tsnow, wrap (&cfsrw_dhashclt::copy_record, cb, b->record));
      return;
    }
    else
      record_cache.remove (id);
  }
  pk_fetch++;
  
  dhc->retrieve (id, wrap (&cfsrw_dhashclt::retrieve_pk_fast_cb, id, cb),
		 pk_skip_lookup > 0);
}

void 
cfsrw_dhashclt::retrieve_pk_fast_cb (chordID id, cb_recordp cb,
                                     ptr<dhash_block> b)
{
  if (b) {
    ptr<log_record> e = log_record::parse (str(b->data, b->len));
    if (!e)
      warn << "parse failed in cache\n";
    cb (e);
  }
  else {
    warn << "fast retrieve on keyhash failed, try regular retrieve\n";
    dhc->retrieve (id, wrap (&cfsrw_dhashclt::retrieve_pk_cb, cb));
  }
}

void 
cfsrw_dhashclt::retrieve_pk_cb (cb_recordp cb, ptr<dhash_block> b)
{
  if (b) {
    ptr<log_record> e = log_record::parse (str(b->data, b->len));
    if (!e)
      warn << "parse failed in cache\n";
    cb (e);
  }
  else
    cb ((ptr<log_record>)0);
}

bool
cfsrw_dhashclt::has_record (chordID n)
{
  cached_block *b = record_cache[n];
  if (b && b->type == cached_block::LOG) {
    b->touches++;
    if (b->touches < 1024)
      return true;
    else
      record_cache.remove (n);
  }
  return false;
}

void
cfsrw_dhashclt::stat()
{
  warn << "record_insert " << record_insert << "\n";
  warn << "record_fetch " << record_fetch << "\n";
  warn << "cached_record_fetch " << cached_record_fetch << "\n";
  warn << "sfsro_insert " << sfsro_insert << "\n";
  warn << "sfsro_fetch " << sfsro_fetch << "\n";
  warn << "cached_sfsro_fetch " << cached_sfsro_fetch << "\n";
  warn << "pk_fetch " << pk_fetch << "\n";
  warn << "pk_insert " << pk_insert << "\n";
  warn << "ch_insert " << ch_insert << "\n";
}

