
#include "log.h"
#include "cache.h"
#include "serialize.h"
#include "cfsa_prot.h"
#include "rxx.h"
#include <time.h>

qhash<sfs_aid, ref<aclnt> > log::cahash;

log::log(chordID head, cb_recordp cb)
  : _head(head)
{
  _has_snapshot = false;
  _snapshot_new = false;
  _readonly = false;
  _head_record = 0;
  _stores = _failed_stores = 0;
  _refreshed = false;
  _loading = false;
  locate(_head, true, wrap(mkref(this), &log::load_head_cb, cb));
}

log::log(chordID head)
  : _head(head)
{
  _has_snapshot = false;
  _snapshot_new = false;
  _readonly = false;
  _head_record = 0;
  _stores = _failed_stores = 0;
  _refreshed = false;
  _loading = false;
  locate(_head, true, wrap(mkref(this), &log::load_head));
}

log::log(recordp e)
  : _head(e->headID())
{
  assert(e->template record<head_log_record>());
  _has_snapshot = false;
  _snapshot_new = false;
  _readonly = false;
  _head_record = 0;
  _stores = _failed_stores = 0;
  _refreshed = false;
  _loading = false;
  load_head(e);
}

void
log::load_head(recordp e)
{
  assert(!_refreshed);
  if (e) {
    if (_appends_e.size() > 0)
      _refreshed = true;
    _head_record = e;
    head_log_record* h = e->template record<head_log_record>();
    assert(h);
    _root = h->root();
    if (h->has_snapshot() && (!_has_snapshot || !_snapshot_new)) {
      _has_snapshot = true;
      _snapshot_new = false;
      _snapshot = h->snapshot();
    }
  }
}

void
log::load_head_cb(cb_recordp cb, recordp e)
{
  load_head(e);
  cb(e);
}

void
log::append(const vervec& v, recordp e, cb_bool cb)
{
  assert(!_readonly);

  e->set_version(v);
  _appends_e.push_back(e);
  _appends_cb.push_back(cb);

  if (!_loading) {
    _loading = true;
    // need to load a fresh head record
    locate(_head, true, wrap(mkref(this), &log::load_head_cb,
	                     wrap(mkref(this), &log::do_append)));
  }
  else if (_refreshed)
    do_append(_head_record);
}

void
log::do_append(recordp e)
{
  bool abort = false;
  if (e == 0)
    abort = true;

  for(unsigned i=0; i<_appends_e.size(); i++) {
    recordp e = _appends_e[i];

    if (!abort) {
      if (e->version().conflicts(_head_record->version())) {
        // oops, another writer, in a more up-to-date view perhaps, has
        // written a head of log. that means this view is out of date
        // and should be made read-only!
        abort = true;
        _readonly = true;
        _head_record = 0;
      }
      else {
        vervec v = e->version();
        _head_record->set_version(v);
      }
    }
    _store_cbs.push_back(_appends_cb[i]);

    if (!abort) {
      e->set_prevID(_head_record->prevID());
      e->set_headID(_head);
      e->set_seqnum(_head_record->seqnum());
      _stores++;
  
      str data = e->serialize();
      chordID id = compute_hash(data.cstr(), data.len());
      _head_record->set_prevID(id);
      _head_record->set_seqnum(_head_record->seqnum()+1);
      store_record
        (id, data, 
         wrap(mkref(this), &log::store_head_after_append), e);
    }
  }
  _appends_e.clear();
  _appends_cb.clear();

  if (abort) {
    _stores++;
    store_head_after_append(false);
  }
}


void
log::store_record_cb (cb_bool cb, bool failed, chordID cid)
{
  if (failed)
    cb(false);
  else
    cb(true);
}

void
log::store_head (recordp e, const str data, cb_bool cb,
                 sfs_pubkey2 k, sfs_sig2 s, bool ok)
{
  if (ok)
    cfsrw_dhashclt::insert_pk
      (_head, k, s, data.cstr(), data.len(),
       wrap(mkref(this), &log::store_record_cb, cb), e);
  else {
    store_record_cb(cb, true, 0);
    cahash.remove(_writer);
  }
}

void
log::store_record(chordID key, const str &data, cb_bool cb, recordp e)
{
  assert(e);
  if (key == _head) {
    sign (data, wrap(mkref(this), &log::store_head, e, data, cb));
    return;
  }
  cfsrw_dhashclt::insert_record
    (key, data.cstr(), data.len(),
     wrap (mkref(this), &log::store_record_cb, cb), e);
}

void
log::head_stored(ptr<vec<cb_bool> >cbs, ptr<head_log_record> h, bool ok)
{
  if (ok && cbs) {
    for (unsigned i=0; i<cbs->size(); i++)
      (*cbs)[i](true);
  }
  else {
    if (cbs) {
      for (unsigned i=0; i<cbs->size(); i++)
        (*cbs)[i](false);
    }
  }
  if (_appends_e.size() > 0) {
    assert(_loading = true);
    _refreshed = true;
    do_append(_head_record);
  }
  else
    _loading = false;
}

void 
log::add_snapshot(chordID s)
{
  _snapshot_new = true;
  _has_snapshot = true;
  _snapshot = s;
}

void
log::store_head_after_append(bool ok)
{
  _stores--;
  if (!ok)
    _failed_stores++;
  if (_stores == 0) {
    if (_failed_stores == 0) {
      _head_record->set_time();
      head_log_record* h = _head_record->template record<head_log_record>();
      if (_has_snapshot) {
	h->snapshot(_snapshot);
	_snapshot_new = false;
      }
      str data = _head_record->serialize();
      ptr<vec<cb_bool> > cbs = New refcounted<vec<cb_bool> >();
      for (unsigned i=0; i<_store_cbs.size(); i++)
	cbs->push_back(_store_cbs[i]);
      _store_cbs.clear();
      ptr<head_log_record> e = New refcounted<head_log_record>(*h);
      store_record
	(_head, data, wrap(mkref(this), &log::head_stored, cbs, e), e);
      // leave loading to true and refreshed to false: if a request
      // comes in while DHash is storing the head block, the request
      // is saved, and we won't try to load a new head block until
      // after the old head block is saved.
      _loading = true;
      _refreshed = false;
    }
    else {
      for (unsigned i=0; i<_store_cbs.size(); i++)
	_store_cbs[i](false);
      _store_cbs.clear();
      // start over, a new request would first grab the version of the
      // head block in DHash
      _loading = false;
      _refreshed = false;
      _head_record = 0;
    }
    _failed_stores = 0;
  }
}

void
log::locate(chordID id, bool keyhashed, cb_recordp cb)
{
  if (!keyhashed)
    cfsrw_dhashclt::retrieve_record(id, cb);
  else
    cfsrw_dhashclt::retrieve_pk(id, cb);
}

void
log::prev_record(recordp e, cb_recordp cb)
{
  locate(e->prevID(), false, cb);
}

void
log::dhash2log(cb_recordp cb, chordID id, ptr<dhash_block> blk)
{
  if (!blk) {
    cb((recordp) 0);
    return;
  }
  recordp e = log_record::parse(str(blk->data, blk->len));
  if (!e) {
    cb((recordp) 0);
    return;
  }
  cb(e);
}

void
log::sign_block_cb(callback<void, sfs_pubkey2, sfs_sig2, bool>::ref cb, 
                   cfsa_signhead_res *res, clnt_stat err)
{
  if (!err && !res->status)
    cb (res->resok->k, res->resok->s, true);
  else {
    sfs_pubkey2 k;
    sfs_sig2 s;
    cb (k, s, false);
  }
  delete res;
}

void
log::sign(const str &data, callback<void, sfs_pubkey2, sfs_sig2, bool>::ref cb)
{
  ptr<aclnt> agentclnt = ca(_writer);
  if (agentclnt) {
    cfsa_signhead_res *res = New cfsa_signhead_res;
    cfsa_signhead_arg a;
    a.keyhash = head();
    a.serialized_head.setsize(data.len());
    memmove(a.serialized_head.base(), data.cstr(), data.len());
    agentclnt->call(CFS_AGENTCTL_SIGNHEAD, &a, res,
	            wrap(mkref(this), &log::sign_block_cb, cb, res));
  }
  else {
    sfs_pubkey2 k;
    sfs_sig2 s;
    cb (k, s, false);
  }
}

ptr<aclnt>
log::ca(sfs_aid aid)
{
  ptr<aclnt> c = cahash[aid];
  if (c)
    return c;

  // get a connection to the user's agent; there may be a better way
  // to do this that does not involve a fork...

  int fds[2];
  int cfd = -1;

  if (socketpair (AF_UNIX, SOCK_STREAM, 0, fds) < 0)
    fatal("socketpair: %m\n");

  pid_t pid = fork();
  if (pid) { // parent
    close (fds[1]);
    char c;
    readfd(fds[0], &c, 1, &cfd);
    int status;
    wait(&status);
  }
  else {
    close (fds[0]);
    setuid(aid);

    cfd = suidgetfd_required("agent");
    if (cfd < 0) {
      warn << "suidgetfd failed\n";
      exit(1);
    }
    if (writefd(fds[1], "", 1, cfd) < 0) {
      warn << "writefd failed\n";
      exit(1);
    }
    exit(0);
  }

  if (cfd < 0) {
    warn << "no file descriptor to sfscd\n";
    return 0;
  }

  ref<axprt_unix> sfscdxprt = axprt_unix::alloc(cfd);
  ref<aclnt> sfscdclnt = aclnt::alloc(sfscdxprt, agent_prog_1);

  int32_t res;
  int afd;
  if (clnt_stat err = sfscdclnt->scall (AGENT_GETAGENT, NULL, &res)) {
    warn << "connect to cdc failed " << err << "\n";
    return 0;
  }
  if ((afd = sfscdxprt->recvfd ()) < 0) {
    warn << "connecting to agent via sfscd: no file descriptor\n";
    return 0;
  }

  ref<aclnt> agentclnt = 
    aclnt::alloc (axprt_stream::alloc (afd), cfs_agentctl_prog_1);
  cahash.insert(aid, agentclnt);

  return agentclnt;
}
