
#include "serialize.h"
#include "cache.h"
#include "log.h"
#include "logiter.h"
#include "view.h"
#include "cfsa_prot.h"
#include "sfscd_prot.h"

bool
logiter::check_and_load_record(int i, recordp e)
{
  recordp olde = _entries[i];
  _entries[i] = e;
  _valid[i] = true;

  if (e == 0 || (olde && e->seqnum() >= olde->seqnum())) {
    _valid[i] = false;
    if (!e) {
      iterinfo info(*this, 0, LI_ERR_NREC, !_comparable, true, true);
      _cb(&info);
      if (info.finished == true)
	return false;
      else
	return true;
    }
    else {
      iterinfo info(*this, e, LI_ERR_SEQN, !_comparable, true, true);
      _cb(&info);
      return false;
    }
  }
  if (olde && !olde->template record<head_log_record>() &&
      (olde->version()).conflicts(e->version())) {
    iterinfo info(*this, e, LI_ERR_SCONF, !_comparable, true, true);
    _cb(&info);
    return false;
  }
  if (!_stop.empty() && e->seqnum() <= _stop[i])
    _valid[i] = false;
  return true;
}

void
logiter::check_ordering()
{
  // there must be an ordering between every two records, otherwise
  // there is either concurrent operations, or conflicts due to stale
  // log head being used
  for (unsigned i=0; i<_entries.size(); i++) {
    if (_valid[i] && !_entries[i]->template record<end_log_record>()) {
      for (unsigned j=0; j<_entries.size(); j++) {
	if (_valid[j] && !_entries[j]->template record<end_log_record>() &&
	    i != j) {
          if ((_entries[i]->version()).equals(_entries[j]->version()) ||
	      (_entries[i]->version()).conflicts(_entries[j]->version())) {
	    _comparable = false;
	    return;
	  }
	}
      }
    }
  }
  _comparable = true;
}

void
logiter::got_materialized_view(ptr<materialized_view> ok)
{
  if (!ok) {
    iterinfo i(*this, 0, LI_ERR_NVIEW, false, true, true);
    _cb(&i);
    return;
  }
  _mview = ok;
  for (unsigned i=0; i < _mview->nlogs(); i++) {
    recordp r = _mview->latest(i);
    _valid.push_back(false);
    _entries.push_back((recordp)0);
    _id.push_back(r->selfID());
    if (!_stopv.empty()) {
      bool found = false;
      for (unsigned j=0; j < _stopv.size() && !found; j++) {
        if (_stopv.head(j) == r->headID()) {
	  _stop.push_back(_stopv.seqnum(j));
	  found = true;
        }
      }
      if (!found)
	_stop.push_back(0);
    }
    if (!check_and_load_record(i, r))
      return;
  }
  iter();
}

void
logiter::got_view(ptr<view> ok)
{
  if (!ok) {
    iterinfo i(*this, 0, LI_ERR_NVIEW, false, true, true);
    _cb(&i);
    return;
  }
  ok->materialize (wrap (mkref(this), &logiter::got_materialized_view));
}

logiter::logiter(cbliter cb, ptr<materialized_view> mv)
  : _cb(cb)
{
  _comparable = true;
  timecb (tsnow, wrap (mkref(this), &logiter::got_materialized_view, mv));
}

logiter::logiter(cbliter cb, ptr<materialized_view> mv, vervec stop)
  : _cb(cb)
{
  _stopv = stop;
  _comparable = true;
  timecb (tsnow, wrap (mkref(this), &logiter::got_materialized_view, mv));
}

logiter::logiter(cbliter cb, chordID v, vervec stop)
  : _cb(cb)
{
  _stopv = stop;
  _comparable = true;
  vNew refcounted<view> (wrap (mkref(this), &logiter::got_view), v);
}

logiter::logiter(cbliter cb, chordID v)
  : _cb(cb)
{
  _comparable = true;
  vNew refcounted<view> (wrap (mkref(this), &logiter::got_view), v);
}

static void
prefetch_record (int n, ptr<log_record> e)
{
  if (n < 4 && e && e->prevID () != 0
      && !cfsrw_dhashclt::has_record (e->prevID ()))
    log::prev_record (e, wrap(prefetch_record, n+1));
}

void
logiter::iter_cb(int vi, ptr<log_record> e)
{
  if (!check_and_load_record(vi, e))
    return;
  // prefetch log record into record cache
  if (e->prevID () != 0 && !cfsrw_dhashclt::has_record (e->prevID ()))
    log::prev_record (e, wrap(prefetch_record, 1));
  iter();
}

void
logiter::iter()
{
  bool got_hi = false;
  unsigned int hi;

  check_ordering();
  for (unsigned int i = 0; i < _entries.size(); i++) {
    if (_valid[i] &&
	(got_hi == false ||
	 _entries[i]->version().ge(_entries[hi]->version()))) {
      hi = i;
      got_hi = true;
    }
  }

  if (got_hi == false) {
    iterinfo i(*this, 0, LI_OK, false, true, true);
    _cb(&i);
    return;
  }
 
  iterinfo i(*this, _entries[hi], LI_OK, !_comparable, false, false);
  _cb(&i);
  if (i.finished)
    return;

  if (i.eol || _entries[hi]->template record<end_log_record>()) {
    _valid[hi] = false;
    iter();
  }
  else {
    _id[hi] = _entries[hi]->prevID();
    log::prev_record (_entries[hi], wrap(mkref(this), &logiter::iter_cb, hi));
  }
}

