
#include "async.h"  // tsnow
#include "record.h"
#include "serialize.h"
#include <time.h>

log_record::log_record (int type)
{
  _type = type;
  set_time();
}

ptr<log_record>
log_record::parse(str ins)
{
  ptr<log_record> e = 0;
  xin *in = New xin(ins);
  
  chordID prev;
  chordID head;
  int seqnum;
  int type;
  int sec, nsec;
  int vsize;
  vervec version;
 
  if (!in->rid(prev))
    goto bad;
  if (!in->rid(head))
    goto bad;
  if (!in->rint(seqnum))
    goto bad;
  if (!in->rint(type))
    goto bad;
  if (!in->rint(sec))
    goto bad;
  if (!in->rint(nsec))
    goto bad;
  if (!in->rint(vsize))
    goto bad;
  for (int i=0; i<vsize; i++) {
    chordID h;
    chordID l;
    int s;
    if (!in->rid(h))
      goto bad;
    if (!in->rid(l))
      goto bad;
    if (!in->rint(s))
      goto bad;
    version.add(h, l, s);
  }

  if(type == T_END)
    e = New refcounted<end_log_record>(in);
  else if(type == T_INODE)
    e = New refcounted<inode_log_record>(in);
  else if(type == T_HEAD)
    e = New refcounted<head_log_record>(in);
  else if(type == T_LINK)
    e = New refcounted<link_log_record>(in);
  else if (type == T_UNLINK)
    e = New refcounted<unlink_log_record>(in);
  else if (type == T_RENAME)
    e = New refcounted<rename_log_record>(in);
  else if(type == T_WRITE)
    e = New refcounted<write_log_record>(in);
  else if(type == T_INTEND)
    e = New refcounted<intend_log_record>(in);
  else if(type == T_ATTR)
    e = New refcounted<attr_log_record>(in);
  else
    warn << "log_record::parse: unknown record type " << type << "\n";
  delete in;

  if (e) {
    e->_prevID = prev;
    e->_headID = head;
    e->_type = type;
    e->_seqnum = (uint)seqnum;
    e->_ts.seconds = sec;
    e->_ts.nseconds = nsec;
    e->_version = version;
    return e;
  }
  else
    return 0;
    
bad:  
  delete in;
  return 0;
}

ptr<log_record>
log_record::dup()
{
  ptr<log_record> r = 0;
  if(type() == T_END)
    r = New refcounted<end_log_record>
      (*this->template record<end_log_record>());
  else if(type() == T_INODE)
    r = New refcounted<inode_log_record>
      (*this->template record<inode_log_record>());
  else if(type() == T_HEAD)
    r = New refcounted<head_log_record>
      (*this->template record<head_log_record>());
  else if(type() == T_LINK)
    r = New refcounted<link_log_record>
      (*this->template record<link_log_record>());
  else if (type() == T_UNLINK)
    r = New refcounted<unlink_log_record>
      (*this->template record<unlink_log_record>());
  else if (type() == T_RENAME)
    r = New refcounted<rename_log_record>
      (*this->template record<rename_log_record>());
  else if(type() == T_WRITE)
    r = New refcounted<write_log_record>
      (*this->template record<write_log_record>());
  else if(type() == T_INTEND)
    r = New refcounted<intend_log_record>
      (*this->template record<intend_log_record>());
  else if(type() == T_ATTR)
    r = New refcounted<attr_log_record>
      (*this->template record<attr_log_record>());
  else
    assert(0);
  return r;
}

void
log_record::serialize1(strbuf &sb) const
{
  sid(sb, _prevID);
  sid(sb, _headID);
  sint(sb, _seqnum);
  sint(sb, _type);
  sint(sb, _ts.seconds);
  sint(sb, _ts.nseconds);
  _version.serialize(sb);
}

cattr::cattr()
{
  valid_size = valid_mode = valid_mtime
    = valid_uid = valid_gid = valid_rdev = 0;
  size = mode = uid = gid = 0;
  mtime.seconds = mtime.nseconds = 0;
  ctime.seconds = ctime.nseconds = 0;
  nlink = 0;
  truncated = false;
}

void
cattr::set_mtime_now()
{
  mtime.seconds = tsnow.tv_sec;
  mtime.nseconds = tsnow.tv_nsec;
  valid_mtime = 1;
}

cattr::cattr(sattr3 *s)
{
  valid_size = valid_mode = valid_mtime 
    = valid_uid = valid_gid = valid_rdev = 0;
  size = mode = uid = gid = mtime.seconds = mtime.nseconds = 0;

  if(s->size.set) {
    valid_size = 1;
    size = (uint64) *s->size.val;
  }
  if(s->mode.set) {
    valid_mode = 1;
    mode = (uint32) *s->mode.val;
  }
  if(s->mtime.set == SET_TO_CLIENT_TIME) {
    valid_mtime = 1;
    mtime = (nfstime3) *s->mtime.time;
  }
  if(s->mtime.set == SET_TO_SERVER_TIME) {
    set_mtime_now();
  }
  if(s->uid.set) {
    valid_uid = 1;
    uid = (uint32) *s->uid.val;
  }
  if(s->gid.set) {
    valid_gid = 1;
    gid = (uint32) *s->gid.val;
  }
}

void
cattr::serialize(strbuf &sb) const
{
  sint(sb, valid_size);
  squad(sb, size);
  sint(sb, valid_mode);
  sint(sb, mode);
  sint(sb, valid_mtime);
  sint(sb, mtime.seconds);
  sint(sb, mtime.nseconds);
  sint(sb, valid_uid);
  sint(sb, uid);
  sint(sb, valid_gid);
  sint(sb, gid);
  sint(sb, valid_rdev);
  sint(sb, rdev.major);
  sint(sb, rdev.minor);
}

void
cattr::xinxin(xin *in)
{
  in->rint(valid_size);
  in->rquad(size);
  in->rint(valid_mode);
  in->rint((int) mode);
  in->rint(valid_mtime);
  in->rint((int) mtime.seconds);
  in->rint((int) mtime.nseconds);
  in->rint(valid_uid);
  in->rint((int) uid);
  in->rint(valid_gid);
  in->rint((int) gid);
  in->rint(valid_rdev);
  in->rint((int) rdev.major);
  in->rint((int) rdev.minor);
}

void
cattr::ca2fa(fattr3& fa) const
{
  fa.size = fa.mode = fa.uid = fa.gid = 0;

  if(valid_size)
    fa.size = size;
  if(valid_mode)
    fa.mode = mode;
  if(valid_mtime)
    fa.mtime = mtime;
  if(valid_uid)
    fa.uid = uid;
  if(valid_gid)
    fa.gid = gid;
  if(valid_rdev)
    fa.rdev = rdev;
  fa.nlink = nlink;
  fa.ctime = ctime;
}

void
cattr::update(cattr *ca)
{
  // set size if not set yet, or when enlarging file
  if(ca->valid_size && !valid_size) {
    size = ca->size;
    valid_size = 1;
  }
  if(ca->valid_mode && !valid_mode) {
    mode = ca->mode;
    valid_mode = 1;
  }
  if(ca->valid_mtime && !valid_mtime) {
    mtime = ca->mtime;
    valid_mtime = 1;
  }
  if(ca->valid_uid && !valid_uid) {
    uid = ca->uid;
    valid_uid = 1;
  }
  if(ca->valid_gid && !valid_gid) {
    gid = ca->gid;
    valid_gid = 1;
  }
  if(ca->valid_rdev && !valid_rdev) {
    rdev = ca->rdev;
    valid_rdev = 1;
  }
}

attr_log_record::attr_log_record(chordID inum, sattr3 *attr)
  : log_record(T_ATTR), _attr(attr)
{
  _inum = inum;
}

attr_log_record::attr_log_record(chordID inum, cattr attr)
  : log_record(T_ATTR), _attr(attr)
{
  _inum = inum;
}

attr_log_record::attr_log_record(xin *in)
{
  in->rid(_inum);
  _attr.xinxin(in);
}

str
attr_log_record::serialize() const
{
  strbuf sb;

  serialize1(sb);
  sid(sb, _inum);
  _attr.serialize(sb);
  return sb;
}

inode_log_record::inode_log_record(int itype, chordID inum)
  : log_record(T_INODE), _attr()
{
  _itype = itype;
  _inum = inum;
}

inode_log_record::inode_log_record(int itype, chordID inum, sattr3* sa)
  : log_record(T_INODE), _attr(sa)
{
  _itype = itype;
  _inum = inum;
}

inode_log_record::inode_log_record(int itype, chordID inum, cattr sa)
  : log_record(T_INODE), _attr(sa)
{
  _itype = itype;
  _inum = inum;
}

inode_log_record::inode_log_record(int itype)
  : log_record(T_INODE), _attr()
{
  _itype = itype;
  // random new i-number.
  char xxb[20];
  rnd.getbytes(xxb, 20);
  mpz_set_rawmag_be(&_inum, xxb, 20);
}


inode_log_record::inode_log_record(int itype, sattr3 *sa)
  : log_record(T_INODE), _attr(sa)
{
  _itype = itype;
  // random new i-number.
  char xxb[20];
  rnd.getbytes(xxb, 20);
  mpz_set_rawmag_be(&_inum, xxb, 20);
}


inode_log_record::inode_log_record(int itype, cattr sa)
  : log_record(T_INODE), _attr(sa)
{
  _itype = itype;
  // random new i-number.
  char xxb[20];
  rnd.getbytes(xxb, 20);
  mpz_set_rawmag_be(&_inum, xxb, 20);
}

inode_log_record::inode_log_record(xin *in)
{
  in->rint(_itype);
  in->rid(_inum);
  _attr.xinxin(in);
}

str
inode_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sint(sb, _itype);
  sid(sb, _inum);
  _attr.serialize(sb);
  return sb;
}

end_log_record::end_log_record()
  : log_record(T_END)
{
  set_prevID(0);
}

end_log_record::end_log_record(xin *)
{
}

str
end_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  return sb;
}

head_log_record::head_log_record(chordID root)
  : log_record(T_HEAD), _has_snapshot(false), _snapshot (0)
{
  _root = root;
}

head_log_record::head_log_record()
  : log_record(), _has_snapshot(false), _snapshot (0)
{
}

head_log_record::head_log_record(xin *in)
{
  int a;
  in->rid(_root);
  in->rint(a);
  if (a) _has_snapshot = true;
  else _has_snapshot = false;
  in->rid(_snapshot);
}

str
head_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sid(sb, _root);
  int a = _has_snapshot ? 1 : 0;
  sint(sb, a);
  sid(sb, _snapshot);
  return sb;
}

dirop_log_record::dirop_log_record
  (int type, chordID inum, int itype, chordID dir, str name)
  : log_record(type), _dir(dir), _inum(inum), _name(name), _itype(itype)
{
}

link_log_record::link_log_record
  (chordID inum, int itype, chordID dir, str name)
  : dirop_log_record (T_LINK, inum, itype, dir, name)
{
}

link_log_record::link_log_record(xin *in)
{
  in->rid(_inum);
  in->rint(_itype);
  in->rid(_dir);
  in->rstr(_name);
}

str
link_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sid(sb, _inum);
  sint(sb, _itype);
  sid(sb, _dir);
  sstr(sb, _name);
  return sb;
}

unlink_log_record::unlink_log_record
  (chordID inum, int itype, chordID dir, str name)
  : dirop_log_record (T_UNLINK, inum, itype, dir, name)
{
}

unlink_log_record::unlink_log_record(xin *in)
{
  in->rid(_inum);
  in->rint(_itype);
  in->rid(_dir);
  in->rstr(_name);
}

str
unlink_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sid(sb, _inum);
  sint(sb, _itype);
  sid(sb, _dir);
  sstr(sb, _name);
  return sb;
}

rename_log_record::rename_log_record
  (chordID inum, int itype, chordID olddir, str oldname,
   chordID dir, str name, chordID delinum)
  : dirop_log_record(T_RENAME, inum, itype, dir, name),
    _olddir(olddir), _oldname(oldname),
    _deleted(true), _delinum(delinum)
{
}

rename_log_record::rename_log_record
  (chordID inum, int itype, chordID olddir, str oldname,
   chordID dir, str name)
  : dirop_log_record(T_RENAME, inum, itype, dir, name),
    _olddir(olddir), _oldname(oldname), _deleted(false)
{
}

rename_log_record::rename_log_record(xin *in)
{
  in->rid(_inum);
  in->rint(_itype);
  in->rid(_olddir);
  in->rstr(_oldname);
  in->rid(_dir);
  in->rstr(_name);
  int n;
  in->rint(n);
  if (n)
    _deleted = true;
  else
    _deleted = false;
  in->rid(_delinum);
}

str
rename_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sid(sb, _inum);
  sint(sb, _itype);
  sid(sb, _olddir);
  sstr(sb, _oldname);
  sid(sb, _dir);
  sstr(sb, _name);
  if (_deleted)
    sint(sb, 1);
  else
    sint(sb, 0);
  sid(sb, _delinum);
  return sb;
}

intend_log_record::intend_log_record(bigint v, chordID d, str name, bool c)
  : log_record(T_INTEND), _verf(v), _dir(d), _name(name)
{
  if (c)
    _canceled = 1;
  else
    _canceled = 0;
}

intend_log_record::intend_log_record(xin *in)
{
  in->rid(_verf);
  in->rid(_dir);
  in->rstr(_name);
  in->rint(_canceled);
}

str
intend_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sid(sb, _verf);
  sid(sb, _dir);
  sstr(sb, _name);
  sint(sb, _canceled);
  return sb;
}

write_log_record::write_log_record(chordID inum, uint64 offset, str data)
  : log_record(T_WRITE), _data(data)
{
  _inum = inum;
  _offset = offset;
}

write_log_record::write_log_record(xin *in)
{
  in->rid(_inum);
  in->rquad(_offset);
  in->rstr(_data);
}

str
write_log_record::serialize() const
{
  strbuf sb;
  serialize1(sb);
  sid(sb, _inum);
  squad(sb, _offset);
  sstr(sb, _data);
  return sb;
}

static inline char *
inames(unsigned i)
{
  switch (i) {
    case inode_log_record::T_FILE:
      return "FILE";
    case inode_log_record::T_DIR:
      return "DIR";
    case inode_log_record::T_SYMLINK:
      return "SYMLINK";
    default:
      break;
  }
  assert(0);
  return 0;
}

static inline char *
tnames(unsigned i)
{
  switch (i) {
    case log_record::T_END:
      return "END";
    case log_record::T_INODE:
      return "INODE";
    case log_record::T_HEAD:
      return "HEAD";
    case log_record::T_LINK:
      return "LINK";
    case log_record::T_UNLINK:
      return "UNLINK";
    case log_record::T_WRITE:
      return "WRITE";
    case log_record::T_ATTR:
      return "ATTR";
    case log_record::T_RENAME:
      return "RENAME";
    case log_record::T_INTEND:
      return "INTEND";
    default:
      break;
  }
  assert(0);
  return 0;
}

const strbuf &
strbuf_cat(const strbuf &b, log_record *e)
{
  b << "type =" << tnames(e->type());
  b << " , ts = " << e->ts().seconds;

  if (rename_log_record *ee = e->template record<rename_log_record>()) {
    b << " " << ee->oldname() << " -> " << ee->name() 
      << " (" << ee->olddir() << " -> " << ee->dir() << ")";
  }
  else if (unlink_log_record *ee = e->template record<unlink_log_record>())
    b << " " << ee->name();
  else if (link_log_record *ee = e->template record<link_log_record>())
    b << " " << ee->name();
  else if (inode_log_record *ee = e->template record<inode_log_record>()) {
    b << " " << inames(ee->itype());
    b << " inum: " << ee->inum();
    b << ", mtime = " 
      << (ee->attr()->valid_mtime ? ee->attr()->mtime.seconds : 0);
  } 
  else if (attr_log_record *ee = e->template record<attr_log_record>()) {
    b << " inum = " << ee->inum()
      << ", mode = " << (ee->attr()->valid_mode ? ee->attr()->mode : 0)
      << ", size = " << (ee->attr()->valid_size ? ee->attr()->size : 0)
      << ", uid = "  << (ee->attr()->valid_uid  ? ee->attr()->uid  : 0)
      << ", gid = "  << (ee->attr()->valid_gid ? ee->attr()->gid : 0)
      << ", mtime = " 
      << (ee->attr()->valid_mtime ? ee->attr()->mtime.seconds : 0);
  }
  return b;
}

