#ifndef CFSRW_RECORD_H
#define CFSRW_RECORD_H

#include "async.h"     // tsnow
#include "nfs3_prot.h" // fattr3
#include "vec.h"
#include "vervec.h"
#include "serialize.h"

template<>
struct hashfn<chordID> {
  hash_t operator() (const chordID &id) const { return id.getui (); }
};

class xin;
class end_log_record;
class head_log_record;
class attr_log_record;
class write_log_record;
class inode_log_record;
class intend_log_record;
class dirop_log_record;
class link_log_record;
class unlink_log_record;
class rename_log_record;

class log_record {
private:
  chordID _prevID;
  chordID _headID;
  int _type;
  uint _seqnum;
  nfstime3 _ts;
  vervec _version;

protected:
  void serialize1(strbuf &sb) const;
  virtual const type_info& cxxtype() const { return typeid(*this); }

public:
  log_record (int type = 0);
  virtual ~log_record() {}
  ptr<log_record> dup();
  
  int type() const { return _type; }
  virtual str serialize() const = 0;

  nfstime3 ts() const { return _ts; }
  void set_time() {
    _ts.seconds = tsnow.tv_sec;
    _ts.nseconds = tsnow.tv_nsec;
  }

  chordID selfID() const {
    str data = serialize();
    return compute_hash(data.cstr(), data.len());
  }
  chordID prevID() const { return _prevID; }
  chordID headID() const { return _headID; }
  uint seqnum() const { return _seqnum; }
  vervec version() const { return _version; }
  
  void set_prevID(chordID id) { _prevID = id; }
  void set_headID(chordID id) { _headID = id; }
  void set_seqnum(uint n)   { _seqnum = n; }
  void set_version(const vervec &v) { _version = v; }

  enum record_type {
    T_END = 1,
    T_INODE,
    T_HEAD,
    T_LINK,
    T_UNLINK,
    T_WRITE,
    T_ATTR,
    T_RENAME,
    T_INTEND
  };

  template<class T> T* record() {
    if (typeid(T) != cxxtype() &&
	// not sure why cxxtype() can return a refcounted type, since
	// supposedly the this pointer is just a plain log_record ptr.
	typeid(refcounted<T>) != cxxtype())
      return 0;
    return static_cast<T*>(this);
  }

  dirop_log_record * dirop_record() {
    if (_type == T_LINK || _type == T_UNLINK || _type == T_RENAME)
      return ((dirop_log_record*) this);
    return 0;
  }

  static ptr<log_record> parse(str);
};

class cattr {
public:
  int valid_size;  uint64 size;
  int valid_mode;  uint32 mode;
  int valid_mtime; nfstime3 mtime;
  int valid_uid;   uint32 uid;
  int valid_gid;   uint32 gid;
  int valid_rdev;  specdata3 rdev;
  nfstime3 ctime;
  int32 nlink;
  bool truncated;

  cattr();
  cattr(sattr3*);
  void serialize(strbuf&) const;
  void xinxin(xin*);
  void ca2fa(fattr3&) const;
  void update(cattr*);
  void set_mtime_now();
};

class attr_log_record : public log_record {
private:
  chordID _inum;
  cattr _attr;

public:
  attr_log_record(const attr_log_record& r) { *this = r; }
  attr_log_record(chordID, sattr3*);
  attr_log_record(chordID, cattr);
  attr_log_record(xin *);
  chordID inum() const { return _inum; }
  cattr* attr() { return &_attr; }
  str serialize() const;
};

class inode_log_record : public log_record {
private:
  int _itype;
  chordID _inum;
  cattr _attr;

public:
  inode_log_record(const inode_log_record& r) { *this = r; }
  inode_log_record(int);
  inode_log_record(int, sattr3*);
  inode_log_record(int, cattr);
  inode_log_record(int, chordID);
  inode_log_record(int, chordID, sattr3*);
  inode_log_record(int, chordID, cattr);
  inode_log_record(xin *);

  int itype() const { return _itype; }
  chordID inum() const { return _inum; }
  cattr* attr() { return &_attr; }

  str serialize() const;

  enum inode_type {
    T_FILE = 1,
    T_DIR,
    T_SYMLINK
  };
};

class end_log_record : public log_record {
public:
  end_log_record();
  end_log_record(const end_log_record& r) { *this = r; }
  end_log_record(xin *);
  str serialize() const;
};

class head_log_record : public log_record {
private:
  bool _has_snapshot;
  chordID _root;
  chordID _snapshot;
  
public:
  head_log_record();
  head_log_record(const head_log_record& h) { *this = h; }
  head_log_record(chordID root);
  head_log_record(xin *);

  chordID root() const { return _root; }
  bool has_snapshot() const { return _has_snapshot; }
  chordID snapshot () const { return _snapshot; }
  void snapshot (chordID s) { _snapshot = s; _has_snapshot = true; }
  str serialize() const;
};

class intend_log_record : public log_record {
private:
  bigint _verf;
  chordID _dir;
  str _name;
  int _canceled;

public:
  intend_log_record(const intend_log_record& r) { *this = r; }
  intend_log_record(bigint, chordID, str, bool);
  intend_log_record(xin *);
  str serialize() const;
  str name() const { return _name; }
  bigint verf() const { return _verf; }
  chordID dir() const { return _dir; }
  bool canceled() const { return (_canceled == 1); }
};

class write_log_record : public log_record {
private:
  chordID _inum;
  uint64 _offset;
  str _data;

public:
  write_log_record(const write_log_record& r) { *this = r; }
  write_log_record(chordID, uint64, str);
  write_log_record(xin *);

  chordID inum() const { return _inum; }
  uint64 offset() const { return _offset; }
  uint32 count() const { return _data.len(); }
  str data() const { return _data; }

  str serialize() const;
};

class dirop_log_record : public log_record {
protected:
  chordID _dir;
  chordID _inum;
  str _name;
  int _itype;     // inode type
  
  dirop_log_record (int type, chordID inum, int itype, 
                    chordID dir, str name);
  dirop_log_record () {}  // used by copy constructors of subclasses

public:
  virtual str serialize() const = 0;
  
  chordID dir()  const { return _dir; }
  chordID inum() const { return _inum; }
  str name()     const { return _name; }
  int itype()    const { return _itype; }
};

class link_log_record : public dirop_log_record {
public:
  link_log_record (const link_log_record &r) { *this = r; }
  link_log_record (chordID inum, int itype, chordID dir, str name);
  link_log_record (xin *);
  str serialize () const;
};

class unlink_log_record : public dirop_log_record {
public:
  unlink_log_record (const unlink_log_record &r) { *this = r; }
  unlink_log_record (chordID inum, int itype, chordID dir, str name);
  unlink_log_record (xin *);
  str serialize () const;
};

class rename_log_record : public dirop_log_record {
private:
  chordID _olddir;
  str _oldname;
  bool _deleted;
  chordID _delinum;

public:
  rename_log_record (const rename_log_record &r) { *this = r; }
  rename_log_record (chordID inum, int itype, chordID olddir, str oldname,
                     chordID dir, str name, chordID delinum);
  rename_log_record (chordID inum, int itype, chordID olddir, str oldname,
                     chordID dir, str name);
  rename_log_record (xin *);
  str serialize () const;
  
  chordID olddir()  const { return _olddir; }
  str oldname()     const { return _oldname; }
  bool deleted()    const { return _deleted; }
  chordID delinum() const { return _delinum; }
};

const strbuf &strbuf_cat(const strbuf &b, log_record *e);

typedef ptr<log_record> recordp;
typedef callback<void, bool>::ref cb_bool;
typedef callback<void, recordp>::ref cb_recordp;

inline uint64
make_fileid(chordID inum)
{
  char ib[20];
  mpz_get_rawmag_be(ib, sizeof(ib), &inum);
  uint64 fileid;
  bcopy(ib, &fileid, sizeof(fileid));
  return(fileid);
}

inline bool
operator<(const nfstime3 &t1, const nfstime3 &t2)
{
  unsigned int ts1 = t1.seconds;
  unsigned int tns1 = t1.nseconds;
  unsigned int ts2 = t2.seconds;
  unsigned int tns2 = t2.nseconds;
  ts1 += (tns1/1000000000);
  tns1 = (tns1%1000000000);
  ts2 += (tns2/1000000000);
  tns2 = (tns2%1000000000);
  return (ts1 < ts2 || (ts1 == ts2 && tns1 < tns2));
}

inline chordID 
nfs2cid (nfs_fh3 *nfh) 
{
  chordID n;
  mpz_set_rawmag_be (&n, nfh->data.base (), nfh->data.size ());
  return n;
}

inline void
cid2nfs (chordID n, nfs_fh3 *nfh)
{
  str raw = n.getraw();
  nfh->data.setsize(raw.len());
  memcpy (nfh->data.base(), raw.cstr(), raw.len());
}

inline void
apply_sattr (fattr3& f, sattr3& s)
{
  bool ctime = false;

  if (s.size.set) {
    f.size = (uint64) *s.size.val;
    ctime = true;
  }

  if (s.mode.set) {
    f.mode = (uint32) *s.mode.val;
    ctime = true;
  }

  if (s.mtime.set == SET_TO_CLIENT_TIME) {
    f.mtime = (nfstime3) *s.mtime.time;
    ctime = true;
  }

  else if (s.mtime.set == SET_TO_SERVER_TIME) {
    f.mtime.seconds = tsnow.tv_sec;
    f.mtime.nseconds = tsnow.tv_nsec;
    ctime = true;
  }

  if (s.uid.set) {
    f.uid = (uint32) *s.uid.val;
    ctime = true;
  }

  if (s.gid.set) {
    f.gid = (uint32) *s.gid.val;
    ctime = true;
  }

  if (ctime) {
    f.ctime.seconds = tsnow.tv_sec;
    f.ctime.nseconds = tsnow.tv_nsec;
  }
}

inline
const strbuf &
strbuf_cat(const strbuf &b, nfs_fh3 fh)
{
  str s = armor32(fh.data.base(), fh.data.size());
  b << s;
  return b;
}

#endif

