
#include "scan.h"
#include "rxx.h"
#include "ranges.h"
#include "debug.h"
#include "logiter.h"

void
logscan::scan_log (cbliter cb, ptr<materialized_view> mv)
{
  vNew refcounted<logiter> (cb, mv);
}

void
logscan::scan_log (cbliter cb, ptr<materialized_view> mv, vervec stop)
{
  vNew refcounted<logiter> (cb, mv, stop);
}

//
// find a file's inode record; there's only one, corresponds with file
// creation. collect incremental file attributes.
//
void
logscan::find_inode (ptr<materialized_view> mv, vervec stop,
                     chordID inum, ptr<cattr> ca, cb_inode cb)
{
  scan_log (wrap (&logscan::find_inode_log_iterator, inum, cb, ca), mv, stop);
}

#define update_ctime(X,Y) \
  if(((X)->ctime < (Y)->ts())) { \
    (X)->ctime = (Y)->ts(); \
  }

#define update_mtime(X,Y) \
  if(!(X)->valid_mtime || \
      ((X)->mtime < (Y)->ts())) { \
    (X)->mtime = (Y)->ts(); \
    (X)->valid_mtime = 1; \
  }

void
logscan::find_inode_log_iterator
(chordID inum, cb_inode cb, ptr<cattr> cfsa, iterinfo *i)
{
  if (i->finished) {
    cb((recordp)0, cfsa, NFS3ERR_ABORTED);
    return;
  }
  recordp e = i->record;

  // handle attributes; can appear in attr_log_record and inode_log_record
  inode_log_record *ie = 0;
  if(e->template record<attr_log_record>() ||
     e->template record<inode_log_record>()) {
    attr_log_record *ae;
    cattr *ca = 0;
    chordID linum = 0;

    if((ae = e->template record<attr_log_record>())) {
      ca = ae->attr();
      linum = ae->inum();
    } else if((ie = e->template record<inode_log_record>())) {
      ca = ie->attr();
      linum = ie->inum();
    }

    // getting attributes from inode record or attribute record
    if(ca && linum == inum) {
      /*
       * size attribute needs attention.
       *
       * Use size attribute if it's the most recent in the log AND it sets the
       * file size LARGER or EQUAL than implied by any subsequent writes.
       * Otherwise ignore it.
       */
      if(ca->valid_size) {
        if(!cfsa->truncated && ca->size >= cfsa->size) {
          cfsa->size = ca->size;
          cfsa->valid_size = true;
        }
        update_mtime(cfsa, e);
        cfsa->truncated = true;
      }

      // Make sure file creation sets mtime.
      if(e->template record<inode_log_record>())
        update_mtime(cfsa, e);

      // Any change to inode sets ctime.
      update_ctime(cfsa, e);

      // transfer other attributes to fs->ca, unless they are set already
      cfsa->update(ca);
    }
  }

  // infer size from write_log_record. writes to cattr
  // don't do write is this file has been truncated
  write_log_record *we = e->template record<write_log_record> ();
  if (we && we->inum() == inum) {
    // if previously truncated, then don't change the size
    if (we->offset() + we->count() > cfsa->size && !cfsa->truncated) {
      cfsa->size = we->offset() + we->count();
      cfsa->valid_size = 1;
      update_ctime(cfsa, we);
    }
    update_mtime(cfsa, we);
  }

  unlink_log_record *ue = e->template record<unlink_log_record> ();
  if (ue) {
    if (ue->dir() == inum) {
      // only effect on directory is to change times.
      update_mtime(cfsa, ue);
      update_ctime(cfsa, ue);
      // remove a directory record that is also a directory
      if (ue->itype() == inode_log_record::T_DIR)
	// unlink from a directory, decr link count
	cfsa->nlink--;
    }
    else if (ue->inum() == inum) {
      // always change file's ctime.
      // never change file's mtime.
      // change file's link count.
      update_ctime(cfsa, ue);
      cfsa->nlink--;
    }
  }

  rename_log_record *re = e->template record<rename_log_record> ();
  if (re) {
    if (re->dir() == inum || re->olddir() == inum) {
      // only effect on directory is to change times.
      update_mtime(cfsa, re);
      update_ctime(cfsa, re);
      // add/remove a directory record that is also a directory
      if (re->itype() == inode_log_record::T_DIR) {
	// rename into a directory, incr link count
	if (re->dir() == inum)
          cfsa->nlink++;
	// rename, dealt with incr link count earlier, here checks if
	// removed a link from same directory
	if (re->olddir() == inum)
	  cfsa->nlink--;
      }
    }
    else if (re->inum() == inum) {
      // always change file's ctime.
      // never change file's mtime.
      // don't change file's link count.
      update_ctime(cfsa, re);
    }
    else if (re->deleted() && re->delinum() == inum) {
      // rename removed a file
      update_ctime(cfsa, re);
      cfsa->nlink--;
    }
  }

  link_log_record *le = e->template record<link_log_record> ();
  if (le) {
    if (le->dir() == inum) {
      // only effect on directory is to change times.
      update_mtime(cfsa, le);
      update_ctime(cfsa, le);
      // add a directory record that is also a directory
      if (le->itype() == inode_log_record::T_DIR)
	// link into a directory, incr link count
        cfsa->nlink++;
    }
    else if (le->inum() == inum) {
      // always change file's ctime.
      // never change file's mtime.
      // change file's link count.
      update_ctime(cfsa, le);
      cfsa->nlink++;
    }
  }

  if (e->template record<inode_log_record>() && ie && ie->inum() == inum) {
    if (inum == i->iter.root())
      // root inum, no link record, but we should incr nlink
      cfsa->nlink++;
    WARN(2) << "in T_INODE, after attrs, mtime = "
            << cfsa->mtime.seconds << "\n";
    cb (e, cfsa, NFS3_OK);
    i->finished = true;
  }
}

//
// lookup a directory by its inum, when successful returns either a
// rename or link log record that links the directory. expects inum to
// be that of a dir (i.e. does not check if it isn't).
//
void
logscan::find_dir_parent
(ptr<materialized_view> mv, vervec stop, chordID inum, cb_chordid cb)
{
  ptr<vec<chordID> > v = 0;
  scan_log(wrap(&logscan::find_dir_parent_log_iterator, inum, cb), mv, stop);
}

void
logscan::find_dir_parent_log_iterator(chordID inum, cb_chordid cb, iterinfo *i)
{
  nfsstat3 status = NFS3_OK;
  recordp e;
  unlink_log_record *ur = 0;
  rename_log_record *rr = 0;
  inode_log_record *nr = 0;
  
  e = i->record;
  if (e)
    nr = e->template record<inode_log_record>();

  if (i->finished || !e || (nr && nr->inum() == inum)) {
    if (nr && nr->inum() == inum)
      status = NFS3ERR_NOENT;
    else
      status = NFS3ERR_ABORTED;
    goto done;
  }
 
  // if inum has been removed, return NOENT. do this before we check
  // if inum() == inum, otherwise delinum() == inum will never be
  // checked
  ur = e->template record<unlink_log_record> ();
  rr = e->template record<rename_log_record> ();
  if ((ur && ur->inum() == inum)
      || (rr && rr->deleted() && rr->delinum() == inum)) {
    status = NFS3ERR_NOENT;
    goto done;
  }
  
  // if not a dirop or dirop is not on the inum in question, return
  if (!e->dirop_record() || e->dirop_record()->inum() != inum)
    return;

  WARN(1) << "found inum: " << inum << " in dir "
          << e->dirop_record()->dir() << "\n";
  cb(e->dirop_record()->dir(), status);
  i->finished = true;
  return;

done:
  WARN(1) << "did not find inum " << inum << "\n"; 
  cb(0, status);
  i->finished = true;
  return;
}

//
// lookup a file name in a directory, given the directory inum.
//
void
logscan::lookup
(ptr<materialized_view> mv, vervec stop,
 chordID dir, str name, cb_lookup cb, bool check_intend)
{
  ptr<vec<chordID> > v = 0;
  if (check_intend)
    v = New refcounted<vec<chordID> >;
  ref<lookup_state> lu = New refcounted<lookup_state>(dir, name, cb, v);
  scan_log(wrap(&logscan::lookup_log_iterator, lu), mv, stop);
}

void
logscan::lookup_log_iterator(ref<lookup_state> lu, iterinfo *i)
{
  nfsstat3 status = NFS3_OK;
  chordID dir = lu->dir;
  str name = lu->name;
  ptr<vec<chordID> > ci = lu->ci;
  
  recordp e;
  
  link_log_record *lr = 0;
  unlink_log_record *ur = 0;
  rename_log_record *rr = 0;
  intend_log_record *ir = 0;
  inode_log_record *nr = 0;
  str lname;
  
  e = i->record;
  if (e)
    nr = e->template record<inode_log_record>();

  if (i->finished || !e || (nr && nr->inum() == dir)) {
    if (nr && nr->inum() == dir)
      status = NFS3ERR_NOENT;
    else
      status = NFS3ERR_ABORTED;
    goto done;
  }
  
  lr = e->template record<link_log_record> ();
  rr = e->template record<rename_log_record> ();
  ur = e->template record<unlink_log_record> ();
  ir = e->template record<intend_log_record> ();

  if (ci && ir) {
    // if we run into a non-canceled intend record that matches, stop
    // scan and trigger callback
    //
    // if we run into a canceled intend record, we need to keep on
    // going, but remember that we have seen a canceled intend record
    // by this user.
    
    if (ir->dir() == dir && ir->name() == name) {
      if (ir->canceled()) {
        ci->push_back(ir->headID());
	return;
      }
      else {
	for (unsigned j = 0; j < ci->size(); j++) {
          if ((*ci)[j] == ir->headID())
	    return;
	}
	// found an intend record that was not canceled
        lu->cb(e, status);
        i->finished = true;
	return;
      }
    }
  }

  // if not right directory or a file by that name has already been
  // deleted or renamed, return.

  if (!(lr || ur || rr) ||
       ((lr && lr->dir() != dir) ||
        (ur && ur->dir() != dir) ||
        (rr && rr->dir() != dir && rr->olddir() != dir)))
    return;

  if ((ur && ur->name() == name) ||
      (rr && rr->oldname() == name && rr->olddir() == dir)) {
    status = NFS3ERR_NOENT;
    goto done;
  }

  if (ur)
    return;
  else if (lr)
    lname = lr->name();
  else if (rr)
    lname = rr->name();
  else
    assert(0);

  if (lname == name) {
    WARN(1) << "found name: " << name << "\n";
    lu->cb(e, status);
    i->finished = true;
    return;
  }
  return;

done:
  WARN(1) << "did not find name " << name << "\n"; 
  lu->cb((recordp)0, status);
  i->finished = true;
  return;
}

//
// read contents of a file
//
void
logscan::read
(ptr<materialized_view> mv, vervec stop, ptr<read_state> rs, cb_read cb)
{
  memset(rs->buf, '\0', rs->count);
  scan_log(wrap(&logscan::do_read_got_chunk, rs, cb), mv, stop);
}

void
logscan::sfrxx(ptr<read_state> rs, write_log_record *e, uint64 gs, uint64 gl)
{
  // We have three ranges:
  //   Read RPC asked for offset/count.
  //   Write log record has offset/count.
  //   Gap in ranges has offset/count (gs/gl).
  // We want to copy the intersection of those ranges
  // into buf[].
  //
  uint64 start = rs->offset;
  if(e->offset() > start)
    start = e->offset();
  if(gs > start)
    start = gs;

  uint64 end = rs->offset + rs->count;
  if(e->offset() + e->count() < end)
    end = e->offset() + e->count();
  if(gs+gl < end)
    end = gs + gl;
  if(end < start)
    return;

  uint64 bi = start - rs->offset; // index into buf[].
  uint64 wi = start - e->offset(); // index into e->data().
  uint64 n = end - start;

  WARN(2) << "sfrxx " << gs << " " << gl << "\n";

  assert(bi <= rs->count && bi + n <= rs->count);
  assert(wi <= e->count() && wi + n <= e->count());

  memcpy(rs->buf + bi, e->data().cstr() + wi, n);
  return;
}

void
logscan::do_read_got_chunk(ptr<read_state> rs, cb_read cb, iterinfo *i)
{
  nfsstat3 status = NFS3_OK;
  recordp e = i->record;
  if (i->finished || e == 0) {
    status = NFS3ERR_ABORTED;
    i->finished = true;
  }
  else if (inode_log_record *ee = e->template record<inode_log_record>()) {
    if (ee->inum() == rs->inum)
      i->finished = true;
  }
  else if (write_log_record *ee = e->template record<write_log_record>()) {
    if (ee->inum() == rs->inum) {
      // copy just bytes in intersection of read/write/gaps:
      rs->rng->gaps(wrap(sfrxx, rs, ee));
      // mark this range as done.
      rs->rng->add(ee->offset(), ee->count());
      // check from _offset to _offset+_count+1. the +1 is there so we
      // can tell if we are at the end of the file or not
      if (rs->rng->check(rs->offset, rs->count + 1))
        i->finished = true;
    }
  }
  else if (attr_log_record *ee = e->template record<attr_log_record>()) {
    if (ee->inum() == rs->inum) {
      cattr *a = ee->attr();
      if (a->valid_size) {
	// size in the first SETATTR is remembered, so we won't return
	// data past the size
	if (!rs->valid_size && a->size > rs->rng->end()) { 
          rs->valid_size = true;
	  rs->size = a->size;
	}
        // only regions inside the a->size can still be read,
        // everything outside are removed by the setattr
	WARN(2) << "do_read_got_chunk: attr size is " << a->size
	        << " request " << rs->offset << "+" << rs->count << "\n";
        if (a->size >= rs->offset && 
            a->size < (rs->offset+rs->count)) {
          rs->rng->add(a->size, rs->offset+rs->count-a->size);
	  WARN(2) << "cfsrw: attr filled buf " << a->size << " " 
		  << rs->offset+rs->count-a->size  << "\n";
	  WARN(2) << "end is now: " << rs->rng->end() << "\n";
	  // check from _offset to _offset+_count+1. the +1 is there
	  // so we can tell if we are at the end of the file or not
          if (rs->rng->check(rs->offset, rs->count + 1))
            i->finished = true;
        }
        else if (a->size < rs->offset) {
          rs->rng->add(rs->offset, rs->count);
          i->finished = true;
        }
      }
    }
  }

  if (i->finished) {
    uint64 c = rs->rng->end() - rs->offset;
    if (c > rs->count)
      c = rs->count;
    bool eof = (rs->offset + c == rs->rng->end());
    if (rs->valid_size) {
      if (rs->offset > rs->size) {
	c = 0;
	eof = true;
      }
      else if (rs->offset+c > rs->size) {
	c = rs->size-rs->offset;
	eof = true;
      }
    }
    WARN(2) << "read return with " << rs->offset << "+" << c << "\n";
    rs->rcount = c;
    rs->eof = eof;
    cb(rs, status);
    i->finished = true;
  }
}

//
// read a directory
//
void
logscan::readdir
(ptr<materialized_view> mv, vervec stop,
 chordID dir, ptr<readdir_state> rs, cb_readdir cb)
{
  scan_log(wrap(&logscan::scan_dir_log_iterator, dir, rs, cb), mv, stop);
}

bool
logscan::readdir_check_dir_entry
(ptr<readdir_state> rs, chordID dir, cb_readdir cb,
 str name, chordID inum, bool finished)
{
  assert(rs);
  if (name == "") {
    if (finished)
      return cb(rs, name, NFS3_OK);
    else
      return cb(rs, name, NFS3ERR_ABORTED);
  }

  // skip entries returned earlier. this won't work for large dir and
  // there are changes between two readdir requests, but then the
  // inconsistency is only for those two requests only, subsequent
  // requests should be fine.
  if (rs->cookie && rs->n < rs->cookie) {
    rs->n++;
    return false;
  }

  // make sure we don't take deleted or renamed inodes into account
  for (unsigned i=0; i<rs->changed.size(); i++) {
    unlink_log_record *ue =
      rs->changed[i]->template record<unlink_log_record> ();
    rename_log_record *re =
      rs->changed[i]->template record<rename_log_record> ();
    if((ue && ue->name() == name) ||
       // skip if
       // 1. rename/move of some sort affected file in this directory
       // 2. when this name was renamed to something else
       (re &&
          ((re->olddir() == dir && re->oldname() == name) ||
           (re->deleted() && re->delinum() == inum))))
      return false;
  }

  // there is an record. call cb, which may stop us from continuing.
  if (cb(rs, name, NFS3_OK))
    return true;

  (*rs->direntp).alloc();
  (*rs->direntp)->name = name;
  (*rs->direntp)->fileid = make_fileid(inum);
  rs->n++;
  (*rs->direntp)->cookie = rs->n;
  rs->direntp = &(*rs->direntp)->nextentry;
  return false;
}

void
logscan::scan_dir_log_iterator
(chordID dir, ptr<readdir_state> rs, cb_readdir cb, iterinfo *i)
{
  recordp e = i->record;
  inode_log_record *n = 0;
  if (e)
    n = e->template record<inode_log_record>();

  if (i->finished || !e || (n && n->inum() == dir)) {
    bool finished = false;
    if (n && n->inum() == dir)
      finished = true;
    readdir_check_dir_entry (rs, dir, cb, "", 0, finished);
    i->finished = true;
    return;
  }

  link_log_record *lr = e->template record<link_log_record>();
  unlink_log_record *ur = e->template record<unlink_log_record>();
  rename_log_record *rr = e->template record<rename_log_record>();

  if (!(lr || ur || rr) ||
       ((lr && lr->dir() != dir) ||
        (ur && ur->dir() != dir) ||
        (rr && rr->dir() != dir && rr->olddir() != dir)))
    return;

  // file deleted or renamed
  if (ur || rr)
    rs->changed.push_back(e);

  // if deleted, bail out (optimization only).
  // if renamed, still consider new name
  if (ur)
    return;

  // found a directory record?
  str name = e->dirop_record()->name();
  chordID inum = e->dirop_record()->inum();
  if (readdir_check_dir_entry (rs, dir, cb, name, inum, false)) {
    i->finished = true;
    WARN(1) << "scan_dir_log_iterator found directory record\n"; 
  }
}

