
#ifndef CFSRW_VIEW_H
#define CFSRW_VIEW_H

#include "refcnt.h"
#include "dhash.h"
#include "serialize.h"

class materialized_view;
class view : public virtual refcount {
private:
  chordID _root;
  vec<chordID> _head;
  vec<chordID> _latest;

  void load_record(ptr<materialized_view>,
                   callback<void, ptr<materialized_view> >::ref, recordp);
  void got_view(callback<void, ptr<view> >::ref cb, ptr<dhash_block> blk);
  void sort();

public:
  view() {}
  view(callback<void, ptr<view> >::ref cb, chordID v);

  chordID root() const { return _root; }
  unsigned nlogs() const { return _head.size(); }
  chordID head(unsigned i) const { return _head[i]; }
  chordID latest(unsigned i) const { return _latest[i]; }

  void root(chordID n) { _root = n; }

  void add_log(chordID h, chordID b) {
    for (unsigned i=0; i<_head.size(); i++) {
      if (_head[i] == h) {
	_latest[i] = b;
	return;
      }
    }
    _head.push_back(h);
    _latest.push_back(b);
    sort();
  }

  str serialize() {
    sort(); // redundancy to deal with my paranoia
    strbuf sb;
    sid(sb, root());
    sint(sb, nlogs());
    for (unsigned i=0; i<nlogs(); i++) {
      sid(sb, head(i));
      sid(sb, latest(i));
    }
    return sb;
  }

  void parse(xin *in) {
    int n;
    in->rid(_root);
    in->rint(n);
    for (int i=0; i<n; i++) {
      chordID h, b;
      in->rid(h);
      in->rid(b);
      add_log(h, b);
    }
  }

  chordID store(callback<void, bool, chordID>::ref cb);
  void materialize(callback<void, ptr<materialized_view> >::ref cb);
};

// a materialized view is a different representation of a view.
// instead of chordID of the log head blocks, a materialized view
// keeps the most recent log record to use for each log. only the view
// object can create a materialized view object.

class materialized_view {
protected:
  friend class view;
  chordID _root;
  vec<chordID> _head;
  vec<ptr<log_record> > _head_record;
  vec<ptr<log_record> > _latest;
  vervec _version;
  unsigned _fetched;

  void add_head_record(ptr<log_record> r) {
    assert(r);
    for(unsigned i=0; i<_head.size(); i++) {
      if (_head[i] == r->headID()) {
        _head_record[i] = r;
	return;
      }
    }
    assert(0);
  }

  void add_record(ptr<log_record> r) {
    assert(r);
    for(unsigned i=0; i<_head.size(); i++) {
      if (_head[i] == r->headID()) {
        _latest[i] = r;
	_fetched++;
	return;
      }
    }
    assert(0);
  }

  bool finished() const { return _fetched == _head.size(); }

  void make_version() {
    _version.clear();
    for (unsigned i=0; i<_head.size(); i++)
      _version.add(_head[i], _latest[i]->selfID(), _latest[i]->seqnum());
  }

  materialized_view() { _fetched = 0; }

public:

  chordID head(unsigned i) const { return _head[i]; }
  unsigned nlogs() const { return _latest.size(); }
  ptr<log_record> latest(unsigned i) const { return _latest[i]; }
  ptr<log_record> head_recordp(unsigned i) const { return _head_record[i]; }

  head_log_record* head_record(unsigned i) const {
    return _head_record[i]->template record<head_log_record>();
  }

  vervec version() const { return _version; }
  chordID root() const { return _root; }
};

#endif

