#!/usr/bin/perl -w

use Env qw(HOME LD_LIBRARY_PATH PATH);

my $verbose = 1;
my $interactive = 1;

# preset
my $SFS_PREFIX = "/usr/local/";
my $CHORD_PREFIX = "/usr/local/";
my $CFSRW_PREFIX = "/usr/local/";
my $CFSRW = "$HOME/.cfsrw/";
my $CFSRW_ROOT = "/cfsrw";
my $LSD_SOCKET = "/tmp/cfsrw-sock";
my $PORT = 11977;
my $HOSTFILE = "$CFSRW/hosts.txt";

my $multhosts = 0;


use Cwd;
my $cwd = cwd();
my $who = `whoami`;
chomp $who;
my $tfseq = 1;

&process_args(@ARGV);
@ARGV = ();

my $SFSKEY_PROG = "sfskey";
my $SFSCD_PROG = "sfscd";
my $LSD_PROG = "lsd";
my $CFSLOG_PROG = "cfslog";
my $CFSAGENT_PROG = "cfsagent";
my $VIEWCTL_PROG = "viewctl";
my $CFSRWCD_PROG = "cfsrwcd";

my @SEARCH_PATH = ( "/usr/local/bin",
                    "/usr/local/sbin",
                    $CHORD_PREFIX,
                    "$CHORD_PREFIX/lsd",
                    "$CHORD_PREFIX/bin",
                    $CFSRW_PREFIX,
                    "$CFSRW_PREFIX/bin",
                    $SFS_PREFIX,
                    "$SFS_PREFIX/bin",
                    "$SFS_PREFIX/sbin",
                    $cwd,
                    "$cwd/..",
                    "$cwd/cfsrwcd",
                    "$cwd/../cfsrwcd" );

my %hosts = ();
my $localhost = `hostname -s`;
chomp $localhost;

&main();

#
# Processes arguments and sets options.
#
sub process_args {
  my @args = @_;

  while(@args) {
    $_ = shift @args;
    if(/^--cfsrw$/)             { $CFSRW = shift @args; next; }
    if(/^--sfsprefix$/)         { $SFS_PREFIX = shift @args; next; }
    if(/^--chordprefix$/)       { $CHORD_PREFIX = shift @args; next; }
    if(/^--cfsrwprefix$/)       { $CFSRW_PREFIX = shift @args; next; }
    if(/^--socket$/)            { $LSD_SOCKET = shift @args; next; }
    if(/^--port$/)              { $PORT = shift @args; next; }
    if(/^--root$/)              { $CFSRW_ROOT = shift @args; next; }
    if(/^--hosts$/)             { $HOSTFILE = shift @args; next; }

    if(/^-v$/ || /^--verbose$/) { $verbose++; next; }
    if(/^\+v$/)                 { $verbose--; next; }
    if(/^-s$/ || /^--silent$/)  { $verbose = 0; next; }

    if(/^-i$/)                  { $interactive = 1; next; }
    if(/^\+i$/)                 { $interactive = 0; next; }

    if(/^-m$/)                  { $multhosts = 1; next; }
    if(/^\+m$/)                 { $multhosts = 0; next; }

    if(/^-h$/)                  { &usage() and exit; }

    print "Unrecognized option $_\n";
    usage();
    exit(1);
  }
}


sub usage() {
  print STDERR <<EOF;
--sfsprefix             : prefix for SFS executables, default $SFS_PREFIX
--chordprefix           : prefix for Chord executables, default $CHORD_PREFIX
--cfsrwprefix           : prefix for CFSRW executables, default $CFSRW_PREFIX
--cfsrw                 : CFSRW config dir, default $CFSRW
--socket                : LSD_SOCKET, default $LSD_SOCKET
--port                  : port, default $PORT
--root                  : cfsrw root directory, default $CFSRW_ROOT
--key                   : key file, default $CFSRW/keyfile.[host]
--hosts                 : hosts file, default $CFSRW/hosts.txt
-v,+v                   : increase, decrease verbosity
-s                      : silent mode, no verbosity
-i,+i                   : interactive or assume defaults
-k,+k                   : skip, don't skip cfsrw directory/file setup
-m,+m                   : multiple hosts, creates shared file system for all hosts
EOF
}

sub sshdo {
  # host, user, command, ssh-options, flag whether or not to return
  # returnvalue or content
  my ($h, $u, $c, $o, $f) = @_;

  not defined $o and $o = "";
  not defined $f and $f = 0;

  if($h eq $localhost and $u eq $who) {
    if($f) {
      return `perl -e '$c and exit 1 or exit 0'`;
    } else {
      return system "perl -e '$c and exit 1 or exit 0'";
    }
  }

  my $tf = "/tmp/cfsrwsetup-$$-$tfseq.pl";
  $tfseq += 1;

  my $rc = rcmd($u, $h, "cat > $tf");
  open(FILE, "| $rc") or die "Cannot write $h:$tf";
  print FILE "$c and exit 1 or exit 0\n";
  close FILE;

  if($h eq $localhost){
      $h = "127.0.0.1";
  }

  $verbose >= 2 and print "Running $c on $u\@$h\n";
  if($f) {
    return `ssh $o $u\@$h 'perl $tf ; rm $tf'`;
  } else {
    return system "ssh $o $u\@$h 'perl $tf ; rm $tf'";
  }
}


sub main {
  # read the hosts file
  if($multhosts) {
    open FILE, "<$HOSTFILE" or die "Cannot open $HOSTFILE";
    while(<FILE>) {
      my ($h, $u, $p) = split /\s+/;
      $hosts{$h}{USER} = $u;
      $hosts{$h}{PATH} = join ":", @SEARCH_PATH;
      $hosts{$h}{PATH} .= ":$p";
    }
  } else {
    # add localhost to hosts
    $hosts{$localhost}{USER} = $who;
    $hosts{$localhost}{PATH} = $ENV{'PATH'} . ":" . join(":", @SEARCH_PATH);
  }


  # host to loginfo key mapping
  my %hostkeys;

  # master host is the one that creates the shared view
  my $master = (keys %hosts)[0];

  # key of view created on master
  my $masterview;

  foreach my $host (keys %hosts) {
    my $user = $hosts{$host}{USER};

    # create ~/.cfsrw
    unless(&sshdo($host, $user, "-d \"$CFSRW/$host/\"")) {
      &ask("Can I create $CFSRW/$host/ on $user\@$host?", "y") or exit;
      &sshdo($host, $user, "mkdir (\"$CFSRW/$host/\", 0755)") or die "Failed";
    }

    # create ~/.cfsrw/etc/ and contents
    unless(&sshdo($host, $user, "-d \"$CFSRW/$host/etc\"")) {
      &ask("Can I create $CFSRW/$host/etc and its contents on $user\@$host?", "y") or exit;
      &sshdo($host, $user, "mkdir (\"$CFSRW/$host/etc\", 0755)") or die "Failed";
      my $cfsrwcd = &sshdo($host, $user, "system \"export PATH=$hosts{$host}{PATH}; type cfsrwcd\"", "", 1);
      $cfsrwcd = (split /\s+/, $cfsrwcd)[2];
      if($cfsrwcd !~ /^\//){
          print STDERR "setup.pl: could not find cfsrwcd on $host\n";
          exit(1);
      }
      my $rc = rcmd($user, $host, "cat >  $CFSRW/$host/etc/sfscd_config");
      open(FILE, "| $rc") or die "Failed: $rc";
      print FILE <<EOF;
Release *
  Protocol chord $cfsrwcd
EOF
      close FILE;
    }

    # create ~/.cfsrw/runinplace/
    unless(&sshdo($host, $user, "-d \"$CFSRW/$host/runinplace\"")) {
      &ask("Can I create $CFSRW/$host/runinplace on $user\@$host?", "y") or exit;
      &sshdo($host, $user, "mkdir (\"$CFSRW/$host/runinplace\", 0777)") or die "Failed";
      &sshdo($host, $user, "chmod 0777, \"$CFSRW/$host/runinplace\"") or die "Failed";
    }


    # create /cfsrw
    unless(&sshdo($host, $user, "-d \"$CFSRW_ROOT\"")) {
      &ask("Can I create $CFSRW_ROOT on $user\@$host?", "y") or exit;
      unless(&sshdo($host, $user, "mkdir (\"$CFSRW_ROOT\", 0755)")) {
        print "Failed. Trying as root.\n";
        &sshdo("root", $host, "mkdir (\"$CFSRW_ROOT\", 0755)") or die "Couldn't create $CFSRW_ROOT on $user\@$host";
      }
    }

    $KEYFILE = "$CFSRW/$host/keyfile";
    $RUN_SCRIPT = "$CFSRW/$host/envrun.sh";

    # create keyfile
    unless(&sshdo($host, $user, "-f \"$KEYFILE\"")) {
      &ask("Can I create $KEYFILE on $user\@$host?", "y") or exit;
      # sfskey returns wrong result
      &sshdo($host, $user, "system \"$SFSKEY_PROG gen -KP $KEYFILE\"", "-t");
      &sshdo($host, $user, "-f \"$KEYFILE\"") or die "Hmm... It seems to have failed.";
    }

    # create script to run programs that need to run in correct environment
    unless(&sshdo($host, $user, "-f \"$RUN_SCRIPT\"")) {
      &ask("Can I create $RUN_SCRIPT on $user\@$host?", "y") or exit;
      my $rc = rcmd($user, $host, "cat > $RUN_SCRIPT");
      open(FILE, "| $rc") or die "Failed: $rc";
      print FILE "#!/bin/sh\n\n";
      if(defined($LD_LIBRARY_PATH)){
          print FILE "export LD_LIBRARY_PATH=$LD_LIBRARY_PATH\n";
      }
      my $p = $hosts{$host}{PATH};
      defined $p or $p = "";
      print FILE <<EOF;
export PATH=$PATH\:$p
export SFS_RUNINPLACE=$CFSRW/$host/
export SFS_ROOT=$CFSRW_ROOT
export SFS_PORT=$PORT
export CFSRW_KEYFILE=$KEYFILE
export LSD_SOCKET=$LSD_SOCKET

\$\*
EOF
      close FILE;
      &sshdo($host, $user, "chmod 0755, \"$RUN_SCRIPT\"");
    }

    # run lsd
    my $run = 1;
    my $procno = &procs($user, $host, "lsd");
    if($procno) {
      &ask("Detected a running lsd on $host ($procno). Run anyway?", "n") or $run = 0;
    }
    if($run) {
      if($multhosts) {
        if($host eq $master) {
          &sshdo($host, $user, "system \"$RUN_SCRIPT $LSD_PROG -j $host\:10001 -p 10001 -S $LSD_SOCKET > lsd.trace 2>&1 &\"", "-f");
        } else {
          &sshdo($host, $user, "system \"$RUN_SCRIPT $LSD_PROG -j $master\:10001 -S $LSD_SOCKET > lsd.trace 2>&1 &\"", "-f");
        }
      } else {
          my $cmd = "$LSD_PROG -j 127.0.0.1:10001 -p 10001 -l 127.0.0.1 -S $LSD_SOCKET > lsd.trace 2>&1 &";
          $verbose >= 1 and print "Running $cmd\n";
          system($cmd);
      }

      $verbose >= 1 and print "Wait 10 seconds for lsd to stabilize.\n";
      sleep 10;
      &sshdo($host, $user, "chmod 0777, \"$LSD_SOCKET\"");
    }

    # run sfscd
    &sshdo($host, $user, "-x \"$RUN_SCRIPT\"") or die "Couldn't locate sfscd.sh at $RUN_SCRIPT";
    $run = 1;
    my $kill = 0;
    
    $procno = &procs($user, $host, "sfscd -d");
    if($procno) {
      &ask("Detected a running sfscd on $host ($procno) . Run anyway?", "y") or $run = 0;
      if($run && ask("Kill the old sfscd on $host ($procno)?", "y")){
          my $x = $procno;
          $x =~ s/,/ /g;
          system(rcmd("root", $host, "kill -9 $x"));
      }
    }
  
    if($run) {

        $verbose >= 1 and print "Running sfscd\n";

        my $rc = rcmd("root", $host, "$RUN_SCRIPT $SFSCD_PROG -d",
                      "sfscd.trace");
        system($rc);

        $verbose >= 1 and print "Waiting for sfscd...\n";
        sleep 5;
    }

    # run cfsagent
    $run = 1;
    $procno = &procs($user, $host, "cfsagent");
    if($procno) {
        if(ask("Detected a running cfsagent on $host ($procno). Restart?", "y")){
            my $x = $procno;
            $x =~ s/,/ /g;
            system(rcmd($user, $host, "kill -9 $x"));
        }
    }
    if($run) {
      $verbose >= 1 and print "Running $CFSAGENT_PROG\n";
      my $rc = rcmd($user, $host, "$RUN_SCRIPT $CFSAGENT_PROG -k $KEYFILE",
                    "cfsagent.trace");
      system($rc);
      sleep(5);
    }

    # create new log
    $verbose >= 1 and print "Creating fresh log on $user\@$host.\n";
    my $key = &sshdo($host, $user, "system \"$RUN_SCRIPT $CFSLOG_PROG -s $KEYFILE | head -1\"", "", 1);
    chomp $key;
    $hostkeys{$host} = $key;

    # create new view
    $verbose >= 1 and print "Creating fresh view on $user\@$host.\n";
    $key = &sshdo($host, $user, "system \"$RUN_SCRIPT $VIEWCTL_PROG -c $key | tail -1\"", "", 1);
    chomp $key;
    $host eq $master and $masterview = $key;

    print "File system on $host :\n";
    print "cd $CFSRW_ROOT/$key\n";
  }

  return unless $multhosts;

  # create a combined view on one of the machines
  $user = $hosts{$master}{USER}; # which user to run as on master
  my $finalkey;
  foreach my $host (keys %hosts) {
    $host eq $master and next;
    $finalkey = &sshdo($master, $user, "system \"$CFSRW/$master/envrun.sh $VIEWCTL_PROG -a $hostkeys{$host} $masterview | tail -1\"", "", 1);
  }

  print "The shared view is: \n";
  print "cd $CFSRW_ROOT/$finalkey\n";
}


sub procs {
  my ($user, $host, $proc) = @_;

  my $procno;
  my $cmd = "ps -ax | grep \"$proc\" | grep -v grep | grep -v defunct | awk '{print \$1}'";
  if($host eq $localhost) {
    $procno = `$cmd`;
  } else {
    $procno = `ssh $user\@$host $cmd`;
  }

  chomp $procno;
  $procno =~ s/\n/, /g;
  return $procno;
}


sub ask {
  my ($question, $default) = @_;
  not defined $default and $default = "y";

  my $answer;
  if($interactive) {
    print "$question \[$default\] ";
    $answer = <>;
    chomp $answer;
  } else {
    $answer = $default;
  }


  $answer eq "" and $answer = $default;
  ($answer eq "y" || $answer eq "yes") and return 1;
  return 0;
}

# Return a command string that runs $cmd on $host as $user.
# The point is to run ssh if need be.
# Another point is to use 127.0.0.1 instead of the local host name,
# when possible, for DHCP hosts.
# If $bg is defined, the command will be run in the background,
# with stdout and stderr redirected to file $bg.
sub rcmd {
    my($user, $host, $cmd, $bg) = @_;
    
    if($host eq $localhost){
        $host = "127.0.0.1";
    }

    if($host eq "127.0.0.1" && $user eq $who){
        if(defined($bg)){
            $cmd .= " > $bg 2>&1 &";
        }
        return $cmd;
    } else {
        if(defined($bg)){
            return "ssh -f -x -a -l $user $host '$cmd' > $bg 2>&1 &";
        } else {
            return "ssh -x -a -l $user $host '$cmd'";
        }
    }
}
