/*
 *
 * Copyright (C) 2002 Athicha Muthitacharoen (athicha@lcs.mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#include "debug.h"
#include "snapshot.h"

snapshot_operation::snapshot_operation (ptr<snapshot> ss)
  : _ss(ss), _outstanding_fetches(0), _fetch_errors(0)
{
}

void
snapshot_operation::getroot
(ptr<snapshot> ss, callback<void, ptr<log_record> >::ref cb)
{
  ptr<snapshot_operation> op = New refcounted<snapshot_operation>(ss);
  op->getroot(cb);
}

void 
snapshot_operation::find_dir_parent
(ptr<snapshot> ss, chordID dir, cb_chordid cb)
{
  ptr<snapshot_operation> op = New refcounted<snapshot_operation>(ss);
  op->find_dir_parent(dir, cb);
}

void 
snapshot_operation::lookup
(ptr<snapshot> ss, chordID dir, str name, cb_lookup cb)
{
  ptr<snapshot_operation> op = New refcounted<snapshot_operation>(ss);
  op->lookup(dir, name, cb);
}

void 
snapshot_operation::find_inode
(ptr<snapshot> ss, chordID inum, ptr<cattr> fa, cb_inode cb)
{
  ptr<snapshot_operation> op = New refcounted<snapshot_operation>(ss);
  op->find_inode(inum, fa, cb);
}

void
snapshot_operation::readdir
(ptr<snapshot> ss, chordID dir, ptr<readdir_state> rs, cb_readdir cb)
{
  ptr<snapshot_operation> op = New refcounted<snapshot_operation>(ss);
  op->readdir(dir, rs, cb);
}

void
snapshot_operation::read (ptr<snapshot> ss, ptr<read_state> rs, cb_read cb)
{
  ptr<snapshot_operation> op = New refcounted<snapshot_operation>(ss);
  op->read(rs, cb);
}

void
snapshot_operation::getroot_fmap_cb
(callback<void, ptr<log_record> >::ref cb, ptr<dbrec> v)
{
  if (v)
    cfsrw_dhashclt::retrieve_sfsro_readonly
      (dbrec2id(v), wrap (mkref(this), &snapshot_operation::getroot_cb, cb));
  else
    (*cb)(0);
}

void
snapshot_operation::getroot (callback<void, ptr<log_record> >::ref cb)
{
  ptr<dbrec> q = id2dbrec (rootdir());
  fmap_db()->lookup
    (q, wrap (mkref(this), &snapshot_operation::getroot_fmap_cb, cb));
}

void 
snapshot_operation::getroot_cb (callback<void, ptr<log_record> >::ref cb, 
		                ptr<sfsro_data> rdata)
{
  ptr<log_record> le = NULL;
  if (!rdata) {
    (*cb)(0);
    return;
  }
  le = set_log_record (true, rdata);
  cb (le);
}

void
snapshot_operation::find_dir_parent_cb (chordID dir, cb_chordid cb,
                                        ptr<sfsro_data> data)
{
  if (!data) {
    cb (0, NFS3ERR_IO);
    return;
  }
  if (data->type != SFSRO_INODE) {
    cb (0, NFS3ERR_IO);
    return;
  }
  sfsro_inode *inode = (*data).inode;
  if (inode->type != SFSRODIR) {
    cb (0, NFS3ERR_IO);
    return;
  }
  chordID pid;
  mpz_set_rawmag_be
    (&pid, inode->reg->parent.base (), SFSRO_FHSIZE);
  cb (pid, NFS3_OK);
}

void
snapshot_operation::find_dir_parent_fmap_cb (chordID dir, cb_chordid cb,
                                             ptr<dbrec> v)
{
  if (!v) {
    cb (0, NFS3ERR_IO);
    return;
  }
  WARN(2) << "found directory: dir = " << id2str (dir) << "\n";
  cfsrw_dhashclt::retrieve_sfsro_readonly
    (dbrec2id(v),
     wrap (mkref(this), &snapshot_operation::find_dir_parent_cb, dir, cb));
}

void 
snapshot_operation::find_dir_parent(chordID dir, cb_chordid cb)
{
  ptr<dbrec> q = id2dbrec (dir);
  fmap_db()->lookup
    (q, wrap (mkref(this),
	      &snapshot_operation::find_dir_parent_fmap_cb, dir, cb));
}

void
snapshot_operation::lookup_fmap_cb (chordID dir, str name,
                                    cb_lookup cb, ptr<dbrec> v)
{
  if (!v) {
    cb ((ptr<log_record>)0, NFS3ERR_IO);
    return;
  }
  WARN(2) << "found directory: dir = " << id2str (dir) << "\n";
  cfsrw_dhashclt::retrieve_sfsro_readonly
    (dbrec2id(v),
     wrap (mkref(this), &snapshot_operation::lookup_cb1, dir, name, cb));
}

void 
snapshot_operation::lookup(chordID dir, str name, cb_lookup cb)
{
  ptr<dbrec> q = id2dbrec (dir);
  fmap_db()->lookup
    (q, wrap (mkref(this), &snapshot_operation::lookup_fmap_cb, dir, name, cb));
}

void
snapshot_operation::lookup_cb1
(chordID dir, str name, cb_lookup cb, ptr<sfsro_data> data)
{
  ptr<log_record> le = NULL;
  if (!data) {
    cb ((ptr<log_record>)0, NFS3ERR_IO);
    return;
  }
  if (data->type != SFSRO_INODE) {
    cb ((ptr<log_record>)0, NFS3ERR_IO);
    return;
  }
  sfsro_inode *inode = (*data).inode;
  if (inode->type != SFSRODIR) {
    cb ((ptr<log_record>)0, NFS3ERR_IO);
    return;
  }
  if (inode->reg->direct.size() > 0) {
    chordID dirblkID;
    mpz_set_rawmag_be
      (&dirblkID, inode->reg->direct[0].base (), SFSRO_FHSIZE);
    WARN(1) << "loading dirblk 0: " << dirblkID << "\n";
    cfsrw_dhashclt::retrieve_sfsro_readonly
      (dirblkID,
       wrap (mkref(this), &snapshot_operation::lookup_cb2,
	     dir, inode, 0, name, cb));  
  }
  else {
    // directory is empty
    cb ((ptr<log_record>)0, NFS3ERR_NOENT);
    return;
  }
}

void 
snapshot_operation::lookup_cb2 (chordID dir, sfsro_inode *inode,
                                uint32 blknum, str name,
				cb_lookup cb, ptr<sfsro_data> data)
{
  ptr<log_record> le = NULL;
  if (!data) {
    cb ((ptr<log_record>)0, NFS3ERR_IO);
    return;
  }
  if (data->type != SFSRO_DIRBLK) {
    cb ((ptr<log_record>)0, NFS3ERR_IO);
    return;
  }

  sfsro_dirent *e = (*(*data).dir).entries;
  while (e && (e->name != name)) {
    WARN(2) << "dir " << dir << " has: " << e->name << "\n";
    e = e->nextentry;
  }
  if (e) {
    WARN(2) << "found entry " << e->name << "\n";
  }
    
  if (!e) {
    if (inode->reg->direct.size() > blknum+1) {
      blknum++;
      chordID dirblkID;
      mpz_set_rawmag_be
        (&dirblkID, inode->reg->direct[blknum].base (), SFSRO_FHSIZE);
      WARN(1) << "loading dirblk " << blknum << ": " << dirblkID << "\n";
      cfsrw_dhashclt::retrieve_sfsro_readonly
        (dirblkID,
         wrap (mkref(this), &snapshot_operation::lookup_cb2,
	       dir, inode, blknum, name, cb));
      return;
    }
    cb ((ptr<log_record>)0, NFS3ERR_NOENT);
  }
  else {
    chordID fhID;
    mpz_set_rawmag_be (&fhID, e->fh.base (), SFSRO_FHSIZE);
    ptr<dbrec> q = id2dbrec (fhID);
    fmap_db()->lookup
      (q, wrap (mkref(this), &snapshot_operation::lookup_fmap_cb3,
		dir, name, cb));
  }
}

void
snapshot_operation::lookup_fmap_cb3 (chordID dir, str name,
                                     cb_lookup cb, ptr<dbrec> v)
{
  if (!v) {
    WARN(2) << "can't load sfsro data\n";
    cb ((ptr<log_record>)0, NFS3ERR_IO);
    return;
  }
  cfsrw_dhashclt::retrieve_sfsro_readonly
    (dbrec2id(v),
     wrap (mkref(this), &snapshot_operation::lookup_cb3, dir, name, cb)); 
}

void 
snapshot_operation::lookup_cb3
(chordID dir, str name, cb_lookup cb, ptr<sfsro_data> data)
{
  ptr<log_record> le = NULL;
  if (!data) {
    cb ((ptr<log_record>)0, NFS3ERR_IO);
    return;
  }
  if (data->type != SFSRO_INODE) {
    cb ((ptr<log_record>)0, NFS3ERR_IO);
    return;
  }
  WARN(2) << "lookup successful\n";
  le = set_log_record (true, data, dir, name);
  cb (le, NFS3_OK);
}

void 
snapshot_operation::get_inode_cb
(chordID inum, ptr<cattr> ca, cb_inode cb, ptr<sfsro_data> data)
{
  ptr<log_record> le = 0;

  if (!data) {
    cb (le, (ptr<cattr>)0, NFS3ERR_IO);
    return;
  }
  if (data->type != SFSRO_INODE) {
    cb (le, (ptr<cattr>)0, NFS3ERR_IO);
    return;
  }
  le = set_log_record (false, data);
  combine_attr (ca, data);
  cb (le, ca, NFS3_OK);
}

void
snapshot_operation::get_inode_fmap_cb (chordID inum, ptr<cattr> fa,
                                       cb_inode cb, ptr<dbrec> v)
{
  if (!v) {
    cb ((ptr<log_record>)0, (ptr<cattr>)0, NFS3ERR_STALE);
    return;
  }
  cfsrw_dhashclt::retrieve_sfsro_readonly
    (dbrec2id(v),
     wrap (mkref(this), &snapshot_operation::get_inode_cb, inum, fa, cb));
}

void 
snapshot_operation::find_inode(chordID inum, ptr<cattr> fa, cb_inode cb)
{
  ptr<dbrec> q = id2dbrec (inum);
  fmap_db()->lookup
    (q, wrap (mkref(this), &snapshot_operation::get_inode_fmap_cb,
	      inum, fa, cb));
}

void
snapshot_operation::got_dir_inode_fmap (chordID dir, ptr<readdir_state> rs,
                                        cb_readdir cb, ptr<dbrec> v)
{
  if (!v) {
    cb (rs, "", NFS3ERR_NOENT);
    return;
  }
  cfsrw_dhashclt::retrieve_sfsro_readonly
    (dbrec2id(v),
     wrap(mkref(this),
          &snapshot_operation::got_dir_inode, dir, rs, cb));
}

void
snapshot_operation::readdir
(chordID dir, ptr<readdir_state> rs, cb_readdir cb)
{
  ptr<dbrec> q = id2dbrec (dir);
  fmap_db()->lookup
    (q, wrap (mkref(this), &snapshot_operation::got_dir_inode_fmap,
	      dir, rs, cb));
}

void 
snapshot_operation::got_dir_inode
(chordID dir, ptr<readdir_state> rs, cb_readdir cb, ptr<sfsro_data> data)
{
  if (!data) {
    cb (rs, "", NFS3ERR_IO);
    return;
  }
  if (data->type != SFSRO_INODE) {
    cb (rs, "", NFS3ERR_IO);
    return;
  }
  sfsro_inode *inode = (*data).inode;
  if (inode->type != SFSRODIR) {
    cb (rs, "", NFS3ERR_NOTDIR);
    return;
  }
  if (inode->reg->direct.size() > 0) {
    chordID dirblkID;
    mpz_set_rawmag_be
      (&dirblkID, inode->reg->direct[0].base (), SFSRO_FHSIZE);
    cfsrw_dhashclt::retrieve_sfsro_readonly
      (dirblkID,
       wrap (mkref(this),
             &snapshot_operation::got_dir_block, dir, inode, 0, rs, cb));
  }
  else {
    cb (rs, "", NFS3_OK);
    return;
  }
}

void
snapshot_operation::got_dir_block (chordID dir, sfsro_inode *inode,
                                   uint32 blknum, ptr<readdir_state> rs,
				   cb_readdir cb, ptr<sfsro_data> data)
{
  if (!data) {
    warn << "did not get dir block " << blknum << "\n";
    cb (rs, "", NFS3ERR_IO);
    return;
  }
  if (data->type != SFSRO_DIRBLK) {
    warn << "wrong type of dir block " << blknum << "\n";
    cb (rs, "", NFS3ERR_IO);
    return;
  }
  
  chordID inum;
  bool skip;
  // bool eof = (*(*data).dir).eof;
  sfsro_dirent *e = (*(*data).dir).entries;
  while (e) {
    WARN(2) << "looking at dirent name " << e->name << "\n";
    skip = false;
    str2chordID (&inum, e->fh.base(), SFSRO_FHSIZE);

    for (uint i=0; i < rs->changed.size(); i++) {
      unlink_log_record *ue =
	rs->changed[i]->template record<unlink_log_record>();
      rename_log_record *re =
	rs->changed[i]->template record<rename_log_record>();
      if ((ue && ue->name() == e->name) ||
	  (re && ((re->olddir() == dir && re->oldname() == e->name) ||
		  (re->deleted() && re->delinum() == inum))))
	skip = true;
    }

    if (!skip) {
      // check if there's enough space in RPC reply, &c. this avoids a
      // crash when rmdir checks if a directory is empty, and also
      // perhaps when a directory is too large to fit in a single
      // READDIR reply.
      if (rs && !dup_entry(&(rs->nfsres.resok->reply.entries), e)) {
        if (rs->cookie && rs->n < rs->cookie)
	  rs->n++;
        else {
          if (cb (rs, e->name, NFS3_OK))
            return;
          WARN(2) << "including " << e->name << "\n";
          (*rs->direntp).alloc();
          (*rs->direntp)->name = e->name;
          (*rs->direntp)->fileid = e->fileid;
          rs->n++;
          (*rs->direntp)->cookie = rs->n;
          rs->direntp = &(*rs->direntp)->nextentry;
        }
      }
    }
    e = e->nextentry;
  }

  if (inode->reg->direct.size() > blknum+1) {
    blknum++;
    chordID dirblkID;
    mpz_set_rawmag_be
      (&dirblkID, inode->reg->direct[blknum].base (), SFSRO_FHSIZE);
    cfsrw_dhashclt::retrieve_sfsro_readonly
      (dirblkID,
       wrap (mkref(this),
             &snapshot_operation::got_dir_block, dir, inode, blknum, rs, cb));
    return;
  }

  cb (rs, "", NFS3_OK);
}

void
snapshot_operation::read_fmap_cb (ptr<read_state> r, cb_read cb, ptr<dbrec> v)
{
  if (!v) {
    rcb (rs, NFS3ERR_IO);
    return;
  }
  cfsrw_dhashclt::retrieve_sfsro_readonly
    (dbrec2id(v), wrap (mkref(this), &snapshot_operation::read_file_cb2));
}

void
snapshot_operation::read (ptr<read_state> r, cb_read cb)
{
  rs = r;
  rcb = cb;
  ptr<dbrec> q = id2dbrec (r->inum);
  fmap_db()->lookup
    (q, wrap (mkref(this), &snapshot_operation::read_fmap_cb, r, cb));
}

void
snapshot_operation::read_file_cb2(ptr<sfsro_data> data)
{
  if (!data) {
    rcb (rs, NFS3ERR_IO);
    return;
  }
  if (data->type != SFSRO_INODE) {
    rcb(rs, NFS3ERR_IO);
    return;
  }
  sfsro_inode *inode = (*data).inode;
  if (inode->type == SFSROLNK) {
    sfsro_inode_lnk *lnode = inode->lnk;
    if (!rs->rng->check (rs->offset, lnode->dest.len ())) {
      memcpy (rs->buf, lnode->dest.cstr (), lnode->dest.len ());
      rs->rcount = lnode->dest.len ();
      rs->eof = true;
      rcb(rs, NFS3_OK);
    }
  }
  else if (inode->type == SFSROREG) {
    sfsro_inode_reg *rnode = inode->reg;
    WARN(2) << "read snapshot " << rs->offset << "+" << rs->count << "\n";
    if (rs->rng->check(rs->offset, rs->count)) {
      // got all the bytes, in here to check if we hit EOF or not
      uint64 c = rs->rng->end() - rs->offset;
      if(c > rs->count)
	c = rs->count;
      bool eof = (rs->offset+c >= rnode->attr.size);
      rs->rcount = c;
      rs->eof = eof;
      rcb(rs, NFS3_OK);
    }
    else {
      // we only want to read what was asked, so bytes before
      // rs->offset are done, and bytes after rs->offset+rs->count,
      // til end of the version of the file stored in snapshot, are
      // done. also, by marking everything from rs->offset+rs->count
      // to eof as done, later on when we test for eof, we know if we
      // have hit eof or not.
      if (rs->offset > 0)
        rs->rng->add (0, rs->offset);
      if ((rs->offset+rs->count) < rnode->attr.size)
        rs->rng->add
	  (rs->offset+rs->count,rnode->attr.size-(rs->offset+rs->count));
      // if requested is more than file size, mark file size to
      // requested size as done. remember that a setattr log record
      // could have expanded the file size before scanning hit the
      // snapshot.
      if ((rs->offset+rs->count) > rnode->attr.size &&
	  rs->offset < rnode->attr.size)
	rs->rng->add(rnode->attr.size, rs->offset+rs->count);
      rs->rng->gaps
	(wrap (mkref(this), &snapshot_operation::fill_buf, rnode));
      check_fill();
    }
  }
}

void
snapshot_operation::fill_buf(sfsro_inode_reg *rnode, uint64 gs, uint64 gl)
{
  // fill buf with data starting from gs to gl or end of block, which
  // ever is smaller. commit to ranges and return.
  
  WARN(2) << "fill_buf: gs " << gs << " gl " << gl 
          << " size " << rnode->attr.size << "\n";
  if (gs >= rnode->attr.size) {
    WARN(2) << "snapshot_operation::fill_buf: gap starts after file ends.\n";
    return;
  }
  chordID cID;
  uint32 blocknum = gs / SNAPSHOT_BLOCKSIZE;
  if (blocknum < SFSRO_NDIR) {
    mpz_set_rawmag_be (&cID, rnode->direct[blocknum].base (), SFSRO_FHSIZE);
    _outstanding_fetches++;
    cfsrw_dhashclt::retrieve_sfsro_readonly
      (cID,
       wrap(mkref(this), &snapshot_operation::got_file_block, gs, gl));
  }
  else if ((blocknum-SFSRO_NDIR) < NFH) {
    mpz_set_rawmag_be (&cID, rnode->indirect.base (), SFSRO_FHSIZE);
    _outstanding_fetches++;
    cfsrw_dhashclt::retrieve_sfsro_readonly
      (cID, wrap (mkref(this), &snapshot_operation::got_sindir_block,
	          blocknum - SFSRO_NDIR, gs, gl));
  }
  else if ((blocknum-SFSRO_NDIR-NFH) < NFH*NFH) {
    mpz_set_rawmag_be (&cID, rnode->double_indirect.base (), SFSRO_FHSIZE);
    uint32 dblock = (blocknum-SFSRO_NDIR-NFH);
    _outstanding_fetches++;
    cfsrw_dhashclt::retrieve_sfsro_readonly
      (cID, wrap (mkref(this), &snapshot_operation::got_dindir_block,
		  dblock / NFH, dblock % NFH, gs, gl));
  }
  else if ((blocknum-SFSRO_NDIR-NFH-NFH*NFH) < NFH*NFH*NFH) {
    mpz_set_rawmag_be
      (&cID, rnode->triple_indirect.base (), SFSRO_FHSIZE);
    uint32 tblock = (blocknum - SFSRO_NDIR - NFH - NFH*NFH);
    uint32 toffst = tblock % (NFH*NFH);
    _outstanding_fetches++;
    cfsrw_dhashclt::retrieve_sfsro_readonly
      (cID, wrap (mkref(this), &snapshot_operation::got_tindir_block, 
  		  tblock / (NFH*NFH), toffst, gs, gl));
  }
  else
    warn << "fill_buf: read offset way too big! \n";
}

void 
snapshot_operation::got_sindir_block
(uint32 blocknum, uint64 gs, uint64 gl, ptr<sfsro_data> data)
{
  _outstanding_fetches--;
  if (!data) {
    _fetch_errors++;
    check_fill();
    return;    
  }
  if (data->type != SFSRO_INDIR) {
    _fetch_errors++;
    check_fill();
    return;
  }
  chordID cID;
  mpz_set_rawmag_be
    (&cID, (*(*data).indir).handles[blocknum].base (), SFSRO_FHSIZE);
  _outstanding_fetches++;
  cfsrw_dhashclt::retrieve_sfsro_readonly
    (cID,
     wrap(mkref(this), &snapshot_operation::got_file_block, gs, gl));
}

void 
snapshot_operation::got_dindir_block
(uint32 blocknum, uint32 blockoffst, uint64 gs, uint64 gl, ptr<sfsro_data> data)
{
  _outstanding_fetches--;
  if (!data) {
    _fetch_errors++;
    check_fill();
    return;    
  }
  if (data->type != SFSRO_INDIR) {
    _fetch_errors++;
    check_fill();
    return;
  }
  chordID cID;
  mpz_set_rawmag_be
    (&cID, (*(*data).indir).handles[blocknum].base (), SFSRO_FHSIZE);
  _outstanding_fetches++;
  cfsrw_dhashclt::retrieve_sfsro_readonly
    (cID, wrap (mkref(this), &snapshot_operation::got_sindir_block,
		blockoffst, gs, gl));
}

void 
snapshot_operation::got_tindir_block
(uint32 tblocknum, uint32 dblocknum, uint64 gs, uint64 gl, ptr<sfsro_data> data)
{
  _outstanding_fetches--;
  if (!data) {
    _fetch_errors++;
    check_fill();
    return;    
  }
  if (data->type != SFSRO_INDIR) {
    _fetch_errors++;
    check_fill();
    return;
  }
  chordID cID;
  mpz_set_rawmag_be
    (&cID, (*(*data).indir).handles[dblocknum].base (), SFSRO_FHSIZE);
  _outstanding_fetches++;
  cfsrw_dhashclt::retrieve_sfsro_readonly
    (cID, wrap (mkref(this), &snapshot_operation::got_dindir_block,
		dblocknum, dblocknum % NFH, gs, gl));
}

void 
snapshot_operation::got_file_block
(uint64 gs, uint64 gl, ptr<sfsro_data> data)
{
  _outstanding_fetches--;
  if (!data) {
    warn << "got_file_block, fetch error\n";
    _fetch_errors++;
    check_fill();
    return;
  }
  if (data->type != SFSRO_FILEBLK) {
    _fetch_errors++;
    check_fill();
    return;
  }
  //uint64 blocknum = gs / SNAPSHOT_BLOCKSIZE;
  //uint64 blockstart = blocknum * SNAPSHOT_BLOCKSIZE;
  //
  // XXX the following logic is horribly broken... we didn't load up
  // the entire "gl" number of bytes
  //
  uint64 start = (gs % SNAPSHOT_BLOCKSIZE);
  uint64 size = rs->offset + rs->count - gs;
  if (size > gl)
    size = gl;
  if (size > SNAPSHOT_BLOCKSIZE - start)
    size = SNAPSHOT_BLOCKSIZE - start;

  uint64 bi = gs - rs->offset;
  uint64 wi = start;

  memcpy (rs->buf + bi, (*(*data).data).base() + wi, size);

  rs->rng->add (gs, size);
  WARN(2) << "fill_buf: filled offset " << gs << " count " << size << "\n";

  check_fill();
}

void
snapshot_operation::check_fill()
{
  if (_outstanding_fetches == 0) {
    // don't need to add 1 to _count here, since we know the size of
    // the file in snapshot
    if (rs->rng->check(rs->offset, rs->count) || _fetch_errors == 0)
      done_fill_buf();
    else
      rcb (rs, NFS3ERR_IO);
  }
}

void 
snapshot_operation::done_fill_buf()
{
  uint64 c = rs->rng->end() - rs->offset;
  if(c > rs->count)
    c = rs->count;
  bool eof = ((rs->offset+c) == rs->rng->end());

  if (rs->valid_size) {
    if (rs->offset > rs->size) {
      c = 0;
      eof = true;
    }
    else if (rs->offset+c > rs->size) {
      c = rs->size-rs->offset;
      eof = true;
    }
  }
  WARN(2) << "done_fill_buf " << c << " bytes, eof? " << eof << "\n";
  rs->eof = eof;
  rs->rcount = c;
  rcb(rs, NFS3_OK);
}


