/*
 *
 * Copyright (C) 2002 Athicha Muthitacharoen (athicha@lcs.mit.edu)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

#ifndef CFSRW_SNAPSHOT_H
#define CFSRW_SNAPSHOT_H

#include "aiod.h"

#include "vec.h"
#include "ihash.h"
#include "dhash.h"
#include "dhash_prot.h"
#include "dbfe.h"

#include "cfsro.h"
#include "cache.h"
#include "record.h"
#include "view.h"
#include "logiter.h"
#include "serialize.h"
#include "vervec.h"
#include "ranges.h"
#include "scan.h"

#define CHORDID_LEN 	   20
#define SNAPSHOT_BLOCKSIZE 8192
#define NFH ((SNAPSHOT_BLOCKSIZE-100) / (sha1::hashsize*2))
#define FMAP_ENTRIES_PER_PAGE ((SNAPSHOT_BLOCKSIZE-100)/(2*CHORDID_LEN))

typedef struct snode {
  chordID inum;
  ptr<sfsro_data> ent;
  ihash_entry<snode> slink;
  snode (chordID, ptr<sfsro_data>);
  ~snode ();
} snode;

inline
snode::snode (chordID i, ptr<sfsro_data> e) : 
  inum (i), ent (e)
{
}

inline
snode::~snode ()
{
  ent = 0;
}

class snapshot;

typedef callback<void, chordID, ptr<snapshot> >::ptr cb_build;
typedef callback<void, snode*>::ptr cb_snode;

typedef ihash<chordID, snode, &snode::inum, &snode::slink>
  snode_hash_t;

class xin;
class snapshot : public virtual refcount {
private:
  friend class snapshot_operation;
  static aiod* a;

  chordID prev_link;
  chordID id; // content hash of this snapshot.
  chordID prev_snapshot;
  chordID vid;

  vervec _version;
  chordID rootdir;

  bool tree_build_started;
  bool tree_stored;
  int inode_stored;
  int outstanding_treeops;
  int treeop_failures;

  int outstanding_inserts;
  int insert_failures;
  int inserts;

  vec< ptr<log_record> > tole; // total order log entries
  snode_hash_t active_snodes; // intermediate fs tree in sfsro format
  
  cb_build bcb;
  void build_from_scratch ();
  void build_from_prev (ptr<snapshot> image);

  void update_tole (vervec, iterinfo *);
  void setinode (ptr<log_record>, sfsro_inode *);
  
  void do_link(chordID, str, ptr<log_record>);
  void do_link_got_dir(str, ptr<log_record>, snode*);
  void do_link_got_node(str, ptr<log_record>, snode*, snode*);
  void do_link_got_dirblk(str, ptr<log_record>, chordID,
                          snode*, snode*, snode*);
  void add_dirent(str, ptr<log_record>, sfsro_inode_reg*,
                  ptr<sfsro_data>, snode*);

  void do_write(ptr<log_record>);
  void do_write_got_node(ptr<log_record>, snode*);
  void do_symlink (ptr<log_record>, sfsro_inode_lnk *);
  void write_file (ptr<log_record>, sfsro_inode_reg *);
  
  void write_file_open_cb (ptr<log_record> e, sfsro_inode_reg *rnode,
                           ptr<aiofh> fh, int err);
  void write_file_write_cb (ptr<log_record> e, sfsro_inode_reg *rnode,
                            ptr<aiofh> fh, ptr<aiobuf> buf,
			    ssize_t sz, int err);
  void write_file_trunc_cb (ptr<log_record> e, sfsro_inode_reg *rnode,
                            ptr<aiofh> fh, int err);
  void write_file_close_cb (int err);

  void do_unlink(int, chordID, str, ptr<log_record>);
  void do_unlink_got_dir(int, str, ptr<log_record>, snode*);
  void do_unlink_got_dirblk (int, str, ptr<log_record>, int, snode*, snode*);
  void remove_dirent (int, str, ptr<log_record>, snode *,
                      sfsro_data *, uint32);
  void remove_dirent_got_fh (int, chordID, ptr<log_record>,
                             sfsro_inode_reg *, snode *);
  void done_unlink(int, ptr<log_record>, bool err = false);
  
  void update_attr (ptr<log_record>);
  void update_attr_got_node(ptr<log_record>, snode*);

  void do_rename (ptr<log_record>);

  snode *update_index (chordID, ptr<sfsro_data>);
  
  void fs_tree();
  void build_tree();
  void store_tree();

  void store (chordID);
  void store_new (snode *);
  chordID store_one_fmap_block (str);

  void insert_and_addtofmap (char *, size_t, chordID, ptr<sfsro_data> d = 0);
  void insert (chordID, char *, size_t, ptr<sfsro_data> d = 0);

  void insert_cb (bool, chordID, bool, chordID);
  void insert_snapshot ();
  void insert_fmapblock_cb (bool, chordID);
  void insert_snapinfo_cb (bool, chordID);
  void store_inode (chordID, sfsro_inode *);

  void store_file (chordID, sfsro_inode *);
  void store_directory (chordID, sfsro_inode *);
  void store_tind (vec<sfs_hash> *fhs, sfs_hash *ti_fh,
                   uint32 off, uint32& blocknum);
  void store_dind (vec<sfs_hash> *fhs, sfs_hash *di_fh,
                   uint32 off, uint32& blocknum);
  void store_sind (vec<sfs_hash> *fhs, sfs_hash *si_fh,
                   uint32 off, uint32& blocknum);
  void store_file_close (chordID inum, sfsro_inode *inode,
                         vec<sfs_hash>* fhs, ptr<aiofh> fd);
  void store_file_close_cb (int);
  void store_file_block (sfsro_inode *inode, uint32 blocknum,
                         const char *block, ssize_t size, sfs_hash *fh);
  void store_file_read_block (chordID inum, sfsro_inode *inode,
                              uint32 blocknum, vec<sfs_hash>* fhs,
			      ptr<aiofh> fd, ptr<aiobuf> buf,
			      ssize_t sz, int err);
  void store_file_open_cb (chordID inum, sfsro_inode *inode,
                           ptr<aiofh> fd, int err);

  void get_info_cb(chordID, ptr<dhash_block> blk);
  void get_fmap_cb(chordID, ptr<dhash_block> blk);
  void get_more_fmap_cb(chordID, ptr<dhash_block> blk);
  void del ();

  ptr<dbfe> fmap_db; // file map database
  ptr<dbfe> fmap_createdb ();

  void fmap_entry_load_cb (chordID sid, int err);
  void fmap_entry_insert_from_prev_cb (ptr<snapshot> image, int err);
  void fmap_entry_insert_on_store_cb (int err);
  void fmap_del_cb (int err);
  void lookup_fmap (chordID, cb_snode);
  void lookup_fmap_cb (chordID, cb_snode, ptr<dbrec>);
  void lookup_fmap_load_block_cb (cb_snode, chordID, ptr<sfsro_data>);
 
  bool fmap_load_done;
  int fmap_entry_inserts;
  int fmap_entry_insert_failures;
  int fmap_block_inserts;
  int fmap_block_insert_failures;

  void snapshot_failed ();

public:
  // load snapshot from DHash
  snapshot (chordID loc, cb_build cb);
  // creates new snapshot from an old image. if image is 0, create
  // from scratch
  snapshot (chordID vid, chordID prev, ptr<snapshot> image, cb_build cb);
  ~snapshot () { del (); }
  
  chordID ID () const { return id; };
  chordID prev () const { return prev_snapshot; }
  chordID view () const { return vid; }
  vervec version () const { return _version; }

private:
  void parse_info (str);
  void parse_fmap (bool, str, chordID);
  str serialize_info ();
  void serialize_and_store ();

};

class snapshot_operation : public virtual refcount {
private:
  ptr<snapshot> _ss;
  int _outstanding_fetches;
  int _fetch_errors;
  
  ptr<read_state> rs;
  cb_read rcb;

  ptr<dbfe> fmap_db() const { return _ss->fmap_db; }
  chordID rootdir() const { return _ss->rootdir; }
 
  void getroot_fmap_cb (callback<void, ptr<log_record> >::ref, ptr<dbrec>);
  void lookup_fmap_cb (chordID, str, cb_lookup, ptr<dbrec>);
  void lookup_fmap_cb3 (chordID, str, cb_lookup, ptr<dbrec>);
  void get_inode_fmap_cb (chordID, ptr<cattr>, cb_inode, ptr<dbrec>);
  void got_dir_inode_fmap (chordID, ptr<readdir_state>, cb_readdir, ptr<dbrec>);
  void read_fmap_cb (ptr<read_state>, cb_read, ptr<dbrec>);

  void find_dir_parent_fmap_cb (chordID dir, cb_chordid cb, ptr<dbrec> v);
  void find_dir_parent_cb (chordID dir, cb_chordid cb, ptr<sfsro_data> data);

  void getroot_cb (callback<void, ptr<log_record> >::ref, ptr<sfsro_data>);
  void lookup_cb1 (chordID, str, cb_lookup, ptr<sfsro_data>);
  void lookup_cb2 (chordID, sfsro_inode *, uint32, str, cb_lookup,
                   ptr<sfsro_data>); 
  void lookup_cb3 (chordID, str, cb_lookup, ptr<sfsro_data>); 
  void get_inode_cb (chordID, ptr<cattr>, cb_inode, ptr<sfsro_data>); 
  void got_dir_inode (chordID, ptr<readdir_state>, cb_readdir, ptr<sfsro_data>);
  void got_dir_block (chordID, sfsro_inode *, uint32, ptr<readdir_state>,
                      cb_readdir, ptr<sfsro_data>);
  
  void read_file_cb2 (ptr<sfsro_data> blk);
  void fill_buf (sfsro_inode_reg *, uint64 gs, uint64 gl);
  void got_sindir_block (uint32, uint64 gs, uint64 gl, ptr<sfsro_data>);
  void got_dindir_block (uint32, uint32, uint64 gs, uint64 gl, ptr<sfsro_data>);
  void got_tindir_block (uint32, uint32, uint64 gs, uint64 gl, ptr<sfsro_data>);
  void got_file_block (uint64 gs, uint64 gl, ptr<sfsro_data>);
  void done_fill_buf ();
  void check_fill ();

  void getroot (callback<void, ptr<log_record> >::ref);
  void find_inode (chordID, ptr<cattr>, cb_inode);
  void readdir (chordID, ptr<readdir_state>, cb_readdir);
  void lookup (chordID, str, cb_lookup cb);
  void find_dir_parent (chordID, cb_chordid);
  void read (ptr<read_state>, cb_read cb);

public:
  snapshot_operation (ptr<snapshot>);
  ~snapshot_operation() {}

  static void getroot(ptr<snapshot>, callback<void, ptr<log_record> >::ref);
  static void readdir(ptr<snapshot>, chordID, ptr<readdir_state>, cb_readdir);
  static void lookup(ptr<snapshot>, chordID, str, cb_lookup);
  static void find_dir_parent(ptr<snapshot>, chordID, cb_chordid);
  static void find_inode(ptr<snapshot>, chordID, ptr<cattr>, cb_inode);
  static void read(ptr<snapshot>, ptr<read_state>, cb_read cb);
};

void init_attr (sfsro_inode_reg *);
void init_attr (sfsro_inode_lnk *);
void set_attr (sfsro_inode_reg *, cattr *);
void set_attr (sfsro_inode_lnk *, cattr *);
void combine_attr (ptr<cattr> fa, ptr<sfsro_data> sdata);
void get_random_ID (chordID *, uint);
str get_dirname (nfspath3);
str get_filename (nfspath3);
ptr<log_record> set_log_record
  (bool, ptr<sfsro_data>, chordID dir=0, str name=NULL);
bool dup_entry (rpc_ptr<entry3> *, sfsro_dirent *);
  
int dbcompare (ref<dbrec> a, ref<dbrec> b);
ref<dbrec> id2dbrec(chordID id);
chordID dbrec2id (ptr<dbrec> r);

inline chordID
content_hash (sfs_hash *ch, char *buf, size_t len)
{
  bzero (ch->base(), ch->size());
  sha1_hash (ch->base(), buf, len);
  chordID n;
  mpz_set_rawmag_be(&n, ch->base(), sha1::hashsize);
  return n;
}

inline str
id2str (chordID id)
{
  strbuf w;
  sid (w, id);
  return armor64A(w);
}

inline void 
str2chordID (chordID *cID, const char *cstr, ssize_t len)
{
  str dearm = dearmor64A (cstr, len);
  mpz_set_rawmag_be(cID, dearm.cstr(), dearm.len());
}

inline void 
set_cfsrwattr (cattr *ca, sfsro_inode_lnk *lnode)
{
  ca->valid_size = (lnode->attr.size == 666 ? 0 : 1);
  ca->size = lnode->attr.size;
  ca->valid_mode = (lnode->attr.mode == 666 ? 0 : 1);
  ca->size = lnode->attr.mode;
  ca->valid_uid = (lnode->attr.uid == 666 ? 0 : 1);
  ca->size = lnode->attr.uid;
  ca->valid_gid = (lnode->attr.gid == 666 ? 0 : 1);
  ca->size = lnode->attr.gid;
  ca->valid_rdev = (lnode->attr.rdev.major == 666 ? 0 : 1);
  ca->rdev.major = lnode->attr.rdev.major;  
  ca->rdev.minor = lnode->attr.rdev.minor;  
}

inline void 
update_mtime (sfsro_inode_reg *inode, nfstime3 ts)
{
  inode->mtime = ts;
}

inline void 
update_ctime (sfsro_inode_reg *inode, nfstime3 ts)
{
  inode->ctime = ts;
}

inline void 
update_mtime (sfsro_inode_lnk *inode, nfstime3 ts)
{
  inode->mtime = ts;
}

inline void 
update_ctime (sfsro_inode_lnk *inode, nfstime3 ts)
{
  inode->ctime = ts;
}

#endif /* CFSRW_SNAPSHOT_H */

