/*
 * Copyright (C) 2002  Robert Morris (rtm@lcs.mit.edu)
 *                     Benjie Chen (benjie@lcs.mit.edu)
 *                     Thomer M. Gil (thomer@lcs.mit.edu)
 *   		       Massachusetts Institute of Technology
 * 
 *  Permission is hereby granted, free of charge, to any person obtaining
 *  a copy of this software and associated documentation files (the
 *  "Software"), to deal in the Software without restriction, including
 *  without limitation the rights to use, copy, modify, merge, publish,
 *  distribute, sublicense, and/or sell copies of the Software, and to
 *  permit persons to whom the Software is furnished to do so, subject to
 *  the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 *  LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 *  OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 *  WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

#include "sfsmisc.h"
#include "dhash_prot.h"
#include "dhash.h"
#include "crypt.h"
#include "parseopt.h"
#include <sys/time.h>
#include "cache.h"
#include "log.h"

dhashclient *dh;

bool verbose_mode = false;

void
get_cb(chordID id, callback<void, int, str>::ref cb, ptr<dhash_block> blk)
{
  if (!blk)
    cb(0, str(""));
  else
    cb(1, str(blk->data, blk->len));
}

void
simple_get(bool ishead, chordID id, callback<void, int, str>::ref cb)
{
  if (ishead) {
    dh->retrieve(id, wrap(get_cb, id, cb));
    return;
  }
  dh->retrieve(id, wrap(get_cb, id, cb));
}

void
get_record_cb(callback<void, ptr<log_record> >::ref cb, int ok, str data)
{
  if (ok) {
    ptr<log_record> e = log_record::parse(data);
    cb(e);
  } 
  else
    cb((ptr<log_record>) 0);
}

void
get_head_record(chordID id, callback<void, ptr<log_record> >::ref cb)
{
  simple_get(true, id, wrap(get_record_cb, cb));
}

void
get_record(chordID id, callback<void, ptr<log_record> >::ref cb)
{
  simple_get(false, id, wrap(get_record_cb, cb));
}

chordID
make_block_id(const void *data, int size) 
{
  long *rd = (long *)data;
  char id[sha1::hashsize];
  sha1_hash (id, rd, size);
  
  chordID ID;
  mpz_set_rawmag_be (&ID, id, sizeof (id));
  return ID;
}


void
store_block_cb(callback<void, chordID, int>::ref cb, bool failure, chordID key)
{
  if (failure) {
    warn << "store_block_cb failed\n";
    exit(0);
  }
  cb(key, 1);
}

void
main_cb(bool ishead, int *n, chordID id, int ok)
{
  if (ishead) {
    char raw[20];
    mpz_get_rawmag_be(raw, sizeof(raw), &id);
    str name = armor64A(raw, sizeof(raw));
    strbuf out;
    out << name << "\n";
    out.tosuio ()->output (1);
  }

  *n -= 1;
  if(*n == 0)
    exit(0);
}


chordID
store1(const str &data, callback<void, chordID, int>::ref cb)
{
  chordID id = make_block_id(data.cstr(), data.len());
  cfsrw_dhashclt::insert_ch
    (data.cstr(), data.len(), wrap(store_block_cb, cb));
  return id;
}

void
store_head(const str &data, callback<void, chordID, int>::ref cb,
           ptr <sfspriv> sk)
{
  assert(sk);
  cfsrw_dhashclt::insert_pk
    (sk, data.cstr(), data.len(), wrap(store_block_cb, cb));
}


void
make_new(ptr <sfspriv> sk)
{
  static int nb = 0;
  log_record *e;
  chordID lid;
  uint seq = 0;
 
  strbuf b;
  sk->export_pubkey (b, false);
  str pk_raw = b;
  chordID head = compute_hash(pk_raw.cstr(), pk_raw.len());

  nb++;
  vervec v;
  e = New end_log_record();
  e->set_version(v);
  e->set_headID(head);
  e->set_seqnum(seq);
  lid = store1(e->serialize(), wrap(main_cb, false, &nb));
  seq++;

  cattr attr;
  attr.valid_mode = 1;
  attr.mode = 0755;
  attr.valid_size = 1;
  attr.size = 0;
  attr.valid_uid = attr.valid_gid = 1;
  attr.uid = attr.gid = 0;

  nb++;
  e = New inode_log_record(inode_log_record::T_DIR, attr);
  v.clear();
  v.add(head,lid,seq-1);
  e->set_version(v);
  e->set_prevID(lid);
  e->set_headID(head);
  e->set_seqnum(seq);
  lid = store1(e->serialize(), wrap(main_cb, false, &nb));
  chordID root_inum = ((inode_log_record*)e)->inum();
  seq++;

  nb++;
  attr.mode = 0644;
  e = New inode_log_record(inode_log_record::T_FILE, attr);
  v.clear();
  v.add(head,lid,seq-1);
  e->set_version(v);
  e->set_prevID(lid);
  e->set_headID(head);
  e->set_seqnum(seq);
  lid = store1(e->serialize(), wrap(main_cb, false, &nb));
  seq++;

  nb++;
  e = New link_log_record(((inode_log_record*)e)->inum(),
			  inode_log_record::T_FILE, root_inum, "TEST");

  v.clear();
  v.add(head,lid,seq-1);
  e->set_version(v);
  e->set_prevID(lid);
  e->set_headID(head);
  e->set_seqnum(seq);
  lid = store1(e->serialize(), wrap(main_cb, false, &nb));
  seq++;

  nb++;
  e = New head_log_record(root_inum);
  v.clear();
  v.add(head,lid,seq-1);
  e->set_version(v);
  e->set_prevID(lid);
  e->set_headID(head);
  e->set_seqnum(seq);
  store_head(e->serialize(), wrap(main_cb, true, &nb), sk);

  // warn << "root_inum " << root_inum << "\n";
}

void
new_attach_cb_1(chordID old, ptr<sfspriv> sk, ptr<log_record> old_e)
{
  strbuf b;
  sk->export_pubkey (b, false);
  str pk_raw = b;
  chordID head = compute_hash (pk_raw.cstr (), pk_raw.len ());

  if(!old_e)
    return;
  
  static int nb = 0;
  log_record *e;
  chordID lid;
  
  head_log_record *re = old_e->template record<head_log_record>();
  chordID root_inum = re->root();

  nb++;
  e = New end_log_record();
  vervec v;
  e->set_version(v);
  e->set_headID(head);
  e->set_seqnum(0);
  lid = store1(e->serialize(), wrap(main_cb, false, &nb));
  
  nb++;
  e = New head_log_record(root_inum);
  v.clear();
  v.add(head,lid,0);
  e->set_version(v);
  e->set_prevID(lid);
  e->set_headID(head);
  e->set_seqnum(1);
  store_head(e->serialize(), wrap(main_cb, true, &nb), sk);
  
  // warn << "root_inum " << root_inum << "\n";
}

void
new_attach(chordID old, ptr <sfspriv> sk)
{
  get_head_record(old, wrap(new_attach_cb_1, old, sk));
}

static void
usage ()
{
  warnx << "usage: " << progname << " [-s <keyfile>] [id]\n";
  warnx << "  " << progname
        << " creates a new head of log. The head of log block is\n"
        << "  stored under the public key specified by the CFSRW_KEYFILE\n"
        << "  environment variable or the -s argument. If a chord ID is\n"
        << "  provided, it is interpreted as the hash of another user's\n"
        << "  public key, and the new log will be part of the same fs.\n"
        << "  Otherwise, a new fs is created\n";
  exit (1);
}

int
main (int argc, char **argv)
{
  char *sk_file = getenv ("CFSRW_KEYFILE");

  setprogname (argv[0]);

  int ch;
  while ((ch = getopt (argc, argv, "s:")) != -1)
    switch (ch) {
    case 's':
      sk_file = optarg;
      break;
    case 'v':
      verbose_mode = true;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if ( (argc > 2) || !sk_file )
    usage ();

  ptr <sfspriv> sk;
  str key = file2wstr (sk_file);
  if (!key) {
    warn << sk_file << ": " << strerror (errno) << "\n";
    fatal ("errors!\n");
  }
  else if (!(sk = sfscrypt.alloc_priv (key, SFS_SIGN))) {
    warn << "could not decode " << sk_file << "\n";
    warn << key << "\n";
    fatal ("errors!\n");
  }

  cfsrw_dhashclt::init();
  sfsconst_init ();
  random_init();

  char *sn = getenv("LSD_SOCKET");
  if(sn == 0){
    warnx << "viewctl: LSD_SOCKET is not set\n";
    exit(1);
  }

  dh = New dhashclient(sn);

  if(argc == 1){
    str dearm = dearmor64A(argv[0]);
    chordID id;
    mpz_set_rawmag_be(&id, dearm.cstr(), dearm.len());
    new_attach(id, sk);
  } else
    make_new(sk);
  
  amain();
}

