/*
 * Copyright (C) 2002  Thomer M. Gil (thomer@lcs.mit.edu)
 *   		       Massachusetts Institute of Technology
 * 
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.
 */

#include "sfsmisc.h"
#include "dhash.h"
#include "dhash_prot.h"
#include "crypt.h"
#include "parseopt.h"
#include "serialize.h"
#include "vervec.h"
#include "cache.h"
#include "log.h"
#include "logiter.h"
#include "rxx.h"
#include "qhash.h"
#include <sys/time.h>


#define REVERSE "my @lines; while(<>) { push @lines, $_; } my $l; while(($l = pop @lines)) { print \"$l\"; }"

typedef struct ivyck_ent {
  str inum;
  ihash_entry<ivyck_ent> slink;

  ivyck_ent (str);
  ~ivyck_ent ();
} ivyck_ent;

ivyck_ent::ivyck_ent(str i) {
  inum = i;
}

// HACK HACK HACK. STORE VERSION VECTORS
typedef struct ivyck_vecent {
  str logid;
  vervec v;
  ihash_entry<ivyck_vecent> slink;

  ivyck_vecent (str, vervec);
  ~ivyck_vecent ();
} ivyck_vecent;

ivyck_vecent::ivyck_vecent(str i, vervec v) {
  logid = i;
  this->v = v;
}

int tmplog = 0;
FILE *ftmplog = 0;
chordID loghead;
chordID head;
uint seq = 0;


// private key to sign new log head with
ptr <sfspriv> sk;

// hash to track inodes
typedef ihash<str, ivyck_ent, &ivyck_ent::inum, &ivyck_ent::slink> ivyck_t;
typedef ihash<str, ivyck_vecent, &ivyck_vecent::logid, &ivyck_vecent::slink> ivyckvec_t;

ivyck_t ivyckinums; // track inodes
ivyck_t ivycklinks; // track links
ivyckvec_t ivyckvecs; // version vectors in logs

// the view to travers
static ptr<view> _view;
static str root_inum;
static chordID chord_root_inum;

static bool verbose = false;
static int _work = 0;
static int _nstore = 0;
static int _hstore = 0;

// used for conflict detection
struct dirop {
  chordID dir;
  str name;
};
static qhash<chordID,chordID> _inode_ops;
static qhash<bigint, chordID> _dir_ops;
static qhash<bigint, dirop> _dirop_mapping;
static vec<chordID> _conflict_inodes;
static vec<bigint> _conflict_dirops;

void
print_vervec(unsigned n, vervec &v, strbuf &out)
{
  for (unsigned i=0; i<v.size(); i++) {
    strbuf hh;
    sid(hh, v.head(i));
    str h = armor64A(hh);
    strbuf ll;
    sid(ll, v.latest(i));
    str l = armor64A(ll);
    for (unsigned j=0; j<n; j++) out << " ";
    out << " verh " << h << "\n";
    for (unsigned j=0; j<n; j++) out << " ";
    out << " verl " << l << "\n";
    for (unsigned j=0; j<n; j++) out << " ";
    out << " vers " << v.seqnum(i) << "\n";
  }
}

void
main_cb(bool ishead, int *n, chordID id, int ok)
{
  if(!ok)
    fatal("Not OK!\n");

  warn << "main_cb\n";

  char raw[20];
  mpz_get_rawmag_be(raw, sizeof(raw), &id);
  str name = armor64A(raw, sizeof(raw));
  strbuf out;
  out << "stored: " << name << "\n";
  out.tosuio ()->output (1);

  (*n)--;
  if(!_hstore && !_nstore) {
    warn << "I think I'm done\n";
    exit(0);
  }
}



void
store_block_cb(callback<void, chordID, int>::ref cb, bool failure, chordID key)
{
  if (failure)
    fatal("store_block_cb failed\n");

  warn << "store_block_cb\n";
  cb(key, 1);
}


chordID
make_block_id(const void *data, int size) 
{
  long *rd = (long *)data;
  char id[sha1::hashsize];
  sha1_hash (id, rd, size);
  
  chordID ID;
  mpz_set_rawmag_be (&ID, id, sizeof (id));
  return ID;
}



chordID
store1(const str &data, callback<void, chordID, int>::ref cb)
{
  chordID id = make_block_id(data.cstr(), data.len());
  cfsrw_dhashclt::insert_ch
    (data.cstr(), data.len(), wrap(store_block_cb, cb));

  char raw[20];
  mpz_get_rawmag_be(raw, sizeof(raw), &id);
  str name = armor64A(raw, sizeof(raw));
  warn << "store1 returning " << name << "\n";
  return id;
}

void
store_head(const str &data, callback<void, chordID, int>::ref cb,
           ptr <sfspriv> sk)
{
  assert(sk);
  cfsrw_dhashclt::insert_pk
    (sk, data.cstr(), data.len(), wrap(store_block_cb, cb));
}


void
newhead()
{
  log_record *e = New head_log_record(chord_root_inum);
  vervec v;
  v.clear();
  e->set_version(v);
  e->set_prevID(head);
  e->set_headID(loghead);
  e->set_seqnum(seq++);
  store_head(e->serialize(), wrap(main_cb, true, &_nstore), sk);
  _nstore++;
}



void
checklinked(str inum, str logid, int type)
{
  ivyck_ent *ent = ivycklinks[inum];

  // root dir is not linked
  if(ent || inum == root_inum)
    return;

  chordID cid;
  str dearm = dearmor64A(inum);
  mpz_set_rawmag_be(&cid, dearm.cstr(), dearm.len());

  // XXX: name? parentdir? link to root directoru
  log_record *e = New link_log_record
    (cid, type, chord_root_inum, inum);

  vervec v = ivyckvecs[logid]->v;
  e->set_version(v);
  e->set_prevID(head);
  e->set_headID(loghead);
  e->set_seqnum(seq++);
  head = store1(e->serialize(), wrap(main_cb, false, &_nstore));
  _nstore++;

  // is now linked
  ent = New ivyck_ent(inum);
  ivycklinks.insert(ent);

  newhead();
}


void
checkinum(str inum, str logid, int type)
{
  ivyck_ent *ent = ivyckinums[inum];
  if(ent)
    return;

  warn << "missing inumber " << inum << "\n";

  chordID cid;
  str dearm = dearmor64A(inum);
  mpz_set_rawmag_be(&cid, dearm.cstr(), dearm.len());


  /*
   * Create T_INODE
   */
  log_record *e = New inode_log_record(type, cid);
  vervec v = ivyckvecs[logid]->v;
  e->set_version(v);
  e->set_prevID(head);
  e->set_headID(loghead);
  e->set_seqnum(seq++);
  head = store1(e->serialize(), wrap(main_cb, false, &_nstore));
  _nstore++;

  // is now inode
  ent = New ivyck_ent(inum);
  ivyckinums.insert(ent);

  newhead();
}


void
ivyck()
{
  // HACK. reverse the log
  strbuf s;
  s << str("perl -e '") << str(REVERSE) << str("' < tmplog.tmp > tmplog2.tmp; mv tmplog2.tmp tmplog.tmp");
  system((const char *)(str)s);

  // open logfile
  ftmplog = fopen("tmplog.tmp", "r");
  if(!ftmplog)
    fatal("Couldn't open tmpfile\n");

  // read head
  char tmp[512];
  fscanf(ftmplog, "%s\n", tmp);
  str head = tmp;

  warnx << "head = " << head << "\n";

  // Now read rest of file
  int seq, type;
  str logid;
  while(!feof(ftmplog)) {
    // read sequence number, inumber, and log record type
    if(fscanf(ftmplog, "%d\t%s\t%d", &seq, tmp, &type) < 3) {
      warn << "Expected seq, inum, and type\n";
      return;
    }
    logid = tmp;
    // warn << "seq = " << seq << " logid = " << logid << " type = " << type << "\n";

    // first log record
    if(type == log_record::T_END)
      continue;

    // last
    if(type == log_record::T_HEAD)
      fatal("done\n");


    // read inumber
    if(fscanf(ftmplog, "\t%s", tmp) < 1)
      fatal("expected inum\n");

    str inum = tmp;
    str dir, olddir;

    // warn << "look at inum = " << inum << "\n";
    switch(type) {
      
      // mark inode as present
      case log_record::T_INODE:
        {
          // read itype
          int itype;
          if(fscanf(ftmplog, "\t%d", &itype) < 1)
            fatal("expected itype\n");
          ivyck_ent *ent = New ivyck_ent(inum);
          ivyckinums.insert(ent);
        }
        break;

      case log_record::T_RENAME:
        {
          char dirc[512], olddirc[512];
          if(fscanf(ftmplog, "\t%s\t%s", olddirc, dirc) < 2)
            fatal("expected olddir, dir\n");
          dir = dirc;
          olddir = olddirc;

          checkinum(olddir, logid, inode_log_record::T_DIR);
          checklinked(olddir, logid, inode_log_record::T_DIR);
          checkinum(dir, logid, inode_log_record::T_DIR);
          checklinked(dir, logid, inode_log_record::T_DIR);
        }
        break;

      case log_record::T_LINK:
        {
          char dirc[512];
          if(fscanf(ftmplog, "\t%s", dirc) < 1)
            fatal("expected dir");
          dir = dirc;

          // XXX: is it a file, directory, or symlink?????
          checkinum(inum, logid, inode_log_record::T_FILE); // XXX
          checkinum(dir, logid, inode_log_record::T_DIR);
          checklinked(dir, logid, inode_log_record::T_DIR);

          ivyck_ent *ent = New ivyck_ent(inum);
          ivycklinks.insert(ent);
        }
        break;

      case log_record::T_UNLINK:
        {
          char dirc[512];
          if(fscanf(ftmplog, "\t%s", dirc) < 1)
            fatal("expected inum and itype");
          dir = dirc;

          // XXX: itype!
          checkinum(inum, logid, inode_log_record::T_FILE);
          // XXX: itype!
          // XXX: does it make sense to fix on unlink?
          checklinked(inum, logid, inode_log_record::T_FILE);
          checkinum(dir, logid, inode_log_record::T_DIR);
          checklinked(dir, logid, inode_log_record::T_DIR);

          ivyck_ent *ent = ivycklinks[inum];
          ivycklinks.remove(ent);
        }
        break;

      case log_record::T_WRITE:
        {
          // XXX: writes only to file, I guess.
          checkinum(inum, logid, inode_log_record::T_FILE);
          checklinked(inum, logid, inode_log_record::T_FILE);
        }
        break;
      case log_record::T_ATTR:
        {
          // XXX: itype!
          checkinum(inum, logid, inode_log_record::T_FILE);

          // XXX: itype!
          checklinked(inum, logid, inode_log_record::T_FILE);
        }
        break;
      default:
        break;
    }
  }
}

void
dump_cb(iterinfo *i)
{
  strbuf out;
  strbuf outf;

  if (i->finished) {
    _work--;

    strbuf r;
    sid(r, loghead);
    str h = armor64A(r);
    if (_work == 0) {
      outf << h << "\n";
      outf.tosuio ()->output (tmplog);
      close(tmplog);
      ivyck();
      return;
    }
  }
        
  recordp e = i->record;
  assert(e);
  chordID hid = e->headID();
  chordID lid = e->selfID();

  char ts[32];
  nfstime3 t = e->ts();
  ctime_r(&((time_t) t.seconds), ts);
  ts[strlen(ts)-1] = '\0';

  strbuf r;
  sid(r, hid);
  str head = armor64A(r);

  strbuf w;
  sid(w, lid);
  str where = armor64A(w);

  outf << e->seqnum() << "\t" << where << "\t" << e->type();
  out << ts << "   seq " << e->seqnum() << "\n                        "
            << "   log " << where << "\n                        "
            << "  head " <<  head << "\n";

  vervec v = e->version();
  ivyck_vecent *ent = New ivyck_vecent(where, v);
  ivyckvecs.insert(ent);

  print_vervec(25, v, out);
  for(unsigned j=0; j<25; j++) out << " ";

  switch(e->type()) {
    case log_record::T_INODE:
      {
        ptr<inode_log_record> l = *reinterpret_cast<ptr<inode_log_record>*>(&e);
        strbuf b;
        sid(b, l->inum());
        outf << "\t" << armor64A(b) << "\t" << l->itype();
        switch(l->itype()) {
          case inode_log_record::T_FILE:
            out << " file ";
            break;
          case inode_log_record::T_DIR:
            out << "  dir ";
            break;
          case inode_log_record::T_SYMLINK:
            out << " syml ";
            break;
          default:
            break;
        }
        out << armor64A(b) << "\n";
      }
      break;
    case log_record::T_INTEND:
      {
        out << "intend ";
        intend_log_record *ie = e->template record<intend_log_record>();
        strbuf x;
        sid(x, ie->verf());
        strbuf y;
        sid(y, ie->dir());
        out << ie->name()  << "\n                            in "
            << armor64A(y) << "\n                          verf "
            << armor64A(x) << "\n                        cancel "
            << ie->canceled() << "\n";
      }
      break;
    case log_record::T_RENAME:
      {
        out << "rname ";
        ptr<rename_log_record> l =
	  *reinterpret_cast<ptr<rename_log_record>*>(&e);
        strbuf x;
        sid(x, l->inum());
        out << armor64A(x) << "\n                            in ";
        strbuf y;
        sid(y, l->dir());
        out << armor64A(y) << "\n                            as "
            << l->name() << "\n";
        out << "                           was " << l->oldname() << "\n";

        strbuf d;
        sid(d, l->olddir());
        outf << "\t" << armor64A(x) << "\t" << "\t" << armor64A(d) << "\t" << armor64A(y);
      }
      break;
    case log_record::T_LINK:
      {
        out << " link ";
        ptr<link_log_record> l = *reinterpret_cast<ptr<link_log_record>*>(&e);
        strbuf x;
        sid(x, l->inum());
        out << armor64A(x) << "\n                            in ";
        strbuf y;
        sid(y, l->dir());
        out << armor64A(y) << "\n                            as "
            << l->name() << "\n";
        outf << "\t" << armor64A(x) << "\t" << armor64A(y);
      }
      break;
    case log_record::T_UNLINK:
      {
        out << "ulink ";
        ptr<unlink_log_record> l =
	  *reinterpret_cast<ptr<unlink_log_record>*>(&e);
        strbuf x;
        sid(x, l->inum());
        out << armor64A(x) << "\n                            in ";
        strbuf y;
        sid(y, l->dir());
        out << armor64A(y) << "\n                            as "
            << l->name() << "\n";
        outf << "\t" << armor64A(x) << "\t" << armor64A(y);
      }
      break;
    case log_record::T_WRITE:
      {
        out << "write ";
        ptr<write_log_record> l = *reinterpret_cast<ptr<write_log_record>*>(&e);
        strbuf x;
        sid(x, l->inum());
        out << armor64A(x) << "\n                            at "
            << l->offset() << ", " << l->count() << " bytes\n";
        outf << "\t" << armor64A(x);
      }
      break;
    case log_record::T_ATTR:
      {
        out << " attr ";
        ptr<attr_log_record> l = *reinterpret_cast<ptr<attr_log_record>*>(&e);
        strbuf x;
        sid(x, l->inum());
        out << armor64A(x) << "\n                          size ";
        outf << "\t" << armor64A(x);
        if (l->attr()->valid_size)
          out << l->attr()->size << " bytes\n";
        else
          out << " not set\n";
      }
      break;
    case log_record::T_END:
    case log_record::T_HEAD:
    default:
      break;
  }

  // if (e->type() != log_record::T_END && e->type() != log_record::T_HEAD) {
    out.tosuio ()->output (1);
    outf << "\n";
    outf.tosuio ()->output (tmplog);
  // }
}

void
got_view(chordID vid, ptr<view> ok)
{
  if (!ok) {
    warn << "cannot load view\n";
    exit(1);
  }

  char raw[20];
  str xx;
  strbuf out;
  chordID n;
  
  n = _view->root();
  mpz_get_rawmag_be(raw, sizeof(raw), &n);
  xx = armor64A(raw, sizeof(raw));
  out << "root inum: " << xx << "\n";
  root_inum = xx;
  chord_root_inum = n;

  _work++;
  New refcounted<logiter>(wrap(&dump_cb), vid);
}

void
usage()
{
  /*
  warnx << "usage: \n";
  warnx << "  " << progname << " [-v] viewid\n"
        << "    prints out information on the given view. if -v is specified,\n"
        << "    iterates and prints all the log entries as well\n";
  warnx << "  " << progname << " -p viewid\n"
        << "    prints out a public key hash in the view\n";
  warnx << "  " << progname << " -c key\n"
        << "    creates a new view with the log stored at the given key\n";
  warnx << "  " << progname << " -a key[,latest] viewid\n"
        << "    adds the log stored at the given key to an existing view\n";
  warnx << "  " << progname << " -d time viewid\n"
        << "    creates a new view that includes all states in the given view\n"
        << "    before specified time. e.g., \"Apr 1 16:18:27 2002\"\n";
  warnx << "  " << progname << " [-v] -x viewid\n"
        << "    scan file system for conflicting operations.\n";
  warnx << "  " << progname << " [-v] -s viewid\n"
        << "    scan file system for conflicting and concurrent operations.\n";
  warnx << "viewid can be the hash of a view's chord ID, or the mount point\n";
  */
  exit(1);
}

int
main (int argc, char **argv)
{
  setprogname (argv[0]);
  char *sk_file = NULL;

  int ch;
  while ((ch = getopt (argc, argv, "vs:")) != -1)
    switch (ch) {
    case 'v':
      verbose = true;
      break;
    case 's':
      sk_file = optarg;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if(!sk_file)
    fatal("create a new idendity file and try again with -s flag");

  // read alternate keyfile
  ptr <sfspriv> sk;
  str key = file2wstr (sk_file);
  if (!key) {
    warn << sk_file << ": " << strerror (errno) << "\n";
    fatal ("errors!\n");
  }
  else if (!(sk = sfscrypt.alloc_priv (key, SFS_SIGN))) {
    warn << "could not decode " << sk_file << "\n";
    warn << key << "\n";
    fatal ("errors!\n");
  }

  // compute loghead
  strbuf b;
  sk->export_pubkey (b, false);
  str pk_raw = b;
  head = compute_hash(pk_raw.cstr(), pk_raw.len());
  loghead = head;

  char *sn = getenv("LSD_SOCKET");
  if(sn == 0)
    fatal("viewctl: LSD_SOCKET is not set\n");

  if((tmplog = open("tmplog.tmp", O_RDONLY | O_WRONLY | O_CREAT | O_TRUNC)) == 0) {
    warnx << "Couldn't open tmpfile\n";
    return -1;
  }

  cfsrw_dhashclt::init();
  sfsconst_init ();
  random_init();

  // store head
  log_record *e = New end_log_record();
  vervec v;
  e->set_version(v);
  e->set_headID(loghead);
  e->set_seqnum(seq++);
  head = store1(e->serialize(), wrap(main_cb, false, &_hstore));
  _hstore++;

  chordID vid;
  if (argc == 1) {
    char sym[1024];
    str dearm;
    rxx path_regexp ("chord:([0-9a-zA-Z+-]*)[/]*$");
    int r = readlink(argv[0], sym, 1024);
    if (r > 0) sym[r] = '\0';
    if (path_regexp.search (argv[0]) || (r > 0 && path_regexp.search(sym)))
      dearm = dearmor64A(path_regexp[1]);
    else
      dearm = dearmor64A(argv[0]);
    if (dearm) 
      mpz_set_rawmag_be(&vid, dearm.cstr(), dearm.len());
    else
      fatal("viewctl: key hash invalid\n");
  }

  _view = New refcounted<view> (wrap(&got_view, vid), vid);
  amain();
}
