/*
 * Copyright (C) 2002  Benjie Chen (benjie@lcs.mit.edu)
 *   		       Massachusetts Institute of Technology
 * 
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.
 */

#include "sfsmisc.h"
#include "dhash.h"
#include "dhash_prot.h"
#include "crypt.h"
#include "parseopt.h"
#include "serialize.h"
#include "vervec.h"
#include "cache.h"
#include "log.h"
#include "logiter.h"
#include "rxx.h"
#include "qhash.h"
#include <sys/time.h>

static ptr<view> _view;

static time_t _before;
static bool verbose = false;
static int _work = 0;

// used for conflict detection
static nfstime3 _ctime;
static vervec _base;
static vervec _conflict_v1;
static vervec _conflict_v2;
static bool _resolve_started = false;
struct dirop {
  chordID dir;
  str name;
};
static qhash<chordID,chordID> _inode_ops;
static qhash<bigint, chordID> _dir_ops;
static qhash<bigint, dirop> _dirop_mapping;
static vec<chordID> _conflict_inodes;
static vec<bigint> _conflict_dirops;

struct split {
  nfstime3 t;
  chordID base;
  chordID ver1;
  chordID ver2;
};
static vec<struct split> _splits;

void
store_cb(bool print, bool failure, chordID id)
{
  _work--;
  if (failure)
    warn << "failed to store view\n";
  else if (print) {
    char raw[20];
    mpz_get_rawmag_be(raw, sizeof(raw), &id);
    str xx = armor64A(raw, sizeof(raw));
    strbuf out;
    out << "chord:" << xx << "\n";
    out.tosuio ()->output (1);
  }
  if (_work == 0)
    exit(0);
}

void
print_vervec(unsigned n, vervec &v, strbuf &out)
{
  for (unsigned i=0; i<v.size(); i++) {
    strbuf hh;
    sid(hh, v.head(i));
    str h = armor64A(hh);
    strbuf ll;
    sid(ll, v.latest(i));
    str l = armor64A(ll);
    for (unsigned j=0; j<n; j++) out << " ";
    out << " verh " << h << "\n";
    for (unsigned j=0; j<n; j++) out << " ";
    out << " verl " << l << "\n";
    for (unsigned j=0; j<n; j++) out << " ";
    out << " vers " << v.seqnum(i) << "\n";
  }
}

ref<view>
construct_limited_view(chordID inum, vervec v1, vervec v2)
{
  // compare v1 against v2. if v1[i] >= v2[i] or log i does not exist in v2,
  // don't limit i in the new view.  if v1[i] <  v2[i], limit i in the new
  // view.

  ref<view> v = New refcounted<view>();
  v->root(inum);
  for (unsigned i=0; i<v1.size(); i++) {
    bool limit = false;
    for (unsigned j=0; j<v2.size(); j++) {
      if (v1.head(i) == v2.head(j)) {
        if (v1.seqnum(i) < v2.seqnum(i))
	  limit = true;
	break;
      }
    }
    if (limit)
      v->add_log(v1.head(i), v1.latest(i));
    else
      v->add_log(v1.head(i), v1.head(i));
  }
  return v;
}

void
split_conflict(vervec v1, vervec v2, strbuf &out)
{
  // construct base view: get the smaller seqnum from each log
  ref<view> view_b = New refcounted<view>();
  view_b->root(_view->root());
  for (unsigned i=0; i<_view->nlogs(); i++)
    view_b->add_log(_view->head(i), _view->latest(i));
  vervec b = vervec::merge_smaller(v1, v2);
  for (unsigned i=0; i<b.size(); i++)
    view_b->add_log(b.head(i), b.latest(i));

  // now generate the other two views
  ref<view> view_1 = construct_limited_view(_view->root(), v1, v2);
  ref<view> view_2 = construct_limited_view(_view->root(), v2, v1);

  _work += 3;
  chordID base = view_b->store(wrap(store_cb, false));
  chordID ver1 = view_1->store(wrap(store_cb, false));
  chordID ver2 = view_2->store(wrap(store_cb, false));

  char raw[20];
  str xx;
  mpz_get_rawmag_be(raw, sizeof(raw), &base);
  xx = armor64A(raw, sizeof(raw));
  out << "base: chord:" << xx << "\n";
  mpz_get_rawmag_be(raw, sizeof(raw), &ver1);
  xx = armor64A(raw, sizeof(raw));
  out << "view: chord:" << xx << "\n";
  mpz_get_rawmag_be(raw, sizeof(raw), &ver2);
  xx = armor64A(raw, sizeof(raw));
  out << "view: chord:" << xx << "\n";
}

void
report_conflicts(nfstime3 t)
{
  strbuf out;
  char ts[32];
  ctime_r(&((time_t)t.seconds), ts);
  ts[strlen(ts)-1] = '\0';
  out << "conflicts starting at: " << ts << "\n";
  for (unsigned i=0; i<_conflict_inodes.size(); i++) {
    strbuf s;
    sid(s, _conflict_inodes[i]);
    str ss = armor64A(s);
    out << "update on inode " << ss << "\n";
  }
  for (unsigned i=0; i<_conflict_dirops.size(); i++) {
    dirop *d = _dirop_mapping[_conflict_dirops[i]];
    assert(d);
    strbuf s;
    sid(s, d->dir);
    str ss = armor64A(s);
    out << "operation on file " << d->name << "\n"
        << "     in directory " << ss << "\n";
  }
  
  if (_conflict_inodes.size() > 0 || _conflict_dirops.size() > 0) 
    split_conflict(_conflict_v1, _conflict_v2, out);
 
  out << "\n";
  out.tosuio ()->output (1);
  
  _conflict_inodes.clear();
  _conflict_dirops.clear();
  _inode_ops.clear();
  _dir_ops.clear();
  _dirop_mapping.clear();
}

void
add_inode_op(chordID inum, ptr<log_record> e)
{
  chordID *n = _inode_ops[inum];
  if (!n)
    _inode_ops.insert(inum, e->headID());
  else if (n && *n != e->headID())
    _conflict_inodes.push_back(inum);
}

void
add_dir_op(chordID inum, str name, ptr<log_record> e)
{
  strbuf s;
  sid(s, inum);
  s << name;
  str ss = s;
  bigint id = compute_hash(ss.cstr(), ss.len());
  chordID *n = _dir_ops[id];
  if (!n) {
    struct dirop d;
    d.dir = inum;
    d.name = name;
    _dir_ops.insert(id, e->headID());
    _dirop_mapping.insert(id, d);
  }
  else if (n && *n != e->headID())
    _conflict_dirops.push_back(id);
}

void
error_cb(iterinfo *i)
{
  if (i->error) {
    strbuf out;
    if (i->error != logiter::ERROR_NO_VIEW) {
      strbuf ll;
      sid(ll, i->hid);
      str l = armor64A(ll);
      strbuf ee;
      sid(ee, i->eid);
      str e = armor64A(ee);

      switch(i->error) {
        case logiter::ERROR_NO_RECORD: 
	  {
	    out << "log " << l << ": missing record\n"
	        << e << ", skipping this log\n";
	    i->finished = false;
	    i->eol = false;
	  }
	  break;
        case logiter::ERROR_BAD_SEQN:
	  {
	    out << "log " << l
	        << ": non decreasing sequence number at record\n"
	        << e << ", skipping this log\n";
	    i->finished = false;
	    i->eol = false;
	  }
	  break;
        case logiter::ERROR_SELF_CONFLICT:
	  {
	    out << "log " << l
	        << ": conflicting version number detected at record\n"
	        << e << ", skipping this log\n";
	    i->finished = false;
	    i->eol = false;
	  }
	  break;
	case logiter::WARN_CONCURRENT:
        case logiter::WARN_CONFLICT:
	  {
	    if (!_resolve_started) {
	      _resolve_started = true;
	      assert(i->record && i->conflict);
	      
	      if (verbose) {
                strbuf ll2;
                sid(ll2, i->conflict->headID());
                str l2 = armor64A(ll2);
                strbuf ee2;
                sid(ee2, i->conflict->selfID());
                str e2 = armor64A(ee2);
	        out << "record " << e 
	            << "of log " << l;
	        if (i->error == logiter::WARN_CONFLICT)
	          out << "\nconflicts with\n";
	        else
	          out << "\nconcurrent with\n";
	        out << "record " << e2
		    << "of log " << l2 << "\n";
	      }
	      
	      vervec v1 = i->record->version();
	      _conflict_v1 = v1;
	      for(unsigned j=0; j<v1.size(); j++) {
                if (v1.head(j) == i->record->headID()) {
		  v1.add(v1.head(j),i->record->selfID(),i->record->seqnum());
		  break;
	        }
	      }
	      vervec v2 = i->conflict->version();
	      _conflict_v2 = v2;
	      for(unsigned j=0; j<v2.size(); j++) {
                if (v2.head(j) == i->conflict->headID()) {
		  v2.add
		    (v2.head(j),i->conflict->selfID(),i->conflict->seqnum());
		  break;
	        }
	      }
              _base = vervec::merge_smaller(v1, v2);
	      if (verbose) {
	        out << "base version:\n";
	        print_vervec(0, _base, out);
                out.tosuio ()->output (1);
	      }
	    }
	    i->finished = false;
	    i->eol = false;
	    return;
	  }
	  break;
        default:
	  break;
      }
    }
    else {
      out << "no view\n";
      out.tosuio ()->output (1);
      exit(0);
    }
  }
  else if (!i->finished && _resolve_started) {
    if (_base.equals(i->record->version()))
      _ctime = i->record->ts();
    if (_base.gt(i->record->version())) {
      report_conflicts(_ctime);
      _resolve_started = false;
    }
    else {
      if (verbose) {
        strbuf out;
        vervec v = i->record->version();
        out << "checking record with version:\n";
        print_vervec(0, v, out);
        out.tosuio ()->output (1);
      }
  
      // detect data conflicts: data conflicts occur if there are ATTR,
      // WRITE or INODE entries on the same inode number, or if there
      // are LINK/UNLINK/RENAME or INTEND entries on the same name.
  
      if (i->record->template record<attr_log_record>())
        add_inode_op
	  (i->record->template record<attr_log_record>()->inum(), i->record);
      else if (i->record->template record<write_log_record>())
        add_inode_op
	  (i->record->template record<write_log_record>()->inum(), i->record);
      else if (i->record->template record<inode_log_record>())
        add_inode_op
	  (i->record->template record<inode_log_record>()->inum(), i->record);
      else if (i->record->template record<intend_log_record>())
        add_dir_op
	  (i->record->template record<intend_log_record>()->dir(),
	   i->record->template record<intend_log_record>()->name(), i->record);
      else if (i->record->template record<link_log_record>()) {
        add_dir_op
	  (i->record->template record<link_log_record>()->dir(),
	   i->record->template record<link_log_record>()->name(), i->record);
      }
      else if (i->record->template record<unlink_log_record>()) {
        add_dir_op
	  (i->record->template record<unlink_log_record>()->dir(),
	   i->record->template record<unlink_log_record>()->name(), i->record);
      }
      else if (i->record->template record<rename_log_record>()) {
        add_dir_op
	  (i->record->template record<rename_log_record>()->dir(),
	   i->record->template record<rename_log_record>()->name(), i->record);
        add_dir_op
	  (i->record->template record<rename_log_record>()->olddir(),
	   i->record->template record<rename_log_record>()->oldname(),
	   i->record);
      }
    }
  }
  else if (i->finished) {
    _work--;
    if (_work == 0)
      exit(0);
  }
}

void
got_view(chordID vid, ptr<view> ok)
{
  if (!ok) {
    warn << "cannot load view\n";
    exit(1);
  }
  _work++;
  New refcounted<logiter>(wrap(&error_cb), vid, error);
}

void
usage()
{
  warnx << "usage: \n";
  warnx << "  " << progname << " [-v] viewid\n"
        << "    prints out information on the given view. if -v is specified,\n"
        << "    iterates and prints all the log entries as well\n";
  warnx << "  " << progname << " -p viewid\n"
        << "    prints out a public key hash in the view\n";
  warnx << "  " << progname << " -c key\n"
        << "    creates a new view with the log stored at the given key\n";
  warnx << "  " << progname << " -a key[,latest] viewid\n"
        << "    adds the log stored at the given key to an existing view\n";
  warnx << "  " << progname << " -d time viewid\n"
        << "    creates a new view that includes all states in the given view\n"
        << "    before specified time. e.g., \"Apr 1 16:18:27 2002\"\n";
  warnx << "  " << progname << " [-v] -x viewid\n"
        << "    scan file system for conflicting operations.\n";
  warnx << "  " << progname << " [-v] -s viewid\n"
        << "    scan file system for conflicting and concurrent operations.\n";
  warnx << "viewid can be the hash of a view's chord ID, or the mount point\n";
  exit(1);
}

int
main (int argc, char **argv)
{
  setprogname (argv[0]);
  if (argc != 2)
    usage();

  char *sn = getenv("LSD_SOCKET");
  if(sn == 0){
    warnx << "viewctl: LSD_SOCKET is not set\n";
    exit(1);
  }
   
  cfsrw_dhashclt::init();

  chordID vid;
  char sym[1024];
  str dearm;
  rxx path_regexp ("chord:([0-9a-zA-Z+-]*)[/]*$");
  int r = readlink(argv[1], sym, 1024);
  if (r > 0) sym[r] = '\0';
  if (path_regexp.search (argv[1]) || (r > 0 && path_regexp.search(sym)))
    dearm = dearmor64A(path_regexp[1]);
  else
    dearm = dearmor64A(argv[1]);
  if (dearm) 
    mpz_set_rawmag_be(&vid, dearm.cstr(), dearm.len());
  else {
    warnx << "viewctl: key hash invalid\n";
    exit(1);
  }
  _view = New refcounted<view>(wrap(&got_view, vid), vid);
  amain();
}

