/*
 * Copyright (C) 2002  Benjie Chen (benjie@lcs.mit.edu)
 *   		       Massachusetts Institute of Technology
 * 
 *  Permission is hereby granted, free of charge, to any person
 *  obtaining a copy of this software and associated documentation
 *  files (the "Software"), to deal in the Software without
 *  restriction, including without limitation the rights to use, copy,
 *  modify, merge, publish, distribute, sublicense, and/or sell copies
 *  of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be
 *  included in all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 *  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 *  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 *  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 *  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 *  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.
 */

#include "sfsmisc.h"
#include "dhash.h"
#include "dhash_prot.h"
#include "crypt.h"
#include "parseopt.h"
#include "serialize.h"
#include "vervec.h"
#include "cache.h"
#include "log.h"
#include "logiter.h"
#include "rxx.h"
#include "qhash.h"
#include <sys/time.h>

static ptr<view> _view;

static vec<chordID> _heads;
static vec<chordID> _latest;
static ptr<logiter> _iter;
static chordID _root;
static bool verbose = false;
static int _work = 0;

void
store_cb(bool print, bool failure, chordID id)
{
  _work--;
  if (failure)
    warn << "failed to store view\n";
  else if (print) {
    char raw[20];
    mpz_get_rawmag_be(raw, sizeof(raw), &id);
    str xx = armor64A(raw, sizeof(raw));
    strbuf out;
    out << "chord:" << xx << "\n";
    out.tosuio ()->output (1);
  }
  if (_work == 0)
    exit(0);
}

void
print_vervec(unsigned n, vervec &v, strbuf &out)
{
  for (unsigned i=0; i<v.size(); i++) {
    strbuf hh;
    sid(hh, v.head(i));
    str h = armor64A(hh);
    strbuf ll;
    sid(ll, v.latest(i));
    str l = armor64A(ll);
    for (unsigned j=0; j<n; j++) out << " ";
    out << " verh " << h << "\n";
    for (unsigned j=0; j<n; j++) out << " ";
    out << " verl " << l << "\n";
    for (unsigned j=0; j<n; j++) out << " ";
    out << " vers " << v.seqnum(i) << "\n";
  }
}

void
dump_cb(iterinfo *i)
{
  if (!i->finished) {
    recordp e = i->record;
    assert(e);
    chordID hid = e->headID();
    chordID lid = e->selfID();

    strbuf out;
    char ts[32];
    nfstime3 t = e->ts();
    ctime_r(&((time_t)t.seconds), ts);
    ts[strlen(ts)-1] = '\0';
    strbuf r;
    sid(r, hid);
    str head = armor64A(r);
    strbuf w;
    sid(w, lid);
    str where = armor64A(w);
         
    out << ts << "   seq " << e->seqnum() << "\n                        "
              << "   log " << where << "\n                        "
	      << "  head " <<  head << "\n";
    vervec v = e->version();
    print_vervec(25, v, out);
    for(unsigned j=0; j<25; j++) out << " ";

    switch(e->type()) {
      case log_record::T_INODE:
	{
          ptr<inode_log_record> l =
	    *reinterpret_cast<ptr<inode_log_record>*>(&e);
	  switch(l->itype()) {
	    case inode_log_record::T_FILE:
	      out << " file ";
	      break;
	    case inode_log_record::T_DIR:
	      out << "  dir ";
	      break;
	    case inode_log_record::T_SYMLINK:
	      out << " syml ";
	      break;
	    default:
	      break;
	  }
	  strbuf b;
	  sid(b, l->inum());
	  out << armor64A(b) << "\n";
	}
	break;
      case log_record::T_INTEND:
	{
	  out << "itend ";
          intend_log_record *ie = e->template record<intend_log_record>();
	  strbuf x;
	  sid(x, ie->verf());
	  strbuf y;
	  sid(y, ie->dir());
	  out << ie->name()  << "\n                            in "
	      << armor64A(y) << "\n                          verf "
	      << armor64A(x) << "\n                        cancel "
	      << ie->canceled() << "\n";
	}
	break;
      case log_record::T_RENAME:
	{
	  out << "rname ";
          ptr<rename_log_record> l =
	    *reinterpret_cast<ptr<rename_log_record>*>(&e);
	  strbuf x;
	  sid(x, l->inum());
	  out << armor64A(x) << "\n                            in ";
	  strbuf y;
	  sid(y, l->dir());
	  out << armor64A(y) << "\n                            as "
	      << l->name() << "\n";
	  out << "                           was " << l->oldname() << "\n";
	}
	break;
      case log_record::T_LINK:
	{
	  out << " link ";
          ptr<link_log_record> l =
	    *reinterpret_cast<ptr<link_log_record>*>(&e);
	  strbuf x;
	  sid(x, l->inum());
	  out << armor64A(x) << "\n                            in ";
	  strbuf y;
	  sid(y, l->dir());
	  out << armor64A(y) << "\n                            as "
	      << l->name() << "\n";
	}
	break;
      case log_record::T_UNLINK:
	{
	  out << "ulink ";
          ptr<unlink_log_record> l =
	    *reinterpret_cast<ptr<unlink_log_record>*>(&e);
	  strbuf x;
	  sid(x, l->inum());
	  out << armor64A(x) << "\n                            in ";
	  strbuf y;
	  sid(y, l->dir());
	  out << armor64A(y) << "\n                            as "
	      << l->name() << "\n";
	}
	break;
      case log_record::T_WRITE:
	{
          out << "write ";
          ptr<write_log_record> l =
	    *reinterpret_cast<ptr<write_log_record>*>(&e);
	  strbuf x;
	  sid(x, l->inum());
	  out << armor64A(x) << "\n                            at "
	      << l->offset() << ", " << l->count() << " bytes\n";
	}
	break;
      case log_record::T_ATTR:
	{
	  out << " attr ";
          ptr<attr_log_record> l =
	    *reinterpret_cast<ptr<attr_log_record>*>(&e);
	  strbuf x;
	  sid(x, l->inum());
	  out << armor64A(x) << "\n                          size ";
	  if (l->attr()->valid_size)
	    out << l->attr()->size << " bytes\n";
	  else
	    out << " not set\n";
	}
	break;
      case log_record::T_END:
      case log_record::T_HEAD:
      default:
	break;
    }
    if (e->type() != log_record::T_END && e->type() != log_record::T_HEAD)
      out.tosuio ()->output (1);
  }
  else if (i->finished) {
    _work--;
    if (_work == 0)
      exit(0);
  }
}

void
new_view(chordID key, recordp l)
{
  if (l && l->type() == log_record::T_HEAD) {
    ref<view> v = New refcounted<view>();
    head_log_record *h = l->template record<head_log_record>();
    v->root(h->root());
    v->add_log(key, key);
    _work++;
    v->store(wrap(store_cb, true));
  }
  else {
    exit(1);
  }
}

void
modify_view(chordID key, chordID latest, ptr<view> ok)
{
  if (!ok) {
    warn << "cannot load view\n";
    exit(1);
  }
  _view->add_log(key,latest);
  _work++;
  _view->store(wrap(store_cb, true));
}

void
got_view(chordID vid, bool printkey, ptr<view> ok)
{
  if (!ok) {
    warn << "cannot load view\n";
    exit(1);
  }

  char raw[20];
  str xx;
  strbuf out;
  chordID n;
  
  if (!printkey) {
    n = _view->root();
    mpz_get_rawmag_be(raw, sizeof(raw), &n);
    xx = armor64A(raw, sizeof(raw));
    out << "root inum: " << xx << "\n";
  }

  for (unsigned i=0; i<_view->nlogs(); i++) {
    n = _view->head(i);
    mpz_get_rawmag_be(raw, sizeof(raw), &n);
    xx = armor64A(raw, sizeof(raw));
    if (!printkey) {
      out << "raw: " << n << "\n";
      out << "key: " << xx << ", latest: ";
      n = _view->latest(i);
      mpz_get_rawmag_be(raw, sizeof(raw), &n);
      xx = armor64A(raw, sizeof(raw));
    }
    out << xx << "\n";
    if (printkey)
      break;
  }
  out.tosuio ()->output (1);

  if (verbose) {
    _work++;
    New refcounted<logiter>(wrap(&dump_cb), vid);
  }
  else
    exit(0);
}

void
usage()
{
  warnx << "usage: \n";
  warnx << "  " << progname << " [-v] viewid\n"
        << "    prints out information on the given view. if -v is specified,\n"
        << "    iterates and prints all the log entries as well\n";
  warnx << "  " << progname << " -c key\n"
        << "    creates a new view with the log stored at the given key\n";
  warnx << "  " << progname << " -a key[,latest] viewid\n"
        << "    adds the log stored at the given key to an existing view\n";
  warnx << "  " << progname << " -p viewid\n"
        << "    prints out a public key hash in the view\n";
  warnx << "viewid can be the hash of a view's chord ID, or the mount point\n";
  exit(1);
}

int
main (int argc, char **argv)
{
  bool printkey = false;
  char *newkey = 0;
  char *addkey = 0;
  setprogname (argv[0]);
  int opt = 0;

  int ch;
  while ((ch = getopt (argc, argv, "a:c:vp")) != -1)
    switch (ch) {
    case 'p':
      opt++;
      printkey = true;
      break;
    case 'c':
      opt++;
      newkey = optarg;
      break;
    case 'a':
      opt++;
      addkey = optarg;
      break;
    case 'v':
      verbose = true;
      break;
    case '?':
    default:
      usage ();
    }
  argc -= optind;
  argv += optind;

  if ((newkey && argc > 0) || (!newkey && argc != 1) || opt > 1)
    usage();

  char *sn = getenv("LSD_SOCKET");
  if(sn == 0){
    warnx << "viewctl: LSD_SOCKET is not set\n";
    exit(1);
  }
   
  cfsrw_dhashclt::init();

  chordID vid;
  if (argc == 1) {
    char sym[1024];
    str dearm;
    rxx path_regexp ("chord:([0-9a-zA-Z+-]*)[/]*$");
    int r = readlink(argv[0], sym, 1024);
    if (r > 0) sym[r] = '\0';
    if (path_regexp.search (argv[0]) || (r > 0 && path_regexp.search(sym)))
      dearm = dearmor64A(path_regexp[1]);
    else
      dearm = dearmor64A(argv[0]);
    if (dearm) 
      mpz_set_rawmag_be(&vid, dearm.cstr(), dearm.len());
    else {
      warnx << "viewctl: key hash invalid\n";
      exit(1);
    }
  }

  if (newkey) {
    str dearm = dearmor64A(newkey);
    if (dearm) {
      chordID key;
      mpz_set_rawmag_be(&key, dearm.cstr(), dearm.len());
      log::locate(key, true, wrap(new_view, key));
    }
    else {
      warnx << "viewctl: key hash invalid\n";
      exit(1);
    }
  }
  else if (addkey) {
    chordID key;
    chordID latest;
    str keystr;
    str lateststr;

    rxx wb_regexp ("([0-9a-zA-Z+-]*),([0-9a-zA-Z+-]*)$");
    if (wb_regexp.search (addkey)) {
      keystr = wb_regexp[1];
      lateststr = wb_regexp[2];
    }
    else {
      keystr = addkey;
      lateststr = addkey;
    }
    str dearm = dearmor64A(keystr);
    if (dearm) {
      mpz_set_rawmag_be(&key, dearm.cstr(), dearm.len());
      dearm = dearmor64A(lateststr);
      mpz_set_rawmag_be(&latest, dearm.cstr(), dearm.len());
      _view = New refcounted<view>(wrap(&modify_view, key, latest), vid);
    }
    else {
      warnx << "viewctl: key hash invalid\n";
      exit(1);
    }
  }
  else
    _view = New refcounted<view>(wrap(&got_view, vid, printkey), vid);
  amain();
}

